/*
 * Qcint is a general GTO integral library for computational chemistry
 * Copyright (C) 2014- Qiming Sun <osirpt.sun@gmail.com>
 *
 * This file is part of Qcint.
 *
 * Qcint is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Description: code generated by  gen-code.cl
 */
#include <pmmintrin.h>
#include <stdlib.h>
#include "cint_bas.h"
#include "cart2sph.h"
#include "g2e.h"
#include "g3c2e.h"
#include "g2c2e.h"
#include "optimizer.h"
#include "cint1e.h"
#include "cint2e.h"
#include "misc.h"
#include "fblas.h"
#include "c2f.h"
/* <i|OVLP |j> */
static void CINTgout1e_cint1e_ovlp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double s[1];
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g0[ix] * g0[iy] * g0[iz];
gout[0] += + s[0];
gout += 1;
}}
FINT cint1e_ovlp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {0, 0, 0, 0, 0, 1, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_ovlp;
return CINT1e_drv(opij, &envs, 1, &c2s_sf_1e); }
C2F_(cint1e_ovlp)
/* <i|NUC |j> */
static void CINTgout1e_cint1e_nuc(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double s[1];
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g0[ix] * g0[iy] * g0[iz];
gout[0] += + s[0];
gout += 1;
}}
FINT cint1e_nuc(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {0, 0, 0, 0, 0, 1, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_nuc;
return CINT1e_nuc_drv(opij, &envs, 1, &c2s_sf_1e); }
C2F_(cint1e_nuc)
/* <SIGMA DOT R i|OVLP |SIGMA DOT R j> */
static void CINTgout1e_cint1e_srsr(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
G1E_R_J(g1, g0, i_l+1, j_l+0);
G1E_R_I(g2, g0, i_l+0, j_l);
G1E_R_I(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + s[5] + (-1*s[7]);
gout[1] += + s[6] + (-1*s[2]);
gout[2] += + s[1] + (-1*s[3]);
gout[3] += + s[0] + s[4] + s[8];
gout += 4;
}}
FINT cint1e_srsr(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_srsr;
return CINT1e_drv(opij, &envs, 1, &c2s_si_1e); }
C2F_(cint1e_srsr)
/* <SIGMA DOT R i|OVLP |j> */
static void CINTgout1e_cint1e_sr(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double s[3];
G1E_R_I(g1, g0, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g1[ix] * g0[iy] * g0[iz];
s[1] = g0[ix] * g1[iy] * g0[iz];
s[2] = g0[ix] * g0[iy] * g1[iz];
gout[0] += + (-1*s[0]);
gout[1] += + (-1*s[1]);
gout[2] += + (-1*s[2]);
gout[3] += 0;
gout += 4;
}}
FINT cint1e_sr(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 0, 0, 1, 4, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_sr;
return CINT1e_drv(opij, &envs, 1, &c2s_si_1ei); }
C2F_(cint1e_sr)
/* <SIGMA DOT R i|OVLP |SIGMA DOT P j> */
static void CINTgout1e_cint1e_srsp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
G1E_D_J(g1, g0, i_l+1, j_l+0);
G1E_R_I(g2, g0, i_l+0, j_l);
G1E_R_I(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + (-1*s[5]) + s[7];
gout[1] += + (-1*s[6]) + s[2];
gout[2] += + (-1*s[1]) + s[3];
gout[3] += + (-1*s[0]) + (-1*s[4]) + (-1*s[8]);
gout += 4;
}}
FINT cint1e_srsp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_srsp;
return CINT1e_drv(opij, &envs, 1, &c2s_si_1ei); }
C2F_(cint1e_srsp)
/* <SIGMA DOT P i|OVLP |SIGMA DOT P j> */
static void CINTgout1e_cint1e_spsp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
G1E_D_J(g1, g0, i_l+1, j_l+0);
G1E_D_I(g2, g0, i_l+0, j_l);
G1E_D_I(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + s[0] + s[4] + s[8];
gout += 1;
}}
FINT cint1e_spsp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 1, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_spsp;
return CINT1e_drv(opij, &envs, 1, &c2s_sf_1e); }
C2F_(cint1e_spsp)
/* <SIGMA DOT P i|OVLP |j> */
static void CINTgout1e_cint1e_sp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double s[3];
G1E_D_I(g1, g0, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g1[ix] * g0[iy] * g0[iz];
s[1] = g0[ix] * g1[iy] * g0[iz];
s[2] = g0[ix] * g0[iy] * g1[iz];
gout[0] += + s[0];
gout[1] += + s[1];
gout[2] += + s[2];
gout[3] += 0;
gout += 4;
}}
FINT cint1e_sp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 0, 0, 1, 4, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_sp;
return CINT1e_drv(opij, &envs, 1, &c2s_si_1e); }
C2F_(cint1e_sp)
/* <SIGMA DOT P i|NUC |SIGMA DOT P j> */
static void CINTgout1e_cint1e_spnucsp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
G1E_D_J(g1, g0, i_l+1, j_l+0);
G1E_D_I(g2, g0, i_l+0, j_l);
G1E_D_I(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + s[5] + (-1*s[7]);
gout[1] += + s[6] + (-1*s[2]);
gout[2] += + s[1] + (-1*s[3]);
gout[3] += + s[0] + s[4] + s[8];
gout += 4;
}}
FINT cint1e_spnucsp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_spnucsp;
return CINT1e_nuc_drv(opij, &envs, 1, &c2s_si_1e); }
C2F_(cint1e_spnucsp)
/* <SIGMA DOT R i|NUC |SIGMA DOT R j> */
static void CINTgout1e_cint1e_srnucsr(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
G1E_R_J(g1, g0, i_l+1, j_l+0);
G1E_R_I(g2, g0, i_l+0, j_l);
G1E_R_I(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + s[5] + (-1*s[7]);
gout[1] += + s[6] + (-1*s[2]);
gout[2] += + s[1] + (-1*s[3]);
gout[3] += + s[0] + s[4] + s[8];
gout += 4;
}}
FINT cint1e_srnucsr(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_srnucsr;
return CINT1e_nuc_drv(opij, &envs, 1, &c2s_si_1e); }
C2F_(cint1e_srnucsr)
/* <G i|OVLP |j> */
static void CINTgout1e_cint1e_govlp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double s[3];
double rirj[3], c[3];
rirj[0] = ri[0] - rj[0];
rirj[1] = ri[1] - rj[1];
rirj[2] = ri[2] - rj[2];
c[0] = 1 * rirj[0];
c[1] = 1 * rirj[1];
c[2] = 1 * rirj[2];
G1E_R0I(g1, g0, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g1[ix] * g0[iy] * g0[iz];
s[1] = g0[ix] * g1[iy] * g0[iz];
s[2] = g0[ix] * g0[iy] * g1[iz];
gout[0] += + c[1]*s[2] + (-1*c[2]*s[1]);
gout[1] += + c[2]*s[0] + (-1*c[0]*s[2]);
gout[2] += + c[0]*s[1] + (-1*c[1]*s[0]);
gout += 3;
}}
FINT cint1e_govlp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 0, 0, 1, 1, 0, 3};
const FINT i_sh = shls[0];
const FINT j_sh = shls[1];
const FINT i_l = bas(ANG_OF, i_sh);
const FINT j_l = bas(ANG_OF, j_sh);if (bas(ATOM_OF, i_sh) == bas(ATOM_OF, j_sh)) {
FINT ip = CINTlen_spinor(i_sh, bas) * bas(NCTR_OF,i_sh);
FINT jp = CINTlen_spinor(j_sh, bas) * bas(NCTR_OF,j_sh);
CINTdset0(ip * jp * OF_CMPLX * ng[TENSOR], opij);
return 0; }
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_govlp;
return CINT1e_drv(opij, &envs, 0.5, &c2s_sf_1ei); }
C2F_(cint1e_govlp)
/* <G i|NUC |j> */
static void CINTgout1e_cint1e_gnuc(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double s[3];
double rirj[3], c[3];
rirj[0] = ri[0] - rj[0];
rirj[1] = ri[1] - rj[1];
rirj[2] = ri[2] - rj[2];
c[0] = 1 * rirj[0];
c[1] = 1 * rirj[1];
c[2] = 1 * rirj[2];
G1E_R0I(g1, g0, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g1[ix] * g0[iy] * g0[iz];
s[1] = g0[ix] * g1[iy] * g0[iz];
s[2] = g0[ix] * g0[iy] * g1[iz];
gout[0] += + c[1]*s[2] + (-1*c[2]*s[1]);
gout[1] += + c[2]*s[0] + (-1*c[0]*s[2]);
gout[2] += + c[0]*s[1] + (-1*c[1]*s[0]);
gout += 3;
}}
FINT cint1e_gnuc(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 0, 0, 1, 1, 0, 3};
const FINT i_sh = shls[0];
const FINT j_sh = shls[1];
const FINT i_l = bas(ANG_OF, i_sh);
const FINT j_l = bas(ANG_OF, j_sh);if (bas(ATOM_OF, i_sh) == bas(ATOM_OF, j_sh)) {
FINT ip = CINTlen_spinor(i_sh, bas) * bas(NCTR_OF,i_sh);
FINT jp = CINTlen_spinor(j_sh, bas) * bas(NCTR_OF,j_sh);
CINTdset0(ip * jp * OF_CMPLX * ng[TENSOR], opij);
return 0; }
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_gnuc;
return CINT1e_nuc_drv(opij, &envs, 0.5, &c2s_sf_1ei); }
C2F_(cint1e_gnuc)
/* <SIGMA CROSS RC i|SIGMA CROSS NABLA-RINV |j> */
static void CINTgout1e_cint1e_cg_sa10sa01(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
double dri[3];
dri[0] = ri[0] - env[PTR_COMMON_ORIG+0];
dri[1] = ri[1] - env[PTR_COMMON_ORIG+1];
dri[2] = ri[2] - env[PTR_COMMON_ORIG+2];
G1E_D_J(g1, g0, i_l+1, j_l+0);
G1E_D_I(g2, g0, i_l+1, j_l+0);
for (ix = 0; ix < envs->g_size * 3; ix++) {g1[ix] += g2[ix];}
G1E_RCI(g2, g0, i_l+0, j_l);
G1E_RCI(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + (-1*s[7]) + s[5];
gout[1] += 0;
gout[2] += 0;
gout[3] += + s[8] + s[4];
gout[4] += + s[6];
gout[5] += + s[5];
gout[6] += + s[8];
gout[7] += + (-1*s[3]);
gout[8] += + (-1*s[3]);
gout[9] += + (-1*s[4]);
gout[10] += + (-1*s[7]);
gout[11] += + (-1*s[6]);
gout[12] += + (-1*s[2]);
gout[13] += + (-1*s[7]);
gout[14] += + (-1*s[8]);
gout[15] += + (-1*s[1]);
gout[16] += 0;
gout[17] += + (-1*s[2]) + s[6];
gout[18] += 0;
gout[19] += + s[8] + s[0];
gout[20] += + s[0];
gout[21] += + s[1];
gout[22] += + s[6];
gout[23] += + (-1*s[7]);
gout[24] += + s[1];
gout[25] += + s[4];
gout[26] += + s[5];
gout[27] += + (-1*s[2]);
gout[28] += + (-1*s[0]);
gout[29] += + (-1*s[3]);
gout[30] += + (-1*s[2]);
gout[31] += + (-1*s[5]);
gout[32] += 0;
gout[33] += 0;
gout[34] += + (-1*s[3]) + s[1];
gout[35] += + s[4] + s[0];
gout += 36;
}}
FINT cint1e_cg_sa10sa01(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 0, 9};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_cg_sa10sa01;
return CINT1e_rinv_drv(opij, &envs, 0.5, &c2s_si_1e); }
C2F_(cint1e_cg_sa10sa01)
/* <RC CROSS SIGMA i|OVLP |SIGMA DOT P j> */
static void CINTgout1e_cint1e_cg_sa10sp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
double dri[3];
dri[0] = ri[0] - env[PTR_COMMON_ORIG+0];
dri[1] = ri[1] - env[PTR_COMMON_ORIG+1];
dri[2] = ri[2] - env[PTR_COMMON_ORIG+2];
G1E_D_J(g1, g0, i_l+1, j_l+0);
G1E_RCI(g2, g0, i_l+0, j_l);
G1E_RCI(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + s[8] + s[4];
gout[1] += + (-1*s[3]);
gout[2] += + (-1*s[6]);
gout[3] += + s[7] + (-1*s[5]);
gout[4] += + (-1*s[1]);
gout[5] += + s[0] + s[8];
gout[6] += + (-1*s[7]);
gout[7] += + (-1*s[6]) + s[2];
gout[8] += + (-1*s[2]);
gout[9] += + (-1*s[5]);
gout[10] += + s[4] + s[0];
gout[11] += + s[3] + (-1*s[1]);
gout += 12;
}}
FINT cint1e_cg_sa10sp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 0, 3};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_cg_sa10sp;
return CINT1e_drv(opij, &envs, 0.5, &c2s_si_1ei); }
C2F_(cint1e_cg_sa10sp)
/* <RC CROSS SIGMA i|NUC |SIGMA DOT P j> */
static void CINTgout1e_cint1e_cg_sa10nucsp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
double dri[3];
dri[0] = ri[0] - env[PTR_COMMON_ORIG+0];
dri[1] = ri[1] - env[PTR_COMMON_ORIG+1];
dri[2] = ri[2] - env[PTR_COMMON_ORIG+2];
G1E_D_J(g1, g0, i_l+1, j_l+0);
G1E_RCI(g2, g0, i_l+0, j_l);
G1E_RCI(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + s[8] + s[4];
gout[1] += + (-1*s[3]);
gout[2] += + (-1*s[6]);
gout[3] += + s[7] + (-1*s[5]);
gout[4] += + (-1*s[1]);
gout[5] += + s[0] + s[8];
gout[6] += + (-1*s[7]);
gout[7] += + (-1*s[6]) + s[2];
gout[8] += + (-1*s[2]);
gout[9] += + (-1*s[5]);
gout[10] += + s[4] + s[0];
gout[11] += + s[3] + (-1*s[1]);
gout += 12;
}}
FINT cint1e_cg_sa10nucsp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 0, 3};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_cg_sa10nucsp;
return CINT1e_nuc_drv(opij, &envs, 0.5, &c2s_si_1ei); }
C2F_(cint1e_cg_sa10nucsp)
/* <SIGMA CROSS R i|SIGMA CROSS NABLA-RINV |j> */
static void CINTgout1e_cint1e_giao_sa10sa01(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
G1E_D_J(g1, g0, i_l+1, j_l+0);
G1E_D_I(g2, g0, i_l+1, j_l+0);
for (ix = 0; ix < envs->g_size * 3; ix++) {g1[ix] += g2[ix];}
G1E_R_I(g2, g0, i_l+0, j_l);
G1E_R_I(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + (-1*s[7]) + s[5];
gout[1] += 0;
gout[2] += 0;
gout[3] += + s[8] + s[4];
gout[4] += + s[6];
gout[5] += + s[5];
gout[6] += + s[8];
gout[7] += + (-1*s[3]);
gout[8] += + (-1*s[3]);
gout[9] += + (-1*s[4]);
gout[10] += + (-1*s[7]);
gout[11] += + (-1*s[6]);
gout[12] += + (-1*s[2]);
gout[13] += + (-1*s[7]);
gout[14] += + (-1*s[8]);
gout[15] += + (-1*s[1]);
gout[16] += 0;
gout[17] += + (-1*s[2]) + s[6];
gout[18] += 0;
gout[19] += + s[8] + s[0];
gout[20] += + s[0];
gout[21] += + s[1];
gout[22] += + s[6];
gout[23] += + (-1*s[7]);
gout[24] += + s[1];
gout[25] += + s[4];
gout[26] += + s[5];
gout[27] += + (-1*s[2]);
gout[28] += + (-1*s[0]);
gout[29] += + (-1*s[3]);
gout[30] += + (-1*s[2]);
gout[31] += + (-1*s[5]);
gout[32] += 0;
gout[33] += 0;
gout[34] += + (-1*s[3]) + s[1];
gout[35] += + s[4] + s[0];
gout += 36;
}}
FINT cint1e_giao_sa10sa01(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 0, 9};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_giao_sa10sa01;
return CINT1e_rinv_drv(opij, &envs, 0.5, &c2s_si_1e); }
C2F_(cint1e_giao_sa10sa01)
/* <R CROSS SIGMA i|OVLP |SIGMA DOT P j> */
static void CINTgout1e_cint1e_giao_sa10sp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
G1E_D_J(g1, g0, i_l+1, j_l+0);
G1E_R_I(g2, g0, i_l+0, j_l);
G1E_R_I(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + s[8] + s[4];
gout[1] += + (-1*s[3]);
gout[2] += + (-1*s[6]);
gout[3] += + s[7] + (-1*s[5]);
gout[4] += + (-1*s[1]);
gout[5] += + s[0] + s[8];
gout[6] += + (-1*s[7]);
gout[7] += + (-1*s[6]) + s[2];
gout[8] += + (-1*s[2]);
gout[9] += + (-1*s[5]);
gout[10] += + s[4] + s[0];
gout[11] += + s[3] + (-1*s[1]);
gout += 12;
}}
FINT cint1e_giao_sa10sp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 0, 3};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_giao_sa10sp;
return CINT1e_drv(opij, &envs, 0.5, &c2s_si_1ei); }
C2F_(cint1e_giao_sa10sp)
/* <R CROSS SIGMA i|NUC |SIGMA DOT P j> */
static void CINTgout1e_cint1e_giao_sa10nucsp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
G1E_D_J(g1, g0, i_l+1, j_l+0);
G1E_R_I(g2, g0, i_l+0, j_l);
G1E_R_I(g3, g1, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + s[8] + s[4];
gout[1] += + (-1*s[3]);
gout[2] += + (-1*s[6]);
gout[3] += + s[7] + (-1*s[5]);
gout[4] += + (-1*s[1]);
gout[5] += + s[0] + s[8];
gout[6] += + (-1*s[7]);
gout[7] += + (-1*s[6]) + s[2];
gout[8] += + (-1*s[2]);
gout[9] += + (-1*s[5]);
gout[10] += + s[4] + s[0];
gout[11] += + s[3] + (-1*s[1]);
gout += 12;
}}
FINT cint1e_giao_sa10nucsp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 0, 3};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_giao_sa10nucsp;
return CINT1e_nuc_drv(opij, &envs, 0.5, &c2s_si_1ei); }
C2F_(cint1e_giao_sa10nucsp)
/* <i|NABLA-RINV CROSS SIGMA |SIGMA DOT P j> */
static void CINTgout1e_cint1e_sa01sp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double s[9];
G1E_D_J(g1, g0, i_l+0, j_l+0);
G1E_D_J(g2, g0, i_l+0, j_l+1);
G1E_D_I(g3, g0, i_l+0, j_l+1);
for (ix = 0; ix < envs->g_size * 3; ix++) {g2[ix] += g3[ix];}
G1E_D_J(g3, g2, i_l+0, j_l+0);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g3[ix] * g0[iy] * g0[iz];
s[1] = g2[ix] * g1[iy] * g0[iz];
s[2] = g2[ix] * g0[iy] * g1[iz];
s[3] = g1[ix] * g2[iy] * g0[iz];
s[4] = g0[ix] * g3[iy] * g0[iz];
s[5] = g0[ix] * g2[iy] * g1[iz];
s[6] = g1[ix] * g0[iy] * g2[iz];
s[7] = g0[ix] * g1[iy] * g2[iz];
s[8] = g0[ix] * g0[iy] * g3[iz];
gout[0] += + s[8] + s[4];
gout[1] += + (-1*s[3]);
gout[2] += + (-1*s[6]);
gout[3] += + s[7] + (-1*s[5]);
gout[4] += + (-1*s[1]);
gout[5] += + s[0] + s[8];
gout[6] += + (-1*s[7]);
gout[7] += + (-1*s[6]) + s[2];
gout[8] += + (-1*s[2]);
gout[9] += + (-1*s[5]);
gout[10] += + s[4] + s[0];
gout[11] += + s[3] + (-1*s[1]);
gout += 12;
}}
FINT cint1e_sa01sp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {0, 2, 0, 0, 2, 4, 0, 3};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_sa01sp;
return CINT1e_rinv_drv(opij, &envs, 1, &c2s_si_1ei); }
C2F_(cint1e_sa01sp)
/* <G SIGMA DOT P i|OVLP |SIGMA DOT P j> */
static void CINTgout1e_cint1e_spgsp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double *g5 = g4  + envs->g_size * 3;
double *g6 = g5  + envs->g_size * 3;
double *g7 = g6  + envs->g_size * 3;
double *g8 = g7  + envs->g_size * 3;
double s[27];
double rirj[3], c[3];
rirj[0] = ri[0] - rj[0];
rirj[1] = ri[1] - rj[1];
rirj[2] = ri[2] - rj[2];
c[0] = 1 * rirj[0];
c[1] = 1 * rirj[1];
c[2] = 1 * rirj[2];
G1E_D_J(g1, g0, i_l+2, j_l+0);
G1E_R0I(g2, g0, i_l+1, j_l);
G1E_R0I(g3, g1, i_l+1, j_l);
G1E_D_I(g4, g0, i_l+0, j_l);
G1E_D_I(g5, g1, i_l+0, j_l);
G1E_D_I(g6, g2, i_l+0, j_l);
G1E_D_I(g7, g3, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g7[ix] * g0[iy] * g0[iz];
s[1] = g6[ix] * g1[iy] * g0[iz];
s[2] = g6[ix] * g0[iy] * g1[iz];
s[3] = g5[ix] * g2[iy] * g0[iz];
s[4] = g4[ix] * g3[iy] * g0[iz];
s[5] = g4[ix] * g2[iy] * g1[iz];
s[6] = g5[ix] * g0[iy] * g2[iz];
s[7] = g4[ix] * g1[iy] * g2[iz];
s[8] = g4[ix] * g0[iy] * g3[iz];
s[9] = g3[ix] * g4[iy] * g0[iz];
s[10] = g2[ix] * g5[iy] * g0[iz];
s[11] = g2[ix] * g4[iy] * g1[iz];
s[12] = g1[ix] * g6[iy] * g0[iz];
s[13] = g0[ix] * g7[iy] * g0[iz];
s[14] = g0[ix] * g6[iy] * g1[iz];
s[15] = g1[ix] * g4[iy] * g2[iz];
s[16] = g0[ix] * g5[iy] * g2[iz];
s[17] = g0[ix] * g4[iy] * g3[iz];
s[18] = g3[ix] * g0[iy] * g4[iz];
s[19] = g2[ix] * g1[iy] * g4[iz];
s[20] = g2[ix] * g0[iy] * g5[iz];
s[21] = g1[ix] * g2[iy] * g4[iz];
s[22] = g0[ix] * g3[iy] * g4[iz];
s[23] = g0[ix] * g2[iy] * g5[iz];
s[24] = g1[ix] * g0[iy] * g6[iz];
s[25] = g0[ix] * g1[iy] * g6[iz];
s[26] = g0[ix] * g0[iy] * g7[iz];
gout[0] += + c[1]*s[17] + (-1*c[2]*s[14]) + (-1*c[1]*s[25]) + c[2]*s[22];
gout[1] += + c[1]*s[24] + (-1*c[1]*s[8]);
gout[2] += + (-1*c[2]*s[4]) + c[2]*s[12];
gout[3] += + c[1]*s[6] + (-1*c[2]*s[3]) + c[1]*s[16] + (-1*c[2]*s[13]) + c[1]*s[26] + (-1*c[2]*s[23]);
gout[4] += + (-1*c[0]*s[17]) + c[0]*s[25];
gout[5] += + c[2]*s[18] + (-1*c[0]*s[24]) + (-1*c[2]*s[2]) + c[0]*s[8];
gout[6] += + c[2]*s[1] + (-1*c[2]*s[9]);
gout[7] += + c[2]*s[0] + (-1*c[0]*s[6]) + c[2]*s[10] + (-1*c[0]*s[16]) + c[2]*s[20] + (-1*c[0]*s[26]);
gout[8] += + c[0]*s[14] + (-1*c[0]*s[22]);
gout[9] += + (-1*c[1]*s[18]) + c[1]*s[2];
gout[10] += + c[0]*s[4] + (-1*c[1]*s[1]) + (-1*c[0]*s[12]) + c[1]*s[9];
gout[11] += + c[0]*s[3] + (-1*c[1]*s[0]) + c[0]*s[13] + (-1*c[1]*s[10]) + c[0]*s[23] + (-1*c[1]*s[20]);
gout += 12;
}}
FINT cint1e_spgsp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {2, 1, 0, 0, 3, 4, 0, 3};
const FINT i_sh = shls[0];
const FINT j_sh = shls[1];
const FINT i_l = bas(ANG_OF, i_sh);
const FINT j_l = bas(ANG_OF, j_sh);if (bas(ATOM_OF, i_sh) == bas(ATOM_OF, j_sh)) {
FINT ip = CINTlen_spinor(i_sh, bas) * bas(NCTR_OF,i_sh);
FINT jp = CINTlen_spinor(j_sh, bas) * bas(NCTR_OF,j_sh);
CINTdset0(ip * jp * OF_CMPLX * ng[TENSOR], opij);
return 0; }
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_spgsp;
return CINT1e_drv(opij, &envs, 0.5, &c2s_si_1ei); }
C2F_(cint1e_spgsp)
/* <G SIGMA DOT P i|NUC |SIGMA DOT P j> */
static void CINTgout1e_cint1e_spgnucsp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double *g5 = g4  + envs->g_size * 3;
double *g6 = g5  + envs->g_size * 3;
double *g7 = g6  + envs->g_size * 3;
double *g8 = g7  + envs->g_size * 3;
double s[27];
double rirj[3], c[3];
rirj[0] = ri[0] - rj[0];
rirj[1] = ri[1] - rj[1];
rirj[2] = ri[2] - rj[2];
c[0] = 1 * rirj[0];
c[1] = 1 * rirj[1];
c[2] = 1 * rirj[2];
G1E_D_J(g1, g0, i_l+2, j_l+0);
G1E_R0I(g2, g0, i_l+1, j_l);
G1E_R0I(g3, g1, i_l+1, j_l);
G1E_D_I(g4, g0, i_l+0, j_l);
G1E_D_I(g5, g1, i_l+0, j_l);
G1E_D_I(g6, g2, i_l+0, j_l);
G1E_D_I(g7, g3, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g7[ix] * g0[iy] * g0[iz];
s[1] = g6[ix] * g1[iy] * g0[iz];
s[2] = g6[ix] * g0[iy] * g1[iz];
s[3] = g5[ix] * g2[iy] * g0[iz];
s[4] = g4[ix] * g3[iy] * g0[iz];
s[5] = g4[ix] * g2[iy] * g1[iz];
s[6] = g5[ix] * g0[iy] * g2[iz];
s[7] = g4[ix] * g1[iy] * g2[iz];
s[8] = g4[ix] * g0[iy] * g3[iz];
s[9] = g3[ix] * g4[iy] * g0[iz];
s[10] = g2[ix] * g5[iy] * g0[iz];
s[11] = g2[ix] * g4[iy] * g1[iz];
s[12] = g1[ix] * g6[iy] * g0[iz];
s[13] = g0[ix] * g7[iy] * g0[iz];
s[14] = g0[ix] * g6[iy] * g1[iz];
s[15] = g1[ix] * g4[iy] * g2[iz];
s[16] = g0[ix] * g5[iy] * g2[iz];
s[17] = g0[ix] * g4[iy] * g3[iz];
s[18] = g3[ix] * g0[iy] * g4[iz];
s[19] = g2[ix] * g1[iy] * g4[iz];
s[20] = g2[ix] * g0[iy] * g5[iz];
s[21] = g1[ix] * g2[iy] * g4[iz];
s[22] = g0[ix] * g3[iy] * g4[iz];
s[23] = g0[ix] * g2[iy] * g5[iz];
s[24] = g1[ix] * g0[iy] * g6[iz];
s[25] = g0[ix] * g1[iy] * g6[iz];
s[26] = g0[ix] * g0[iy] * g7[iz];
gout[0] += + c[1]*s[17] + (-1*c[2]*s[14]) + (-1*c[1]*s[25]) + c[2]*s[22];
gout[1] += + c[1]*s[24] + (-1*c[2]*s[21]) + (-1*c[1]*s[8]) + c[2]*s[5];
gout[2] += + c[1]*s[7] + (-1*c[2]*s[4]) + (-1*c[1]*s[15]) + c[2]*s[12];
gout[3] += + c[1]*s[6] + (-1*c[2]*s[3]) + c[1]*s[16] + (-1*c[2]*s[13]) + c[1]*s[26] + (-1*c[2]*s[23]);
gout[4] += + c[2]*s[11] + (-1*c[0]*s[17]) + (-1*c[2]*s[19]) + c[0]*s[25];
gout[5] += + c[2]*s[18] + (-1*c[0]*s[24]) + (-1*c[2]*s[2]) + c[0]*s[8];
gout[6] += + c[2]*s[1] + (-1*c[0]*s[7]) + (-1*c[2]*s[9]) + c[0]*s[15];
gout[7] += + c[2]*s[0] + (-1*c[0]*s[6]) + c[2]*s[10] + (-1*c[0]*s[16]) + c[2]*s[20] + (-1*c[0]*s[26]);
gout[8] += + c[0]*s[14] + (-1*c[1]*s[11]) + (-1*c[0]*s[22]) + c[1]*s[19];
gout[9] += + c[0]*s[21] + (-1*c[1]*s[18]) + (-1*c[0]*s[5]) + c[1]*s[2];
gout[10] += + c[0]*s[4] + (-1*c[1]*s[1]) + (-1*c[0]*s[12]) + c[1]*s[9];
gout[11] += + c[0]*s[3] + (-1*c[1]*s[0]) + c[0]*s[13] + (-1*c[1]*s[10]) + c[0]*s[23] + (-1*c[1]*s[20]);
gout += 12;
}}
FINT cint1e_spgnucsp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {2, 1, 0, 0, 3, 4, 0, 3};
const FINT i_sh = shls[0];
const FINT j_sh = shls[1];
const FINT i_l = bas(ANG_OF, i_sh);
const FINT j_l = bas(ANG_OF, j_sh);if (bas(ATOM_OF, i_sh) == bas(ATOM_OF, j_sh)) {
FINT ip = CINTlen_spinor(i_sh, bas) * bas(NCTR_OF,i_sh);
FINT jp = CINTlen_spinor(j_sh, bas) * bas(NCTR_OF,j_sh);
CINTdset0(ip * jp * OF_CMPLX * ng[TENSOR], opij);
return 0; }
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_spgnucsp;
return CINT1e_nuc_drv(opij, &envs, 0.5, &c2s_si_1ei); }
C2F_(cint1e_spgnucsp)
/* <G SIGMA DOT P i|NABLA-RINV CROSS SIGMA |j> */
static void CINTgout1e_cint1e_spgsa01(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double *g5 = g4  + envs->g_size * 3;
double *g6 = g5  + envs->g_size * 3;
double *g7 = g6  + envs->g_size * 3;
double *g8 = g7  + envs->g_size * 3;
double s[27];
double rirj[3], c[3];
rirj[0] = ri[0] - rj[0];
rirj[1] = ri[1] - rj[1];
rirj[2] = ri[2] - rj[2];
c[0] = 1 * rirj[0];
c[1] = 1 * rirj[1];
c[2] = 1 * rirj[2];
G1E_D_J(g1, g0, i_l+2, j_l+0);
G1E_D_I(g2, g0, i_l+2, j_l+0);
for (ix = 0; ix < envs->g_size * 3; ix++) {g1[ix] += g2[ix];}
G1E_R0I(g2, g0, i_l+1, j_l);
G1E_R0I(g3, g1, i_l+1, j_l);
G1E_D_I(g4, g0, i_l+0, j_l);
G1E_D_I(g5, g1, i_l+0, j_l);
G1E_D_I(g6, g2, i_l+0, j_l);
G1E_D_I(g7, g3, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g7[ix] * g0[iy] * g0[iz];
s[1] = g6[ix] * g1[iy] * g0[iz];
s[2] = g6[ix] * g0[iy] * g1[iz];
s[3] = g5[ix] * g2[iy] * g0[iz];
s[4] = g4[ix] * g3[iy] * g0[iz];
s[5] = g4[ix] * g2[iy] * g1[iz];
s[6] = g5[ix] * g0[iy] * g2[iz];
s[7] = g4[ix] * g1[iy] * g2[iz];
s[8] = g4[ix] * g0[iy] * g3[iz];
s[9] = g3[ix] * g4[iy] * g0[iz];
s[10] = g2[ix] * g5[iy] * g0[iz];
s[11] = g2[ix] * g4[iy] * g1[iz];
s[12] = g1[ix] * g6[iy] * g0[iz];
s[13] = g0[ix] * g7[iy] * g0[iz];
s[14] = g0[ix] * g6[iy] * g1[iz];
s[15] = g1[ix] * g4[iy] * g2[iz];
s[16] = g0[ix] * g5[iy] * g2[iz];
s[17] = g0[ix] * g4[iy] * g3[iz];
s[18] = g3[ix] * g0[iy] * g4[iz];
s[19] = g2[ix] * g1[iy] * g4[iz];
s[20] = g2[ix] * g0[iy] * g5[iz];
s[21] = g1[ix] * g2[iy] * g4[iz];
s[22] = g0[ix] * g3[iy] * g4[iz];
s[23] = g0[ix] * g2[iy] * g5[iz];
s[24] = g1[ix] * g0[iy] * g6[iz];
s[25] = g0[ix] * g1[iy] * g6[iz];
s[26] = g0[ix] * g0[iy] * g7[iz];
gout[0] += + (-1*c[1]*s[16]) + c[2]*s[13] + (-1*c[1]*s[26]) + c[2]*s[23];
gout[1] += + c[1]*s[7] + (-1*c[2]*s[4]);
gout[2] += + c[1]*s[8] + (-1*c[2]*s[5]);
gout[3] += + c[1]*s[17] + (-1*c[2]*s[14]) + (-1*c[1]*s[25]) + c[2]*s[22];
gout[4] += + c[1]*s[15] + (-1*c[2]*s[12]);
gout[5] += + (-1*c[1]*s[26]) + c[2]*s[23] + (-1*c[1]*s[6]) + c[2]*s[3];
gout[6] += + c[1]*s[17] + (-1*c[2]*s[14]);
gout[7] += + (-1*c[1]*s[8]) + c[2]*s[5] + c[1]*s[24] + (-1*c[2]*s[21]);
gout[8] += + c[1]*s[24] + (-1*c[2]*s[21]);
gout[9] += + c[1]*s[25] + (-1*c[2]*s[22]);
gout[10] += + (-1*c[1]*s[6]) + c[2]*s[3] + (-1*c[1]*s[16]) + c[2]*s[13];
gout[11] += + c[1]*s[7] + (-1*c[2]*s[4]) + (-1*c[1]*s[15]) + c[2]*s[12];
gout[12] += + (-1*c[2]*s[10]) + c[0]*s[16] + (-1*c[2]*s[20]) + c[0]*s[26];
gout[13] += + c[2]*s[1] + (-1*c[0]*s[7]);
gout[14] += + c[2]*s[2] + (-1*c[0]*s[8]);
gout[15] += + c[2]*s[11] + (-1*c[0]*s[17]) + (-1*c[2]*s[19]) + c[0]*s[25];
gout[16] += + c[2]*s[9] + (-1*c[0]*s[15]);
gout[17] += + (-1*c[2]*s[20]) + c[0]*s[26] + (-1*c[2]*s[0]) + c[0]*s[6];
gout[18] += + c[2]*s[11] + (-1*c[0]*s[17]);
gout[19] += + (-1*c[2]*s[2]) + c[0]*s[8] + c[2]*s[18] + (-1*c[0]*s[24]);
gout[20] += + c[2]*s[18] + (-1*c[0]*s[24]);
gout[21] += + c[2]*s[19] + (-1*c[0]*s[25]);
gout[22] += + (-1*c[2]*s[0]) + c[0]*s[6] + (-1*c[2]*s[10]) + c[0]*s[16];
gout[23] += + c[2]*s[1] + (-1*c[0]*s[7]) + (-1*c[2]*s[9]) + c[0]*s[15];
gout[24] += + (-1*c[0]*s[13]) + c[1]*s[10] + (-1*c[0]*s[23]) + c[1]*s[20];
gout[25] += + c[0]*s[4] + (-1*c[1]*s[1]);
gout[26] += + c[0]*s[5] + (-1*c[1]*s[2]);
gout[27] += + c[0]*s[14] + (-1*c[1]*s[11]) + (-1*c[0]*s[22]) + c[1]*s[19];
gout[28] += + c[0]*s[12] + (-1*c[1]*s[9]);
gout[29] += + (-1*c[0]*s[23]) + c[1]*s[20] + (-1*c[0]*s[3]) + c[1]*s[0];
gout[30] += + c[0]*s[14] + (-1*c[1]*s[11]);
gout[31] += + (-1*c[0]*s[5]) + c[1]*s[2] + c[0]*s[21] + (-1*c[1]*s[18]);
gout[32] += + c[0]*s[21] + (-1*c[1]*s[18]);
gout[33] += + c[0]*s[22] + (-1*c[1]*s[19]);
gout[34] += + (-1*c[0]*s[3]) + c[1]*s[0] + (-1*c[0]*s[13]) + c[1]*s[10];
gout[35] += + c[0]*s[4] + (-1*c[1]*s[1]) + (-1*c[0]*s[12]) + c[1]*s[9];
gout += 36;
}}
FINT cint1e_spgsa01(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {2, 1, 0, 0, 3, 4, 0, 9};
const FINT i_sh = shls[0];
const FINT j_sh = shls[1];
const FINT i_l = bas(ANG_OF, i_sh);
const FINT j_l = bas(ANG_OF, j_sh);if (bas(ATOM_OF, i_sh) == bas(ATOM_OF, j_sh)) {
FINT ip = CINTlen_spinor(i_sh, bas) * bas(NCTR_OF,i_sh);
FINT jp = CINTlen_spinor(j_sh, bas) * bas(NCTR_OF,j_sh);
CINTdset0(ip * jp * OF_CMPLX * ng[TENSOR], opij);
return 0; }
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_spgsa01;
return CINT1e_rinv_drv(opij, &envs, 0.5, &c2s_si_1e); }
C2F_(cint1e_spgsa01)
/* <k SIGMA DOT P i|R12 |SIGMA DOT P j l> : i,j \in electron 1; k,l \in electron 2
 * = (SIGMA DOT P i SIGMA DOT P j|R12 |k l) */
static void CINTgout2e_cint2e_spsp1(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double s[9];
G2E_D_J(g1, g0, i_l+1, j_l+0, k_l, l_l);
G2E_D_I(g2, g0, i_l+0, j_l, k_l, l_l);
G2E_D_I(g3, g1, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g3[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g2[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g2[ix+0]*g0[iy+0]*g1[iz+0];
s[3] = + g1[ix+0]*g2[iy+0]*g0[iz+0];
s[4] = + g0[ix+0]*g3[iy+0]*g0[iz+0];
s[5] = + g0[ix+0]*g2[iy+0]*g1[iz+0];
s[6] = + g1[ix+0]*g0[iy+0]*g2[iz+0];
s[7] = + g0[ix+0]*g1[iy+0]*g2[iz+0];
s[8] = + g0[ix+0]*g0[iy+0]*g3[iz+0];
break;
case 2:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 3:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+2] * g0[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+2] * g2[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+2] * g3[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+2] * g2[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+2] * g0[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+2] * g1[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+2] * g0[iy+2] * g3[iz+2];
break;
case 4:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 5:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+4] * g0[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+4] * g2[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+4] * g3[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+4] * g2[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+4] * g0[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+4] * g1[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+4] * g0[iy+4] * g3[iz+4];
break;
case 6:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 7:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+6] * g0[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+6] * g2[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+6] * g3[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+6] * g2[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+6] * g0[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+6] * g1[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+6] * g0[iy+6] * g3[iz+6];
break;
default:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g3[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g2[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g2[ix+i] * g0[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[3] += g1[ix+i] * g2[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g3+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[4] += g0[ix+i] * g3[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[5] += g0[ix+i] * g2[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[6] += g1[ix+i] * g0[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[7] += g0[ix+i] * g1[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[8] += g0[ix+i] * g0[iy+i] * g3[iz+i];}
break;}
if (gout_empty) {
gout[0] = + s[5] + (-1*s[7]);
gout[1] = + s[6] + (-1*s[2]);
gout[2] = + s[1] + (-1*s[3]);
gout[3] = + s[0] + s[4] + s[8];
gout += 4;
} else {
gout[0] += + s[5] + (-1*s[7]);
gout[1] += + s[6] + (-1*s[2]);
gout[2] += + s[1] + (-1*s[3]);
gout[3] += + s[0] + s[4] + s[8];
gout += 4;
}}}
void cint2e_spsp1_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 1, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_spsp1(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 1, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_spsp1;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_sf_2e2);
}
OPTIMIZER2F_(cint2e_spsp1_optimizer);
C2Fo_(cint2e_spsp1)
/* <SIGMA DOT P k SIGMA DOT P i|R12 |SIGMA DOT P j SIGMA DOT P l> : i,j \in electron 1; k,l \in electron 2
 * = (SIGMA DOT P i SIGMA DOT P j|R12 |SIGMA DOT P k SIGMA DOT P l) */
static void CINTgout2e_cint2e_spsp1spsp2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double *g5 = g4 + envs->g_size * 3;
double *g6 = g5 + envs->g_size * 3;
double *g7 = g6 + envs->g_size * 3;
double *g8 = g7 + envs->g_size * 3;
double *g9 = g8 + envs->g_size * 3;
double *g10 = g9 + envs->g_size * 3;
double *g11 = g10 + envs->g_size * 3;
double *g12 = g11 + envs->g_size * 3;
double *g13 = g12 + envs->g_size * 3;
double *g14 = g13 + envs->g_size * 3;
double *g15 = g14 + envs->g_size * 3;
double *g16 = g15 + envs->g_size * 3;
double s[81];
G2E_D_L(g1, g0, i_l+1, j_l+1, k_l+1, l_l+0);
G2E_D_K(g2, g0, i_l+1, j_l+1, k_l+0, l_l);
G2E_D_K(g3, g1, i_l+1, j_l+1, k_l+0, l_l);
G2E_D_J(g4, g0, i_l+1, j_l+0, k_l, l_l);
G2E_D_J(g5, g1, i_l+1, j_l+0, k_l, l_l);
G2E_D_J(g6, g2, i_l+1, j_l+0, k_l, l_l);
G2E_D_J(g7, g3, i_l+1, j_l+0, k_l, l_l);
G2E_D_I(g8, g0, i_l+0, j_l, k_l, l_l);
G2E_D_I(g9, g1, i_l+0, j_l, k_l, l_l);
G2E_D_I(g10, g2, i_l+0, j_l, k_l, l_l);
G2E_D_I(g11, g3, i_l+0, j_l, k_l, l_l);
G2E_D_I(g12, g4, i_l+0, j_l, k_l, l_l);
G2E_D_I(g13, g5, i_l+0, j_l, k_l, l_l);
G2E_D_I(g14, g6, i_l+0, j_l, k_l, l_l);
G2E_D_I(g15, g7, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
CINTdset0(81, s);
for (i = 0; i < envs->nrys_roots; i++) {
s[0] += g15[ix+i] * g0[iy+i] * g0[iz+i];
s[1] += g14[ix+i] * g1[iy+i] * g0[iz+i];
s[2] += g14[ix+i] * g0[iy+i] * g1[iz+i];
s[3] += g13[ix+i] * g2[iy+i] * g0[iz+i];
s[4] += g12[ix+i] * g3[iy+i] * g0[iz+i];
s[5] += g12[ix+i] * g2[iy+i] * g1[iz+i];
s[6] += g13[ix+i] * g0[iy+i] * g2[iz+i];
s[7] += g12[ix+i] * g1[iy+i] * g2[iz+i];
s[8] += g12[ix+i] * g0[iy+i] * g3[iz+i];
s[9] += g11[ix+i] * g4[iy+i] * g0[iz+i];
s[10] += g10[ix+i] * g5[iy+i] * g0[iz+i];
s[11] += g10[ix+i] * g4[iy+i] * g1[iz+i];
s[12] += g9[ix+i] * g6[iy+i] * g0[iz+i];
s[13] += g8[ix+i] * g7[iy+i] * g0[iz+i];
s[14] += g8[ix+i] * g6[iy+i] * g1[iz+i];
s[15] += g9[ix+i] * g4[iy+i] * g2[iz+i];
s[16] += g8[ix+i] * g5[iy+i] * g2[iz+i];
s[17] += g8[ix+i] * g4[iy+i] * g3[iz+i];
s[18] += g11[ix+i] * g0[iy+i] * g4[iz+i];
s[19] += g10[ix+i] * g1[iy+i] * g4[iz+i];
s[20] += g10[ix+i] * g0[iy+i] * g5[iz+i];
s[21] += g9[ix+i] * g2[iy+i] * g4[iz+i];
s[22] += g8[ix+i] * g3[iy+i] * g4[iz+i];
s[23] += g8[ix+i] * g2[iy+i] * g5[iz+i];
s[24] += g9[ix+i] * g0[iy+i] * g6[iz+i];
s[25] += g8[ix+i] * g1[iy+i] * g6[iz+i];
s[26] += g8[ix+i] * g0[iy+i] * g7[iz+i];
s[27] += g7[ix+i] * g8[iy+i] * g0[iz+i];
s[28] += g6[ix+i] * g9[iy+i] * g0[iz+i];
s[29] += g6[ix+i] * g8[iy+i] * g1[iz+i];
s[30] += g5[ix+i] * g10[iy+i] * g0[iz+i];
s[31] += g4[ix+i] * g11[iy+i] * g0[iz+i];
s[32] += g4[ix+i] * g10[iy+i] * g1[iz+i];
s[33] += g5[ix+i] * g8[iy+i] * g2[iz+i];
s[34] += g4[ix+i] * g9[iy+i] * g2[iz+i];
s[35] += g4[ix+i] * g8[iy+i] * g3[iz+i];
s[36] += g3[ix+i] * g12[iy+i] * g0[iz+i];
s[37] += g2[ix+i] * g13[iy+i] * g0[iz+i];
s[38] += g2[ix+i] * g12[iy+i] * g1[iz+i];
s[39] += g1[ix+i] * g14[iy+i] * g0[iz+i];
s[40] += g0[ix+i] * g15[iy+i] * g0[iz+i];
s[41] += g0[ix+i] * g14[iy+i] * g1[iz+i];
s[42] += g1[ix+i] * g12[iy+i] * g2[iz+i];
s[43] += g0[ix+i] * g13[iy+i] * g2[iz+i];
s[44] += g0[ix+i] * g12[iy+i] * g3[iz+i];
s[45] += g3[ix+i] * g8[iy+i] * g4[iz+i];
s[46] += g2[ix+i] * g9[iy+i] * g4[iz+i];
s[47] += g2[ix+i] * g8[iy+i] * g5[iz+i];
s[48] += g1[ix+i] * g10[iy+i] * g4[iz+i];
s[49] += g0[ix+i] * g11[iy+i] * g4[iz+i];
s[50] += g0[ix+i] * g10[iy+i] * g5[iz+i];
s[51] += g1[ix+i] * g8[iy+i] * g6[iz+i];
s[52] += g0[ix+i] * g9[iy+i] * g6[iz+i];
s[53] += g0[ix+i] * g8[iy+i] * g7[iz+i];
s[54] += g7[ix+i] * g0[iy+i] * g8[iz+i];
s[55] += g6[ix+i] * g1[iy+i] * g8[iz+i];
s[56] += g6[ix+i] * g0[iy+i] * g9[iz+i];
s[57] += g5[ix+i] * g2[iy+i] * g8[iz+i];
s[58] += g4[ix+i] * g3[iy+i] * g8[iz+i];
s[59] += g4[ix+i] * g2[iy+i] * g9[iz+i];
s[60] += g5[ix+i] * g0[iy+i] * g10[iz+i];
s[61] += g4[ix+i] * g1[iy+i] * g10[iz+i];
s[62] += g4[ix+i] * g0[iy+i] * g11[iz+i];
s[63] += g3[ix+i] * g4[iy+i] * g8[iz+i];
s[64] += g2[ix+i] * g5[iy+i] * g8[iz+i];
s[65] += g2[ix+i] * g4[iy+i] * g9[iz+i];
s[66] += g1[ix+i] * g6[iy+i] * g8[iz+i];
s[67] += g0[ix+i] * g7[iy+i] * g8[iz+i];
s[68] += g0[ix+i] * g6[iy+i] * g9[iz+i];
s[69] += g1[ix+i] * g4[iy+i] * g10[iz+i];
s[70] += g0[ix+i] * g5[iy+i] * g10[iz+i];
s[71] += g0[ix+i] * g4[iy+i] * g11[iz+i];
s[72] += g3[ix+i] * g0[iy+i] * g12[iz+i];
s[73] += g2[ix+i] * g1[iy+i] * g12[iz+i];
s[74] += g2[ix+i] * g0[iy+i] * g13[iz+i];
s[75] += g1[ix+i] * g2[iy+i] * g12[iz+i];
s[76] += g0[ix+i] * g3[iy+i] * g12[iz+i];
s[77] += g0[ix+i] * g2[iy+i] * g13[iz+i];
s[78] += g1[ix+i] * g0[iy+i] * g14[iz+i];
s[79] += g0[ix+i] * g1[iy+i] * g14[iz+i];
s[80] += g0[ix+i] * g0[iy+i] * g15[iz+i];
}
if (gout_empty) {
gout[0] = + s[50] + (-1*s[68]) + (-1*s[52]) + s[70];
gout[1] = + s[59] + (-1*s[23]) + (-1*s[61]) + s[25];
gout[2] = + s[14] + (-1*s[32]) + (-1*s[16]) + s[34];
gout[3] = + s[5] + s[41] + s[77] + (-1*s[7]) + (-1*s[43]) + (-1*s[79]);
gout[4] = + s[51] + (-1*s[69]) + (-1*s[47]) + s[65];
gout[5] = + s[60] + (-1*s[24]) + (-1*s[56]) + s[20];
gout[6] = + s[15] + (-1*s[33]) + (-1*s[11]) + s[29];
gout[7] = + s[6] + s[42] + s[78] + (-1*s[2]) + (-1*s[38]) + (-1*s[74]);
gout[8] = + s[46] + (-1*s[64]) + (-1*s[48]) + s[66];
gout[9] = + s[55] + (-1*s[19]) + (-1*s[57]) + s[21];
gout[10] = + s[10] + (-1*s[28]) + (-1*s[12]) + s[30];
gout[11] = + s[1] + s[37] + s[73] + (-1*s[3]) + (-1*s[39]) + (-1*s[75]);
gout[12] = + s[45] + (-1*s[63]) + s[49] + (-1*s[67]) + s[53] + (-1*s[71]);
gout[13] = + s[54] + (-1*s[18]) + s[58] + (-1*s[22]) + s[62] + (-1*s[26]);
gout[14] = + s[9] + (-1*s[27]) + s[13] + (-1*s[31]) + s[17] + (-1*s[35]);
gout[15] = + s[0] + s[36] + s[72] + s[4] + s[40] + s[76] + s[8] + s[44] + s[80];
gout += 16;
} else {
gout[0] += + s[50] + (-1*s[68]) + (-1*s[52]) + s[70];
gout[1] += + s[59] + (-1*s[23]) + (-1*s[61]) + s[25];
gout[2] += + s[14] + (-1*s[32]) + (-1*s[16]) + s[34];
gout[3] += + s[5] + s[41] + s[77] + (-1*s[7]) + (-1*s[43]) + (-1*s[79]);
gout[4] += + s[51] + (-1*s[69]) + (-1*s[47]) + s[65];
gout[5] += + s[60] + (-1*s[24]) + (-1*s[56]) + s[20];
gout[6] += + s[15] + (-1*s[33]) + (-1*s[11]) + s[29];
gout[7] += + s[6] + s[42] + s[78] + (-1*s[2]) + (-1*s[38]) + (-1*s[74]);
gout[8] += + s[46] + (-1*s[64]) + (-1*s[48]) + s[66];
gout[9] += + s[55] + (-1*s[19]) + (-1*s[57]) + s[21];
gout[10] += + s[10] + (-1*s[28]) + (-1*s[12]) + s[30];
gout[11] += + s[1] + s[37] + s[73] + (-1*s[3]) + (-1*s[39]) + (-1*s[75]);
gout[12] += + s[45] + (-1*s[63]) + s[49] + (-1*s[67]) + s[53] + (-1*s[71]);
gout[13] += + s[54] + (-1*s[18]) + s[58] + (-1*s[22]) + s[62] + (-1*s[26]);
gout[14] += + s[9] + (-1*s[27]) + s[13] + (-1*s[31]) + s[17] + (-1*s[35]);
gout[15] += + s[0] + s[36] + s[72] + s[4] + s[40] + s[76] + s[8] + s[44] + s[80];
gout += 16;
}}}
void cint2e_spsp1spsp2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 1, 1, 4, 4, 4, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_spsp1spsp2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 1, 1, 1, 4, 4, 4, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_spsp1spsp2;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_spsp1spsp2_optimizer);
C2Fo_(cint2e_spsp1spsp2)
/* <k SIGMA DOT R i|R12 |SIGMA DOT R j l> : i,j \in electron 1; k,l \in electron 2
 * = (SIGMA DOT R i SIGMA DOT R j|R12 |k l) */
static void CINTgout2e_cint2e_srsr1(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double s[9];
G2E_R_J(g1, g0, i_l+1, j_l+0, k_l, l_l);
G2E_R_I(g2, g0, i_l+0, j_l, k_l, l_l);
G2E_R_I(g3, g1, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g3[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g2[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g2[ix+0]*g0[iy+0]*g1[iz+0];
s[3] = + g1[ix+0]*g2[iy+0]*g0[iz+0];
s[4] = + g0[ix+0]*g3[iy+0]*g0[iz+0];
s[5] = + g0[ix+0]*g2[iy+0]*g1[iz+0];
s[6] = + g1[ix+0]*g0[iy+0]*g2[iz+0];
s[7] = + g0[ix+0]*g1[iy+0]*g2[iz+0];
s[8] = + g0[ix+0]*g0[iy+0]*g3[iz+0];
break;
case 2:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 3:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+2] * g0[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+2] * g2[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+2] * g3[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+2] * g2[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+2] * g0[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+2] * g1[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+2] * g0[iy+2] * g3[iz+2];
break;
case 4:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 5:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+4] * g0[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+4] * g2[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+4] * g3[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+4] * g2[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+4] * g0[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+4] * g1[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+4] * g0[iy+4] * g3[iz+4];
break;
case 6:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 7:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+6] * g0[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+6] * g2[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+6] * g3[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+6] * g2[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+6] * g0[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+6] * g1[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+6] * g0[iy+6] * g3[iz+6];
break;
default:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g3[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g2[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g2[ix+i] * g0[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[3] += g1[ix+i] * g2[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g3+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[4] += g0[ix+i] * g3[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[5] += g0[ix+i] * g2[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[6] += g1[ix+i] * g0[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[7] += g0[ix+i] * g1[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[8] += g0[ix+i] * g0[iy+i] * g3[iz+i];}
break;}
if (gout_empty) {
gout[0] = + s[5] + (-1*s[7]);
gout[1] = + s[6] + (-1*s[2]);
gout[2] = + s[1] + (-1*s[3]);
gout[3] = + s[0] + s[4] + s[8];
gout += 4;
} else {
gout[0] += + s[5] + (-1*s[7]);
gout[1] += + s[6] + (-1*s[2]);
gout[2] += + s[1] + (-1*s[3]);
gout[3] += + s[0] + s[4] + s[8];
gout += 4;
}}}
void cint2e_srsr1_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 1, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_srsr1(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 1, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_srsr1;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_sf_2e2);
}
OPTIMIZER2F_(cint2e_srsr1_optimizer);
C2Fo_(cint2e_srsr1)
/* <SIGMA DOT R k SIGMA DOT R i|R12 |SIGMA DOT R j SIGMA DOT R l> : i,j \in electron 1; k,l \in electron 2
 * = (SIGMA DOT R i SIGMA DOT R j|R12 |SIGMA DOT R k SIGMA DOT R l) */
static void CINTgout2e_cint2e_srsr1srsr2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double *g5 = g4 + envs->g_size * 3;
double *g6 = g5 + envs->g_size * 3;
double *g7 = g6 + envs->g_size * 3;
double *g8 = g7 + envs->g_size * 3;
double *g9 = g8 + envs->g_size * 3;
double *g10 = g9 + envs->g_size * 3;
double *g11 = g10 + envs->g_size * 3;
double *g12 = g11 + envs->g_size * 3;
double *g13 = g12 + envs->g_size * 3;
double *g14 = g13 + envs->g_size * 3;
double *g15 = g14 + envs->g_size * 3;
double *g16 = g15 + envs->g_size * 3;
double s[81];
G2E_R_L(g1, g0, i_l+1, j_l+1, k_l+1, l_l+0);
G2E_R_K(g2, g0, i_l+1, j_l+1, k_l+0, l_l);
G2E_R_K(g3, g1, i_l+1, j_l+1, k_l+0, l_l);
G2E_R_J(g4, g0, i_l+1, j_l+0, k_l, l_l);
G2E_R_J(g5, g1, i_l+1, j_l+0, k_l, l_l);
G2E_R_J(g6, g2, i_l+1, j_l+0, k_l, l_l);
G2E_R_J(g7, g3, i_l+1, j_l+0, k_l, l_l);
G2E_R_I(g8, g0, i_l+0, j_l, k_l, l_l);
G2E_R_I(g9, g1, i_l+0, j_l, k_l, l_l);
G2E_R_I(g10, g2, i_l+0, j_l, k_l, l_l);
G2E_R_I(g11, g3, i_l+0, j_l, k_l, l_l);
G2E_R_I(g12, g4, i_l+0, j_l, k_l, l_l);
G2E_R_I(g13, g5, i_l+0, j_l, k_l, l_l);
G2E_R_I(g14, g6, i_l+0, j_l, k_l, l_l);
G2E_R_I(g15, g7, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
CINTdset0(81, s);
for (i = 0; i < envs->nrys_roots; i++) {
s[0] += g15[ix+i] * g0[iy+i] * g0[iz+i];
s[1] += g14[ix+i] * g1[iy+i] * g0[iz+i];
s[2] += g14[ix+i] * g0[iy+i] * g1[iz+i];
s[3] += g13[ix+i] * g2[iy+i] * g0[iz+i];
s[4] += g12[ix+i] * g3[iy+i] * g0[iz+i];
s[5] += g12[ix+i] * g2[iy+i] * g1[iz+i];
s[6] += g13[ix+i] * g0[iy+i] * g2[iz+i];
s[7] += g12[ix+i] * g1[iy+i] * g2[iz+i];
s[8] += g12[ix+i] * g0[iy+i] * g3[iz+i];
s[9] += g11[ix+i] * g4[iy+i] * g0[iz+i];
s[10] += g10[ix+i] * g5[iy+i] * g0[iz+i];
s[11] += g10[ix+i] * g4[iy+i] * g1[iz+i];
s[12] += g9[ix+i] * g6[iy+i] * g0[iz+i];
s[13] += g8[ix+i] * g7[iy+i] * g0[iz+i];
s[14] += g8[ix+i] * g6[iy+i] * g1[iz+i];
s[15] += g9[ix+i] * g4[iy+i] * g2[iz+i];
s[16] += g8[ix+i] * g5[iy+i] * g2[iz+i];
s[17] += g8[ix+i] * g4[iy+i] * g3[iz+i];
s[18] += g11[ix+i] * g0[iy+i] * g4[iz+i];
s[19] += g10[ix+i] * g1[iy+i] * g4[iz+i];
s[20] += g10[ix+i] * g0[iy+i] * g5[iz+i];
s[21] += g9[ix+i] * g2[iy+i] * g4[iz+i];
s[22] += g8[ix+i] * g3[iy+i] * g4[iz+i];
s[23] += g8[ix+i] * g2[iy+i] * g5[iz+i];
s[24] += g9[ix+i] * g0[iy+i] * g6[iz+i];
s[25] += g8[ix+i] * g1[iy+i] * g6[iz+i];
s[26] += g8[ix+i] * g0[iy+i] * g7[iz+i];
s[27] += g7[ix+i] * g8[iy+i] * g0[iz+i];
s[28] += g6[ix+i] * g9[iy+i] * g0[iz+i];
s[29] += g6[ix+i] * g8[iy+i] * g1[iz+i];
s[30] += g5[ix+i] * g10[iy+i] * g0[iz+i];
s[31] += g4[ix+i] * g11[iy+i] * g0[iz+i];
s[32] += g4[ix+i] * g10[iy+i] * g1[iz+i];
s[33] += g5[ix+i] * g8[iy+i] * g2[iz+i];
s[34] += g4[ix+i] * g9[iy+i] * g2[iz+i];
s[35] += g4[ix+i] * g8[iy+i] * g3[iz+i];
s[36] += g3[ix+i] * g12[iy+i] * g0[iz+i];
s[37] += g2[ix+i] * g13[iy+i] * g0[iz+i];
s[38] += g2[ix+i] * g12[iy+i] * g1[iz+i];
s[39] += g1[ix+i] * g14[iy+i] * g0[iz+i];
s[40] += g0[ix+i] * g15[iy+i] * g0[iz+i];
s[41] += g0[ix+i] * g14[iy+i] * g1[iz+i];
s[42] += g1[ix+i] * g12[iy+i] * g2[iz+i];
s[43] += g0[ix+i] * g13[iy+i] * g2[iz+i];
s[44] += g0[ix+i] * g12[iy+i] * g3[iz+i];
s[45] += g3[ix+i] * g8[iy+i] * g4[iz+i];
s[46] += g2[ix+i] * g9[iy+i] * g4[iz+i];
s[47] += g2[ix+i] * g8[iy+i] * g5[iz+i];
s[48] += g1[ix+i] * g10[iy+i] * g4[iz+i];
s[49] += g0[ix+i] * g11[iy+i] * g4[iz+i];
s[50] += g0[ix+i] * g10[iy+i] * g5[iz+i];
s[51] += g1[ix+i] * g8[iy+i] * g6[iz+i];
s[52] += g0[ix+i] * g9[iy+i] * g6[iz+i];
s[53] += g0[ix+i] * g8[iy+i] * g7[iz+i];
s[54] += g7[ix+i] * g0[iy+i] * g8[iz+i];
s[55] += g6[ix+i] * g1[iy+i] * g8[iz+i];
s[56] += g6[ix+i] * g0[iy+i] * g9[iz+i];
s[57] += g5[ix+i] * g2[iy+i] * g8[iz+i];
s[58] += g4[ix+i] * g3[iy+i] * g8[iz+i];
s[59] += g4[ix+i] * g2[iy+i] * g9[iz+i];
s[60] += g5[ix+i] * g0[iy+i] * g10[iz+i];
s[61] += g4[ix+i] * g1[iy+i] * g10[iz+i];
s[62] += g4[ix+i] * g0[iy+i] * g11[iz+i];
s[63] += g3[ix+i] * g4[iy+i] * g8[iz+i];
s[64] += g2[ix+i] * g5[iy+i] * g8[iz+i];
s[65] += g2[ix+i] * g4[iy+i] * g9[iz+i];
s[66] += g1[ix+i] * g6[iy+i] * g8[iz+i];
s[67] += g0[ix+i] * g7[iy+i] * g8[iz+i];
s[68] += g0[ix+i] * g6[iy+i] * g9[iz+i];
s[69] += g1[ix+i] * g4[iy+i] * g10[iz+i];
s[70] += g0[ix+i] * g5[iy+i] * g10[iz+i];
s[71] += g0[ix+i] * g4[iy+i] * g11[iz+i];
s[72] += g3[ix+i] * g0[iy+i] * g12[iz+i];
s[73] += g2[ix+i] * g1[iy+i] * g12[iz+i];
s[74] += g2[ix+i] * g0[iy+i] * g13[iz+i];
s[75] += g1[ix+i] * g2[iy+i] * g12[iz+i];
s[76] += g0[ix+i] * g3[iy+i] * g12[iz+i];
s[77] += g0[ix+i] * g2[iy+i] * g13[iz+i];
s[78] += g1[ix+i] * g0[iy+i] * g14[iz+i];
s[79] += g0[ix+i] * g1[iy+i] * g14[iz+i];
s[80] += g0[ix+i] * g0[iy+i] * g15[iz+i];
}
if (gout_empty) {
gout[0] = + s[50] + (-1*s[68]) + (-1*s[52]) + s[70];
gout[1] = + s[59] + (-1*s[23]) + (-1*s[61]) + s[25];
gout[2] = + s[14] + (-1*s[32]) + (-1*s[16]) + s[34];
gout[3] = + s[5] + s[41] + s[77] + (-1*s[7]) + (-1*s[43]) + (-1*s[79]);
gout[4] = + s[51] + (-1*s[69]) + (-1*s[47]) + s[65];
gout[5] = + s[60] + (-1*s[24]) + (-1*s[56]) + s[20];
gout[6] = + s[15] + (-1*s[33]) + (-1*s[11]) + s[29];
gout[7] = + s[6] + s[42] + s[78] + (-1*s[2]) + (-1*s[38]) + (-1*s[74]);
gout[8] = + s[46] + (-1*s[64]) + (-1*s[48]) + s[66];
gout[9] = + s[55] + (-1*s[19]) + (-1*s[57]) + s[21];
gout[10] = + s[10] + (-1*s[28]) + (-1*s[12]) + s[30];
gout[11] = + s[1] + s[37] + s[73] + (-1*s[3]) + (-1*s[39]) + (-1*s[75]);
gout[12] = + s[45] + (-1*s[63]) + s[49] + (-1*s[67]) + s[53] + (-1*s[71]);
gout[13] = + s[54] + (-1*s[18]) + s[58] + (-1*s[22]) + s[62] + (-1*s[26]);
gout[14] = + s[9] + (-1*s[27]) + s[13] + (-1*s[31]) + s[17] + (-1*s[35]);
gout[15] = + s[0] + s[36] + s[72] + s[4] + s[40] + s[76] + s[8] + s[44] + s[80];
gout += 16;
} else {
gout[0] += + s[50] + (-1*s[68]) + (-1*s[52]) + s[70];
gout[1] += + s[59] + (-1*s[23]) + (-1*s[61]) + s[25];
gout[2] += + s[14] + (-1*s[32]) + (-1*s[16]) + s[34];
gout[3] += + s[5] + s[41] + s[77] + (-1*s[7]) + (-1*s[43]) + (-1*s[79]);
gout[4] += + s[51] + (-1*s[69]) + (-1*s[47]) + s[65];
gout[5] += + s[60] + (-1*s[24]) + (-1*s[56]) + s[20];
gout[6] += + s[15] + (-1*s[33]) + (-1*s[11]) + s[29];
gout[7] += + s[6] + s[42] + s[78] + (-1*s[2]) + (-1*s[38]) + (-1*s[74]);
gout[8] += + s[46] + (-1*s[64]) + (-1*s[48]) + s[66];
gout[9] += + s[55] + (-1*s[19]) + (-1*s[57]) + s[21];
gout[10] += + s[10] + (-1*s[28]) + (-1*s[12]) + s[30];
gout[11] += + s[1] + s[37] + s[73] + (-1*s[3]) + (-1*s[39]) + (-1*s[75]);
gout[12] += + s[45] + (-1*s[63]) + s[49] + (-1*s[67]) + s[53] + (-1*s[71]);
gout[13] += + s[54] + (-1*s[18]) + s[58] + (-1*s[22]) + s[62] + (-1*s[26]);
gout[14] += + s[9] + (-1*s[27]) + s[13] + (-1*s[31]) + s[17] + (-1*s[35]);
gout[15] += + s[0] + s[36] + s[72] + s[4] + s[40] + s[76] + s[8] + s[44] + s[80];
gout += 16;
}}}
void cint2e_srsr1srsr2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 1, 1, 4, 4, 4, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_srsr1srsr2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 1, 1, 1, 4, 4, 4, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_srsr1srsr2;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_srsr1srsr2_optimizer);
C2Fo_(cint2e_srsr1srsr2)
/* <k RC CROSS SIGMA i|R12 |SIGMA DOT P j l> : i,j \in electron 1; k,l \in electron 2
 * = (RC CROSS SIGMA i SIGMA DOT P j|R12 |k l) */
static void CINTgout2e_cint2e_cg_sa10sp1(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double s[9];
double dri[3];
dri[0] = ri[0] - env[PTR_COMMON_ORIG+0];
dri[1] = ri[1] - env[PTR_COMMON_ORIG+1];
dri[2] = ri[2] - env[PTR_COMMON_ORIG+2];
G2E_D_J(g1, g0, i_l+1, j_l+0, k_l, l_l);
G2E_RCI(g2, g0, i_l+0, j_l, k_l, l_l);
G2E_RCI(g3, g1, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g3[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g2[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g2[ix+0]*g0[iy+0]*g1[iz+0];
s[3] = + g1[ix+0]*g2[iy+0]*g0[iz+0];
s[4] = + g0[ix+0]*g3[iy+0]*g0[iz+0];
s[5] = + g0[ix+0]*g2[iy+0]*g1[iz+0];
s[6] = + g1[ix+0]*g0[iy+0]*g2[iz+0];
s[7] = + g0[ix+0]*g1[iy+0]*g2[iz+0];
s[8] = + g0[ix+0]*g0[iy+0]*g3[iz+0];
break;
case 2:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 3:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+2] * g0[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+2] * g2[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+2] * g3[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+2] * g2[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+2] * g0[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+2] * g1[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+2] * g0[iy+2] * g3[iz+2];
break;
case 4:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 5:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+4] * g0[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+4] * g2[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+4] * g3[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+4] * g2[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+4] * g0[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+4] * g1[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+4] * g0[iy+4] * g3[iz+4];
break;
case 6:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 7:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+6] * g0[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+6] * g2[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+6] * g3[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+6] * g2[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+6] * g0[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+6] * g1[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+6] * g0[iy+6] * g3[iz+6];
break;
default:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g3[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g2[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g2[ix+i] * g0[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[3] += g1[ix+i] * g2[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g3+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[4] += g0[ix+i] * g3[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[5] += g0[ix+i] * g2[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[6] += g1[ix+i] * g0[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[7] += g0[ix+i] * g1[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[8] += g0[ix+i] * g0[iy+i] * g3[iz+i];}
break;}
if (gout_empty) {
gout[0] = + s[8] + s[4];
gout[1] = + (-1*s[3]);
gout[2] = + (-1*s[6]);
gout[3] = + s[7] + (-1*s[5]);
gout[4] = + (-1*s[1]);
gout[5] = + s[0] + s[8];
gout[6] = + (-1*s[7]);
gout[7] = + (-1*s[6]) + s[2];
gout[8] = + (-1*s[2]);
gout[9] = + (-1*s[5]);
gout[10] = + s[4] + s[0];
gout[11] = + s[3] + (-1*s[1]);
gout += 12;
} else {
gout[0] += + s[8] + s[4];
gout[1] += + (-1*s[3]);
gout[2] += + (-1*s[6]);
gout[3] += + s[7] + (-1*s[5]);
gout[4] += + (-1*s[1]);
gout[5] += + s[0] + s[8];
gout[6] += + (-1*s[7]);
gout[7] += + (-1*s[6]) + s[2];
gout[8] += + (-1*s[2]);
gout[9] += + (-1*s[5]);
gout[10] += + s[4] + s[0];
gout[11] += + s[3] + (-1*s[1]);
gout += 12;
}}}
void cint2e_cg_sa10sp1_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 1, 3};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_cg_sa10sp1(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 1, 3};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_cg_sa10sp1;
envs.common_factor *= 0.5;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1i, &c2s_sf_2e2);
}
OPTIMIZER2F_(cint2e_cg_sa10sp1_optimizer);
C2Fo_(cint2e_cg_sa10sp1)
/* <SIGMA DOT P k RC CROSS SIGMA i|R12 |SIGMA DOT P j SIGMA DOT P l> : i,j \in electron 1; k,l \in electron 2
 * = (RC CROSS SIGMA i SIGMA DOT P j|R12 |SIGMA DOT P k SIGMA DOT P l) */
static void CINTgout2e_cint2e_cg_sa10sp1spsp2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double *g5 = g4 + envs->g_size * 3;
double *g6 = g5 + envs->g_size * 3;
double *g7 = g6 + envs->g_size * 3;
double *g8 = g7 + envs->g_size * 3;
double *g9 = g8 + envs->g_size * 3;
double *g10 = g9 + envs->g_size * 3;
double *g11 = g10 + envs->g_size * 3;
double *g12 = g11 + envs->g_size * 3;
double *g13 = g12 + envs->g_size * 3;
double *g14 = g13 + envs->g_size * 3;
double *g15 = g14 + envs->g_size * 3;
double *g16 = g15 + envs->g_size * 3;
double s[81];
double dri[3];
dri[0] = ri[0] - env[PTR_COMMON_ORIG+0];
dri[1] = ri[1] - env[PTR_COMMON_ORIG+1];
dri[2] = ri[2] - env[PTR_COMMON_ORIG+2];
G2E_D_L(g1, g0, i_l+1, j_l+1, k_l+1, l_l+0);
G2E_D_K(g2, g0, i_l+1, j_l+1, k_l+0, l_l);
G2E_D_K(g3, g1, i_l+1, j_l+1, k_l+0, l_l);
G2E_D_J(g4, g0, i_l+1, j_l+0, k_l, l_l);
G2E_D_J(g5, g1, i_l+1, j_l+0, k_l, l_l);
G2E_D_J(g6, g2, i_l+1, j_l+0, k_l, l_l);
G2E_D_J(g7, g3, i_l+1, j_l+0, k_l, l_l);
G2E_RCI(g8, g0, i_l+0, j_l, k_l, l_l);
G2E_RCI(g9, g1, i_l+0, j_l, k_l, l_l);
G2E_RCI(g10, g2, i_l+0, j_l, k_l, l_l);
G2E_RCI(g11, g3, i_l+0, j_l, k_l, l_l);
G2E_RCI(g12, g4, i_l+0, j_l, k_l, l_l);
G2E_RCI(g13, g5, i_l+0, j_l, k_l, l_l);
G2E_RCI(g14, g6, i_l+0, j_l, k_l, l_l);
G2E_RCI(g15, g7, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
CINTdset0(81, s);
for (i = 0; i < envs->nrys_roots; i++) {
s[0] += g15[ix+i] * g0[iy+i] * g0[iz+i];
s[1] += g14[ix+i] * g1[iy+i] * g0[iz+i];
s[2] += g14[ix+i] * g0[iy+i] * g1[iz+i];
s[3] += g13[ix+i] * g2[iy+i] * g0[iz+i];
s[4] += g12[ix+i] * g3[iy+i] * g0[iz+i];
s[5] += g12[ix+i] * g2[iy+i] * g1[iz+i];
s[6] += g13[ix+i] * g0[iy+i] * g2[iz+i];
s[7] += g12[ix+i] * g1[iy+i] * g2[iz+i];
s[8] += g12[ix+i] * g0[iy+i] * g3[iz+i];
s[9] += g11[ix+i] * g4[iy+i] * g0[iz+i];
s[10] += g10[ix+i] * g5[iy+i] * g0[iz+i];
s[11] += g10[ix+i] * g4[iy+i] * g1[iz+i];
s[12] += g9[ix+i] * g6[iy+i] * g0[iz+i];
s[13] += g8[ix+i] * g7[iy+i] * g0[iz+i];
s[14] += g8[ix+i] * g6[iy+i] * g1[iz+i];
s[15] += g9[ix+i] * g4[iy+i] * g2[iz+i];
s[16] += g8[ix+i] * g5[iy+i] * g2[iz+i];
s[17] += g8[ix+i] * g4[iy+i] * g3[iz+i];
s[18] += g11[ix+i] * g0[iy+i] * g4[iz+i];
s[19] += g10[ix+i] * g1[iy+i] * g4[iz+i];
s[20] += g10[ix+i] * g0[iy+i] * g5[iz+i];
s[21] += g9[ix+i] * g2[iy+i] * g4[iz+i];
s[22] += g8[ix+i] * g3[iy+i] * g4[iz+i];
s[23] += g8[ix+i] * g2[iy+i] * g5[iz+i];
s[24] += g9[ix+i] * g0[iy+i] * g6[iz+i];
s[25] += g8[ix+i] * g1[iy+i] * g6[iz+i];
s[26] += g8[ix+i] * g0[iy+i] * g7[iz+i];
s[27] += g7[ix+i] * g8[iy+i] * g0[iz+i];
s[28] += g6[ix+i] * g9[iy+i] * g0[iz+i];
s[29] += g6[ix+i] * g8[iy+i] * g1[iz+i];
s[30] += g5[ix+i] * g10[iy+i] * g0[iz+i];
s[31] += g4[ix+i] * g11[iy+i] * g0[iz+i];
s[32] += g4[ix+i] * g10[iy+i] * g1[iz+i];
s[33] += g5[ix+i] * g8[iy+i] * g2[iz+i];
s[34] += g4[ix+i] * g9[iy+i] * g2[iz+i];
s[35] += g4[ix+i] * g8[iy+i] * g3[iz+i];
s[36] += g3[ix+i] * g12[iy+i] * g0[iz+i];
s[37] += g2[ix+i] * g13[iy+i] * g0[iz+i];
s[38] += g2[ix+i] * g12[iy+i] * g1[iz+i];
s[39] += g1[ix+i] * g14[iy+i] * g0[iz+i];
s[40] += g0[ix+i] * g15[iy+i] * g0[iz+i];
s[41] += g0[ix+i] * g14[iy+i] * g1[iz+i];
s[42] += g1[ix+i] * g12[iy+i] * g2[iz+i];
s[43] += g0[ix+i] * g13[iy+i] * g2[iz+i];
s[44] += g0[ix+i] * g12[iy+i] * g3[iz+i];
s[45] += g3[ix+i] * g8[iy+i] * g4[iz+i];
s[46] += g2[ix+i] * g9[iy+i] * g4[iz+i];
s[47] += g2[ix+i] * g8[iy+i] * g5[iz+i];
s[48] += g1[ix+i] * g10[iy+i] * g4[iz+i];
s[49] += g0[ix+i] * g11[iy+i] * g4[iz+i];
s[50] += g0[ix+i] * g10[iy+i] * g5[iz+i];
s[51] += g1[ix+i] * g8[iy+i] * g6[iz+i];
s[52] += g0[ix+i] * g9[iy+i] * g6[iz+i];
s[53] += g0[ix+i] * g8[iy+i] * g7[iz+i];
s[54] += g7[ix+i] * g0[iy+i] * g8[iz+i];
s[55] += g6[ix+i] * g1[iy+i] * g8[iz+i];
s[56] += g6[ix+i] * g0[iy+i] * g9[iz+i];
s[57] += g5[ix+i] * g2[iy+i] * g8[iz+i];
s[58] += g4[ix+i] * g3[iy+i] * g8[iz+i];
s[59] += g4[ix+i] * g2[iy+i] * g9[iz+i];
s[60] += g5[ix+i] * g0[iy+i] * g10[iz+i];
s[61] += g4[ix+i] * g1[iy+i] * g10[iz+i];
s[62] += g4[ix+i] * g0[iy+i] * g11[iz+i];
s[63] += g3[ix+i] * g4[iy+i] * g8[iz+i];
s[64] += g2[ix+i] * g5[iy+i] * g8[iz+i];
s[65] += g2[ix+i] * g4[iy+i] * g9[iz+i];
s[66] += g1[ix+i] * g6[iy+i] * g8[iz+i];
s[67] += g0[ix+i] * g7[iy+i] * g8[iz+i];
s[68] += g0[ix+i] * g6[iy+i] * g9[iz+i];
s[69] += g1[ix+i] * g4[iy+i] * g10[iz+i];
s[70] += g0[ix+i] * g5[iy+i] * g10[iz+i];
s[71] += g0[ix+i] * g4[iy+i] * g11[iz+i];
s[72] += g3[ix+i] * g0[iy+i] * g12[iz+i];
s[73] += g2[ix+i] * g1[iy+i] * g12[iz+i];
s[74] += g2[ix+i] * g0[iy+i] * g13[iz+i];
s[75] += g1[ix+i] * g2[iy+i] * g12[iz+i];
s[76] += g0[ix+i] * g3[iy+i] * g12[iz+i];
s[77] += g0[ix+i] * g2[iy+i] * g13[iz+i];
s[78] += g1[ix+i] * g0[iy+i] * g14[iz+i];
s[79] += g0[ix+i] * g1[iy+i] * g14[iz+i];
s[80] += g0[ix+i] * g0[iy+i] * g15[iz+i];
}
if (gout_empty) {
gout[0] = + s[77] + s[41] + (-1*s[79]) + (-1*s[43]);
gout[1] = + (-1*s[32]) + s[34];
gout[2] = + (-1*s[59]) + s[61];
gout[3] = + s[68] + (-1*s[50]) + (-1*s[70]) + s[52];
gout[4] = + s[78] + s[42] + (-1*s[74]) + (-1*s[38]);
gout[5] = + (-1*s[33]) + s[29];
gout[6] = + (-1*s[60]) + s[56];
gout[7] = + s[69] + (-1*s[51]) + (-1*s[65]) + s[47];
gout[8] = + s[73] + s[37] + (-1*s[75]) + (-1*s[39]);
gout[9] = + (-1*s[28]) + s[30];
gout[10] = + (-1*s[55]) + s[57];
gout[11] = + s[64] + (-1*s[46]) + (-1*s[66]) + s[48];
gout[12] = + s[72] + s[36] + s[76] + s[40] + s[80] + s[44];
gout[13] = + (-1*s[27]) + (-1*s[31]) + (-1*s[35]);
gout[14] = + (-1*s[54]) + (-1*s[58]) + (-1*s[62]);
gout[15] = + s[63] + (-1*s[45]) + s[67] + (-1*s[49]) + s[71] + (-1*s[53]);
gout[16] = + (-1*s[14]) + s[16];
gout[17] = + s[5] + s[77] + (-1*s[7]) + (-1*s[79]);
gout[18] = + (-1*s[68]) + s[70];
gout[19] = + (-1*s[59]) + s[23] + s[61] + (-1*s[25]);
gout[20] = + (-1*s[15]) + s[11];
gout[21] = + s[6] + s[78] + (-1*s[2]) + (-1*s[74]);
gout[22] = + (-1*s[69]) + s[65];
gout[23] = + (-1*s[60]) + s[24] + s[56] + (-1*s[20]);
gout[24] = + (-1*s[10]) + s[12];
gout[25] = + s[1] + s[73] + (-1*s[3]) + (-1*s[75]);
gout[26] = + (-1*s[64]) + s[66];
gout[27] = + (-1*s[55]) + s[19] + s[57] + (-1*s[21]);
gout[28] = + (-1*s[9]) + (-1*s[13]) + (-1*s[17]);
gout[29] = + s[0] + s[72] + s[4] + s[76] + s[8] + s[80];
gout[30] = + (-1*s[63]) + (-1*s[67]) + (-1*s[71]);
gout[31] = + (-1*s[54]) + s[18] + (-1*s[58]) + s[22] + (-1*s[62]) + s[26];
gout[32] = + (-1*s[23]) + s[25];
gout[33] = + (-1*s[50]) + s[52];
gout[34] = + s[41] + s[5] + (-1*s[43]) + (-1*s[7]);
gout[35] = + s[32] + (-1*s[14]) + (-1*s[34]) + s[16];
gout[36] = + (-1*s[24]) + s[20];
gout[37] = + (-1*s[51]) + s[47];
gout[38] = + s[42] + s[6] + (-1*s[38]) + (-1*s[2]);
gout[39] = + s[33] + (-1*s[15]) + (-1*s[29]) + s[11];
gout[40] = + (-1*s[19]) + s[21];
gout[41] = + (-1*s[46]) + s[48];
gout[42] = + s[37] + s[1] + (-1*s[39]) + (-1*s[3]);
gout[43] = + s[28] + (-1*s[10]) + (-1*s[30]) + s[12];
gout[44] = + (-1*s[18]) + (-1*s[22]) + (-1*s[26]);
gout[45] = + (-1*s[45]) + (-1*s[49]) + (-1*s[53]);
gout[46] = + s[36] + s[0] + s[40] + s[4] + s[44] + s[8];
gout[47] = + s[27] + (-1*s[9]) + s[31] + (-1*s[13]) + s[35] + (-1*s[17]);
gout += 48;
} else {
gout[0] += + s[77] + s[41] + (-1*s[79]) + (-1*s[43]);
gout[1] += + (-1*s[32]) + s[34];
gout[2] += + (-1*s[59]) + s[61];
gout[3] += + s[68] + (-1*s[50]) + (-1*s[70]) + s[52];
gout[4] += + s[78] + s[42] + (-1*s[74]) + (-1*s[38]);
gout[5] += + (-1*s[33]) + s[29];
gout[6] += + (-1*s[60]) + s[56];
gout[7] += + s[69] + (-1*s[51]) + (-1*s[65]) + s[47];
gout[8] += + s[73] + s[37] + (-1*s[75]) + (-1*s[39]);
gout[9] += + (-1*s[28]) + s[30];
gout[10] += + (-1*s[55]) + s[57];
gout[11] += + s[64] + (-1*s[46]) + (-1*s[66]) + s[48];
gout[12] += + s[72] + s[36] + s[76] + s[40] + s[80] + s[44];
gout[13] += + (-1*s[27]) + (-1*s[31]) + (-1*s[35]);
gout[14] += + (-1*s[54]) + (-1*s[58]) + (-1*s[62]);
gout[15] += + s[63] + (-1*s[45]) + s[67] + (-1*s[49]) + s[71] + (-1*s[53]);
gout[16] += + (-1*s[14]) + s[16];
gout[17] += + s[5] + s[77] + (-1*s[7]) + (-1*s[79]);
gout[18] += + (-1*s[68]) + s[70];
gout[19] += + (-1*s[59]) + s[23] + s[61] + (-1*s[25]);
gout[20] += + (-1*s[15]) + s[11];
gout[21] += + s[6] + s[78] + (-1*s[2]) + (-1*s[74]);
gout[22] += + (-1*s[69]) + s[65];
gout[23] += + (-1*s[60]) + s[24] + s[56] + (-1*s[20]);
gout[24] += + (-1*s[10]) + s[12];
gout[25] += + s[1] + s[73] + (-1*s[3]) + (-1*s[75]);
gout[26] += + (-1*s[64]) + s[66];
gout[27] += + (-1*s[55]) + s[19] + s[57] + (-1*s[21]);
gout[28] += + (-1*s[9]) + (-1*s[13]) + (-1*s[17]);
gout[29] += + s[0] + s[72] + s[4] + s[76] + s[8] + s[80];
gout[30] += + (-1*s[63]) + (-1*s[67]) + (-1*s[71]);
gout[31] += + (-1*s[54]) + s[18] + (-1*s[58]) + s[22] + (-1*s[62]) + s[26];
gout[32] += + (-1*s[23]) + s[25];
gout[33] += + (-1*s[50]) + s[52];
gout[34] += + s[41] + s[5] + (-1*s[43]) + (-1*s[7]);
gout[35] += + s[32] + (-1*s[14]) + (-1*s[34]) + s[16];
gout[36] += + (-1*s[24]) + s[20];
gout[37] += + (-1*s[51]) + s[47];
gout[38] += + s[42] + s[6] + (-1*s[38]) + (-1*s[2]);
gout[39] += + s[33] + (-1*s[15]) + (-1*s[29]) + s[11];
gout[40] += + (-1*s[19]) + s[21];
gout[41] += + (-1*s[46]) + s[48];
gout[42] += + s[37] + s[1] + (-1*s[39]) + (-1*s[3]);
gout[43] += + s[28] + (-1*s[10]) + (-1*s[30]) + s[12];
gout[44] += + (-1*s[18]) + (-1*s[22]) + (-1*s[26]);
gout[45] += + (-1*s[45]) + (-1*s[49]) + (-1*s[53]);
gout[46] += + s[36] + s[0] + s[40] + s[4] + s[44] + s[8];
gout[47] += + s[27] + (-1*s[9]) + s[31] + (-1*s[13]) + s[35] + (-1*s[17]);
gout += 48;
}}}
void cint2e_cg_sa10sp1spsp2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 1, 1, 4, 4, 4, 3};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_cg_sa10sp1spsp2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 1, 1, 1, 4, 4, 4, 3};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_cg_sa10sp1spsp2;
envs.common_factor *= 0.5;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1i, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_cg_sa10sp1spsp2_optimizer);
C2Fo_(cint2e_cg_sa10sp1spsp2)
/* <k R CROSS SIGMA i|R12 |SIGMA DOT P j l> : i,j \in electron 1; k,l \in electron 2
 * = (R CROSS SIGMA i SIGMA DOT P j|R12 |k l) */
static void CINTgout2e_cint2e_giao_sa10sp1(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double s[9];
G2E_D_J(g1, g0, i_l+1, j_l+0, k_l, l_l);
G2E_R_I(g2, g0, i_l+0, j_l, k_l, l_l);
G2E_R_I(g3, g1, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g3[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g2[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g2[ix+0]*g0[iy+0]*g1[iz+0];
s[3] = + g1[ix+0]*g2[iy+0]*g0[iz+0];
s[4] = + g0[ix+0]*g3[iy+0]*g0[iz+0];
s[5] = + g0[ix+0]*g2[iy+0]*g1[iz+0];
s[6] = + g1[ix+0]*g0[iy+0]*g2[iz+0];
s[7] = + g0[ix+0]*g1[iy+0]*g2[iz+0];
s[8] = + g0[ix+0]*g0[iy+0]*g3[iz+0];
break;
case 2:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 3:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+2] * g0[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+2] * g2[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+2] * g3[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+2] * g2[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+2] * g0[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+2] * g1[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+2] * g0[iy+2] * g3[iz+2];
break;
case 4:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 5:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+4] * g0[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+4] * g2[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+4] * g3[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+4] * g2[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+4] * g0[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+4] * g1[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+4] * g0[iy+4] * g3[iz+4];
break;
case 6:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 7:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+6] * g0[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+6] * g2[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+6] * g3[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+6] * g2[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+6] * g0[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+6] * g1[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+6] * g0[iy+6] * g3[iz+6];
break;
default:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g3[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g2[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g2[ix+i] * g0[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[3] += g1[ix+i] * g2[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g3+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[4] += g0[ix+i] * g3[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[5] += g0[ix+i] * g2[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[6] += g1[ix+i] * g0[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[7] += g0[ix+i] * g1[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[8] += g0[ix+i] * g0[iy+i] * g3[iz+i];}
break;}
if (gout_empty) {
gout[0] = + s[8] + s[4];
gout[1] = + (-1*s[3]);
gout[2] = + (-1*s[6]);
gout[3] = + s[7] + (-1*s[5]);
gout[4] = + (-1*s[1]);
gout[5] = + s[0] + s[8];
gout[6] = + (-1*s[7]);
gout[7] = + (-1*s[6]) + s[2];
gout[8] = + (-1*s[2]);
gout[9] = + (-1*s[5]);
gout[10] = + s[4] + s[0];
gout[11] = + s[3] + (-1*s[1]);
gout += 12;
} else {
gout[0] += + s[8] + s[4];
gout[1] += + (-1*s[3]);
gout[2] += + (-1*s[6]);
gout[3] += + s[7] + (-1*s[5]);
gout[4] += + (-1*s[1]);
gout[5] += + s[0] + s[8];
gout[6] += + (-1*s[7]);
gout[7] += + (-1*s[6]) + s[2];
gout[8] += + (-1*s[2]);
gout[9] += + (-1*s[5]);
gout[10] += + s[4] + s[0];
gout[11] += + s[3] + (-1*s[1]);
gout += 12;
}}}
void cint2e_giao_sa10sp1_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 1, 3};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_giao_sa10sp1(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 1, 0, 0, 2, 4, 1, 3};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_giao_sa10sp1;
envs.common_factor *= 0.5;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1i, &c2s_sf_2e2);
}
OPTIMIZER2F_(cint2e_giao_sa10sp1_optimizer);
C2Fo_(cint2e_giao_sa10sp1)
/* <SIGMA DOT P k R CROSS SIGMA i|R12 |SIGMA DOT P j SIGMA DOT P l> : i,j \in electron 1; k,l \in electron 2
 * = (R CROSS SIGMA i SIGMA DOT P j|R12 |SIGMA DOT P k SIGMA DOT P l) */
static void CINTgout2e_cint2e_giao_sa10sp1spsp2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double *g5 = g4 + envs->g_size * 3;
double *g6 = g5 + envs->g_size * 3;
double *g7 = g6 + envs->g_size * 3;
double *g8 = g7 + envs->g_size * 3;
double *g9 = g8 + envs->g_size * 3;
double *g10 = g9 + envs->g_size * 3;
double *g11 = g10 + envs->g_size * 3;
double *g12 = g11 + envs->g_size * 3;
double *g13 = g12 + envs->g_size * 3;
double *g14 = g13 + envs->g_size * 3;
double *g15 = g14 + envs->g_size * 3;
double *g16 = g15 + envs->g_size * 3;
double s[81];
G2E_D_L(g1, g0, i_l+1, j_l+1, k_l+1, l_l+0);
G2E_D_K(g2, g0, i_l+1, j_l+1, k_l+0, l_l);
G2E_D_K(g3, g1, i_l+1, j_l+1, k_l+0, l_l);
G2E_D_J(g4, g0, i_l+1, j_l+0, k_l, l_l);
G2E_D_J(g5, g1, i_l+1, j_l+0, k_l, l_l);
G2E_D_J(g6, g2, i_l+1, j_l+0, k_l, l_l);
G2E_D_J(g7, g3, i_l+1, j_l+0, k_l, l_l);
G2E_R_I(g8, g0, i_l+0, j_l, k_l, l_l);
G2E_R_I(g9, g1, i_l+0, j_l, k_l, l_l);
G2E_R_I(g10, g2, i_l+0, j_l, k_l, l_l);
G2E_R_I(g11, g3, i_l+0, j_l, k_l, l_l);
G2E_R_I(g12, g4, i_l+0, j_l, k_l, l_l);
G2E_R_I(g13, g5, i_l+0, j_l, k_l, l_l);
G2E_R_I(g14, g6, i_l+0, j_l, k_l, l_l);
G2E_R_I(g15, g7, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
CINTdset0(81, s);
for (i = 0; i < envs->nrys_roots; i++) {
s[0] += g15[ix+i] * g0[iy+i] * g0[iz+i];
s[1] += g14[ix+i] * g1[iy+i] * g0[iz+i];
s[2] += g14[ix+i] * g0[iy+i] * g1[iz+i];
s[3] += g13[ix+i] * g2[iy+i] * g0[iz+i];
s[4] += g12[ix+i] * g3[iy+i] * g0[iz+i];
s[5] += g12[ix+i] * g2[iy+i] * g1[iz+i];
s[6] += g13[ix+i] * g0[iy+i] * g2[iz+i];
s[7] += g12[ix+i] * g1[iy+i] * g2[iz+i];
s[8] += g12[ix+i] * g0[iy+i] * g3[iz+i];
s[9] += g11[ix+i] * g4[iy+i] * g0[iz+i];
s[10] += g10[ix+i] * g5[iy+i] * g0[iz+i];
s[11] += g10[ix+i] * g4[iy+i] * g1[iz+i];
s[12] += g9[ix+i] * g6[iy+i] * g0[iz+i];
s[13] += g8[ix+i] * g7[iy+i] * g0[iz+i];
s[14] += g8[ix+i] * g6[iy+i] * g1[iz+i];
s[15] += g9[ix+i] * g4[iy+i] * g2[iz+i];
s[16] += g8[ix+i] * g5[iy+i] * g2[iz+i];
s[17] += g8[ix+i] * g4[iy+i] * g3[iz+i];
s[18] += g11[ix+i] * g0[iy+i] * g4[iz+i];
s[19] += g10[ix+i] * g1[iy+i] * g4[iz+i];
s[20] += g10[ix+i] * g0[iy+i] * g5[iz+i];
s[21] += g9[ix+i] * g2[iy+i] * g4[iz+i];
s[22] += g8[ix+i] * g3[iy+i] * g4[iz+i];
s[23] += g8[ix+i] * g2[iy+i] * g5[iz+i];
s[24] += g9[ix+i] * g0[iy+i] * g6[iz+i];
s[25] += g8[ix+i] * g1[iy+i] * g6[iz+i];
s[26] += g8[ix+i] * g0[iy+i] * g7[iz+i];
s[27] += g7[ix+i] * g8[iy+i] * g0[iz+i];
s[28] += g6[ix+i] * g9[iy+i] * g0[iz+i];
s[29] += g6[ix+i] * g8[iy+i] * g1[iz+i];
s[30] += g5[ix+i] * g10[iy+i] * g0[iz+i];
s[31] += g4[ix+i] * g11[iy+i] * g0[iz+i];
s[32] += g4[ix+i] * g10[iy+i] * g1[iz+i];
s[33] += g5[ix+i] * g8[iy+i] * g2[iz+i];
s[34] += g4[ix+i] * g9[iy+i] * g2[iz+i];
s[35] += g4[ix+i] * g8[iy+i] * g3[iz+i];
s[36] += g3[ix+i] * g12[iy+i] * g0[iz+i];
s[37] += g2[ix+i] * g13[iy+i] * g0[iz+i];
s[38] += g2[ix+i] * g12[iy+i] * g1[iz+i];
s[39] += g1[ix+i] * g14[iy+i] * g0[iz+i];
s[40] += g0[ix+i] * g15[iy+i] * g0[iz+i];
s[41] += g0[ix+i] * g14[iy+i] * g1[iz+i];
s[42] += g1[ix+i] * g12[iy+i] * g2[iz+i];
s[43] += g0[ix+i] * g13[iy+i] * g2[iz+i];
s[44] += g0[ix+i] * g12[iy+i] * g3[iz+i];
s[45] += g3[ix+i] * g8[iy+i] * g4[iz+i];
s[46] += g2[ix+i] * g9[iy+i] * g4[iz+i];
s[47] += g2[ix+i] * g8[iy+i] * g5[iz+i];
s[48] += g1[ix+i] * g10[iy+i] * g4[iz+i];
s[49] += g0[ix+i] * g11[iy+i] * g4[iz+i];
s[50] += g0[ix+i] * g10[iy+i] * g5[iz+i];
s[51] += g1[ix+i] * g8[iy+i] * g6[iz+i];
s[52] += g0[ix+i] * g9[iy+i] * g6[iz+i];
s[53] += g0[ix+i] * g8[iy+i] * g7[iz+i];
s[54] += g7[ix+i] * g0[iy+i] * g8[iz+i];
s[55] += g6[ix+i] * g1[iy+i] * g8[iz+i];
s[56] += g6[ix+i] * g0[iy+i] * g9[iz+i];
s[57] += g5[ix+i] * g2[iy+i] * g8[iz+i];
s[58] += g4[ix+i] * g3[iy+i] * g8[iz+i];
s[59] += g4[ix+i] * g2[iy+i] * g9[iz+i];
s[60] += g5[ix+i] * g0[iy+i] * g10[iz+i];
s[61] += g4[ix+i] * g1[iy+i] * g10[iz+i];
s[62] += g4[ix+i] * g0[iy+i] * g11[iz+i];
s[63] += g3[ix+i] * g4[iy+i] * g8[iz+i];
s[64] += g2[ix+i] * g5[iy+i] * g8[iz+i];
s[65] += g2[ix+i] * g4[iy+i] * g9[iz+i];
s[66] += g1[ix+i] * g6[iy+i] * g8[iz+i];
s[67] += g0[ix+i] * g7[iy+i] * g8[iz+i];
s[68] += g0[ix+i] * g6[iy+i] * g9[iz+i];
s[69] += g1[ix+i] * g4[iy+i] * g10[iz+i];
s[70] += g0[ix+i] * g5[iy+i] * g10[iz+i];
s[71] += g0[ix+i] * g4[iy+i] * g11[iz+i];
s[72] += g3[ix+i] * g0[iy+i] * g12[iz+i];
s[73] += g2[ix+i] * g1[iy+i] * g12[iz+i];
s[74] += g2[ix+i] * g0[iy+i] * g13[iz+i];
s[75] += g1[ix+i] * g2[iy+i] * g12[iz+i];
s[76] += g0[ix+i] * g3[iy+i] * g12[iz+i];
s[77] += g0[ix+i] * g2[iy+i] * g13[iz+i];
s[78] += g1[ix+i] * g0[iy+i] * g14[iz+i];
s[79] += g0[ix+i] * g1[iy+i] * g14[iz+i];
s[80] += g0[ix+i] * g0[iy+i] * g15[iz+i];
}
if (gout_empty) {
gout[0] = + s[77] + s[41] + (-1*s[79]) + (-1*s[43]);
gout[1] = + (-1*s[32]) + s[34];
gout[2] = + (-1*s[59]) + s[61];
gout[3] = + s[68] + (-1*s[50]) + (-1*s[70]) + s[52];
gout[4] = + s[78] + s[42] + (-1*s[74]) + (-1*s[38]);
gout[5] = + (-1*s[33]) + s[29];
gout[6] = + (-1*s[60]) + s[56];
gout[7] = + s[69] + (-1*s[51]) + (-1*s[65]) + s[47];
gout[8] = + s[73] + s[37] + (-1*s[75]) + (-1*s[39]);
gout[9] = + (-1*s[28]) + s[30];
gout[10] = + (-1*s[55]) + s[57];
gout[11] = + s[64] + (-1*s[46]) + (-1*s[66]) + s[48];
gout[12] = + s[72] + s[36] + s[76] + s[40] + s[80] + s[44];
gout[13] = + (-1*s[27]) + (-1*s[31]) + (-1*s[35]);
gout[14] = + (-1*s[54]) + (-1*s[58]) + (-1*s[62]);
gout[15] = + s[63] + (-1*s[45]) + s[67] + (-1*s[49]) + s[71] + (-1*s[53]);
gout[16] = + (-1*s[14]) + s[16];
gout[17] = + s[5] + s[77] + (-1*s[7]) + (-1*s[79]);
gout[18] = + (-1*s[68]) + s[70];
gout[19] = + (-1*s[59]) + s[23] + s[61] + (-1*s[25]);
gout[20] = + (-1*s[15]) + s[11];
gout[21] = + s[6] + s[78] + (-1*s[2]) + (-1*s[74]);
gout[22] = + (-1*s[69]) + s[65];
gout[23] = + (-1*s[60]) + s[24] + s[56] + (-1*s[20]);
gout[24] = + (-1*s[10]) + s[12];
gout[25] = + s[1] + s[73] + (-1*s[3]) + (-1*s[75]);
gout[26] = + (-1*s[64]) + s[66];
gout[27] = + (-1*s[55]) + s[19] + s[57] + (-1*s[21]);
gout[28] = + (-1*s[9]) + (-1*s[13]) + (-1*s[17]);
gout[29] = + s[0] + s[72] + s[4] + s[76] + s[8] + s[80];
gout[30] = + (-1*s[63]) + (-1*s[67]) + (-1*s[71]);
gout[31] = + (-1*s[54]) + s[18] + (-1*s[58]) + s[22] + (-1*s[62]) + s[26];
gout[32] = + (-1*s[23]) + s[25];
gout[33] = + (-1*s[50]) + s[52];
gout[34] = + s[41] + s[5] + (-1*s[43]) + (-1*s[7]);
gout[35] = + s[32] + (-1*s[14]) + (-1*s[34]) + s[16];
gout[36] = + (-1*s[24]) + s[20];
gout[37] = + (-1*s[51]) + s[47];
gout[38] = + s[42] + s[6] + (-1*s[38]) + (-1*s[2]);
gout[39] = + s[33] + (-1*s[15]) + (-1*s[29]) + s[11];
gout[40] = + (-1*s[19]) + s[21];
gout[41] = + (-1*s[46]) + s[48];
gout[42] = + s[37] + s[1] + (-1*s[39]) + (-1*s[3]);
gout[43] = + s[28] + (-1*s[10]) + (-1*s[30]) + s[12];
gout[44] = + (-1*s[18]) + (-1*s[22]) + (-1*s[26]);
gout[45] = + (-1*s[45]) + (-1*s[49]) + (-1*s[53]);
gout[46] = + s[36] + s[0] + s[40] + s[4] + s[44] + s[8];
gout[47] = + s[27] + (-1*s[9]) + s[31] + (-1*s[13]) + s[35] + (-1*s[17]);
gout += 48;
} else {
gout[0] += + s[77] + s[41] + (-1*s[79]) + (-1*s[43]);
gout[1] += + (-1*s[32]) + s[34];
gout[2] += + (-1*s[59]) + s[61];
gout[3] += + s[68] + (-1*s[50]) + (-1*s[70]) + s[52];
gout[4] += + s[78] + s[42] + (-1*s[74]) + (-1*s[38]);
gout[5] += + (-1*s[33]) + s[29];
gout[6] += + (-1*s[60]) + s[56];
gout[7] += + s[69] + (-1*s[51]) + (-1*s[65]) + s[47];
gout[8] += + s[73] + s[37] + (-1*s[75]) + (-1*s[39]);
gout[9] += + (-1*s[28]) + s[30];
gout[10] += + (-1*s[55]) + s[57];
gout[11] += + s[64] + (-1*s[46]) + (-1*s[66]) + s[48];
gout[12] += + s[72] + s[36] + s[76] + s[40] + s[80] + s[44];
gout[13] += + (-1*s[27]) + (-1*s[31]) + (-1*s[35]);
gout[14] += + (-1*s[54]) + (-1*s[58]) + (-1*s[62]);
gout[15] += + s[63] + (-1*s[45]) + s[67] + (-1*s[49]) + s[71] + (-1*s[53]);
gout[16] += + (-1*s[14]) + s[16];
gout[17] += + s[5] + s[77] + (-1*s[7]) + (-1*s[79]);
gout[18] += + (-1*s[68]) + s[70];
gout[19] += + (-1*s[59]) + s[23] + s[61] + (-1*s[25]);
gout[20] += + (-1*s[15]) + s[11];
gout[21] += + s[6] + s[78] + (-1*s[2]) + (-1*s[74]);
gout[22] += + (-1*s[69]) + s[65];
gout[23] += + (-1*s[60]) + s[24] + s[56] + (-1*s[20]);
gout[24] += + (-1*s[10]) + s[12];
gout[25] += + s[1] + s[73] + (-1*s[3]) + (-1*s[75]);
gout[26] += + (-1*s[64]) + s[66];
gout[27] += + (-1*s[55]) + s[19] + s[57] + (-1*s[21]);
gout[28] += + (-1*s[9]) + (-1*s[13]) + (-1*s[17]);
gout[29] += + s[0] + s[72] + s[4] + s[76] + s[8] + s[80];
gout[30] += + (-1*s[63]) + (-1*s[67]) + (-1*s[71]);
gout[31] += + (-1*s[54]) + s[18] + (-1*s[58]) + s[22] + (-1*s[62]) + s[26];
gout[32] += + (-1*s[23]) + s[25];
gout[33] += + (-1*s[50]) + s[52];
gout[34] += + s[41] + s[5] + (-1*s[43]) + (-1*s[7]);
gout[35] += + s[32] + (-1*s[14]) + (-1*s[34]) + s[16];
gout[36] += + (-1*s[24]) + s[20];
gout[37] += + (-1*s[51]) + s[47];
gout[38] += + s[42] + s[6] + (-1*s[38]) + (-1*s[2]);
gout[39] += + s[33] + (-1*s[15]) + (-1*s[29]) + s[11];
gout[40] += + (-1*s[19]) + s[21];
gout[41] += + (-1*s[46]) + s[48];
gout[42] += + s[37] + s[1] + (-1*s[39]) + (-1*s[3]);
gout[43] += + s[28] + (-1*s[10]) + (-1*s[30]) + s[12];
gout[44] += + (-1*s[18]) + (-1*s[22]) + (-1*s[26]);
gout[45] += + (-1*s[45]) + (-1*s[49]) + (-1*s[53]);
gout[46] += + s[36] + s[0] + s[40] + s[4] + s[44] + s[8];
gout[47] += + s[27] + (-1*s[9]) + s[31] + (-1*s[13]) + s[35] + (-1*s[17]);
gout += 48;
}}}
void cint2e_giao_sa10sp1spsp2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 1, 1, 1, 4, 4, 4, 3};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_giao_sa10sp1spsp2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 1, 1, 1, 4, 4, 4, 3};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_giao_sa10sp1spsp2;
envs.common_factor *= 0.5;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1i, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_giao_sa10sp1spsp2_optimizer);
C2Fo_(cint2e_giao_sa10sp1spsp2)
/* <k G i|R12 |j l> : i,j \in electron 1; k,l \in electron 2
 * = (G i j|R12 |k l) */
static void CINTgout2e_cint2e_g1(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double s[3];
double rirj[3];
rirj[0] = ri[0] - rj[0];
rirj[1] = ri[1] - rj[1];
rirj[2] = ri[2] - rj[2];
double c[3];
c[0] = 1 * rirj[0];
c[1] = 1 * rirj[1];
c[2] = 1 * rirj[2];
G2E_R0I(g1, g0, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g1[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g0[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g0[ix+0]*g0[iy+0]*g1[iz+0];
break;
case 2:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
break;
case 3:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g1[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g0[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g0[ix+2] * g0[iy+2] * g1[iz+2];
break;
case 4:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
break;
case 5:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g1[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g0[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g0[ix+4] * g0[iy+4] * g1[iz+4];
break;
case 6:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
break;
case 7:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g1[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g0[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g0[ix+6] * g0[iy+6] * g1[iz+6];
break;
default:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g1[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g0[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g0[ix+i] * g0[iy+i] * g1[iz+i];}
break;}
if (gout_empty) {
gout[0] = + c[1]*s[2] + (-1*c[2]*s[1]);
gout[1] = + c[2]*s[0] + (-1*c[0]*s[2]);
gout[2] = + c[0]*s[1] + (-1*c[1]*s[0]);
gout += 3;
} else {
gout[0] += + c[1]*s[2] + (-1*c[2]*s[1]);
gout[1] += + c[2]*s[0] + (-1*c[0]*s[2]);
gout[2] += + c[0]*s[1] + (-1*c[1]*s[0]);
gout += 3;
}}}
void cint2e_g1_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 0, 0, 1, 1, 1, 3};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_g1(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 0, 0, 0, 1, 1, 1, 3};
const FINT i_sh = shls[0];
const FINT j_sh = shls[1];
const FINT k_sh = shls[2];
const FINT l_sh = shls[3];
if (bas(ATOM_OF, i_sh) == bas(ATOM_OF, j_sh)) {
FINT ip = CINTlen_spinor(i_sh, bas) * bas(NCTR_OF,i_sh);
FINT jp = CINTlen_spinor(j_sh, bas) * bas(NCTR_OF,j_sh);
FINT kp = CINTlen_spinor(k_sh, bas) * bas(NCTR_OF,k_sh);
FINT lp = CINTlen_spinor(l_sh, bas) * bas(NCTR_OF,l_sh);
CINTdset0(kp * ip * jp * lp * OF_CMPLX * ng[TENSOR], opijkl);
return 0; }
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_g1;
envs.common_factor *= 0.5;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_sf_2e1i, &c2s_sf_2e2);
}
OPTIMIZER2F_(cint2e_g1_optimizer);
C2Fo_(cint2e_g1)
/* <k G SIGMA DOT P i|R12 |SIGMA DOT P j l> : i,j \in electron 1; k,l \in electron 2
 * = (G SIGMA DOT P i SIGMA DOT P j|R12 |k l) */
static void CINTgout2e_cint2e_spgsp1(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double *g5 = g4 + envs->g_size * 3;
double *g6 = g5 + envs->g_size * 3;
double *g7 = g6 + envs->g_size * 3;
double *g8 = g7 + envs->g_size * 3;
double s[27];
double rirj[3];
rirj[0] = ri[0] - rj[0];
rirj[1] = ri[1] - rj[1];
rirj[2] = ri[2] - rj[2];
double c[3];
c[0] = 1 * rirj[0];
c[1] = 1 * rirj[1];
c[2] = 1 * rirj[2];
G2E_D_J(g1, g0, i_l+2, j_l+0, k_l, l_l);
G2E_R0I(g2, g0, i_l+1, j_l, k_l, l_l);
G2E_R0I(g3, g1, i_l+1, j_l, k_l, l_l);
G2E_D_I(g4, g0, i_l+0, j_l, k_l, l_l);
G2E_D_I(g5, g1, i_l+0, j_l, k_l, l_l);
G2E_D_I(g6, g2, i_l+0, j_l, k_l, l_l);
G2E_D_I(g7, g3, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
CINTdset0(27, s);
for (i = 0; i < envs->nrys_roots; i++) {
s[0] += g7[ix+i] * g0[iy+i] * g0[iz+i];
s[1] += g6[ix+i] * g1[iy+i] * g0[iz+i];
s[2] += g6[ix+i] * g0[iy+i] * g1[iz+i];
s[3] += g5[ix+i] * g2[iy+i] * g0[iz+i];
s[4] += g4[ix+i] * g3[iy+i] * g0[iz+i];
s[5] += g4[ix+i] * g2[iy+i] * g1[iz+i];
s[6] += g5[ix+i] * g0[iy+i] * g2[iz+i];
s[7] += g4[ix+i] * g1[iy+i] * g2[iz+i];
s[8] += g4[ix+i] * g0[iy+i] * g3[iz+i];
s[9] += g3[ix+i] * g4[iy+i] * g0[iz+i];
s[10] += g2[ix+i] * g5[iy+i] * g0[iz+i];
s[11] += g2[ix+i] * g4[iy+i] * g1[iz+i];
s[12] += g1[ix+i] * g6[iy+i] * g0[iz+i];
s[13] += g0[ix+i] * g7[iy+i] * g0[iz+i];
s[14] += g0[ix+i] * g6[iy+i] * g1[iz+i];
s[15] += g1[ix+i] * g4[iy+i] * g2[iz+i];
s[16] += g0[ix+i] * g5[iy+i] * g2[iz+i];
s[17] += g0[ix+i] * g4[iy+i] * g3[iz+i];
s[18] += g3[ix+i] * g0[iy+i] * g4[iz+i];
s[19] += g2[ix+i] * g1[iy+i] * g4[iz+i];
s[20] += g2[ix+i] * g0[iy+i] * g5[iz+i];
s[21] += g1[ix+i] * g2[iy+i] * g4[iz+i];
s[22] += g0[ix+i] * g3[iy+i] * g4[iz+i];
s[23] += g0[ix+i] * g2[iy+i] * g5[iz+i];
s[24] += g1[ix+i] * g0[iy+i] * g6[iz+i];
s[25] += g0[ix+i] * g1[iy+i] * g6[iz+i];
s[26] += g0[ix+i] * g0[iy+i] * g7[iz+i];
}
if (gout_empty) {
gout[0] = + c[1]*s[17] + (-1*c[2]*s[14]) + (-1*c[1]*s[25]) + c[2]*s[22];
gout[1] = + c[1]*s[24] + (-1*c[2]*s[21]) + (-1*c[1]*s[8]) + c[2]*s[5];
gout[2] = + c[1]*s[7] + (-1*c[2]*s[4]) + (-1*c[1]*s[15]) + c[2]*s[12];
gout[3] = + c[1]*s[6] + (-1*c[2]*s[3]) + c[1]*s[16] + (-1*c[2]*s[13]) + c[1]*s[26] + (-1*c[2]*s[23]);
gout[4] = + c[2]*s[11] + (-1*c[0]*s[17]) + (-1*c[2]*s[19]) + c[0]*s[25];
gout[5] = + c[2]*s[18] + (-1*c[0]*s[24]) + (-1*c[2]*s[2]) + c[0]*s[8];
gout[6] = + c[2]*s[1] + (-1*c[0]*s[7]) + (-1*c[2]*s[9]) + c[0]*s[15];
gout[7] = + c[2]*s[0] + (-1*c[0]*s[6]) + c[2]*s[10] + (-1*c[0]*s[16]) + c[2]*s[20] + (-1*c[0]*s[26]);
gout[8] = + c[0]*s[14] + (-1*c[1]*s[11]) + (-1*c[0]*s[22]) + c[1]*s[19];
gout[9] = + c[0]*s[21] + (-1*c[1]*s[18]) + (-1*c[0]*s[5]) + c[1]*s[2];
gout[10] = + c[0]*s[4] + (-1*c[1]*s[1]) + (-1*c[0]*s[12]) + c[1]*s[9];
gout[11] = + c[0]*s[3] + (-1*c[1]*s[0]) + c[0]*s[13] + (-1*c[1]*s[10]) + c[0]*s[23] + (-1*c[1]*s[20]);
gout += 12;
} else {
gout[0] += + c[1]*s[17] + (-1*c[2]*s[14]) + (-1*c[1]*s[25]) + c[2]*s[22];
gout[1] += + c[1]*s[24] + (-1*c[2]*s[21]) + (-1*c[1]*s[8]) + c[2]*s[5];
gout[2] += + c[1]*s[7] + (-1*c[2]*s[4]) + (-1*c[1]*s[15]) + c[2]*s[12];
gout[3] += + c[1]*s[6] + (-1*c[2]*s[3]) + c[1]*s[16] + (-1*c[2]*s[13]) + c[1]*s[26] + (-1*c[2]*s[23]);
gout[4] += + c[2]*s[11] + (-1*c[0]*s[17]) + (-1*c[2]*s[19]) + c[0]*s[25];
gout[5] += + c[2]*s[18] + (-1*c[0]*s[24]) + (-1*c[2]*s[2]) + c[0]*s[8];
gout[6] += + c[2]*s[1] + (-1*c[0]*s[7]) + (-1*c[2]*s[9]) + c[0]*s[15];
gout[7] += + c[2]*s[0] + (-1*c[0]*s[6]) + c[2]*s[10] + (-1*c[0]*s[16]) + c[2]*s[20] + (-1*c[0]*s[26]);
gout[8] += + c[0]*s[14] + (-1*c[1]*s[11]) + (-1*c[0]*s[22]) + c[1]*s[19];
gout[9] += + c[0]*s[21] + (-1*c[1]*s[18]) + (-1*c[0]*s[5]) + c[1]*s[2];
gout[10] += + c[0]*s[4] + (-1*c[1]*s[1]) + (-1*c[0]*s[12]) + c[1]*s[9];
gout[11] += + c[0]*s[3] + (-1*c[1]*s[0]) + c[0]*s[13] + (-1*c[1]*s[10]) + c[0]*s[23] + (-1*c[1]*s[20]);
gout += 12;
}}}
void cint2e_spgsp1_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {2, 1, 0, 0, 3, 4, 1, 3};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_spgsp1(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {2, 1, 0, 0, 3, 4, 1, 3};
const FINT i_sh = shls[0];
const FINT j_sh = shls[1];
const FINT k_sh = shls[2];
const FINT l_sh = shls[3];
if (bas(ATOM_OF, i_sh) == bas(ATOM_OF, j_sh)) {
FINT ip = CINTlen_spinor(i_sh, bas) * bas(NCTR_OF,i_sh);
FINT jp = CINTlen_spinor(j_sh, bas) * bas(NCTR_OF,j_sh);
FINT kp = CINTlen_spinor(k_sh, bas) * bas(NCTR_OF,k_sh);
FINT lp = CINTlen_spinor(l_sh, bas) * bas(NCTR_OF,l_sh);
CINTdset0(kp * ip * jp * lp * OF_CMPLX * ng[TENSOR], opijkl);
return 0; }
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_spgsp1;
envs.common_factor *= 0.5;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1i, &c2s_sf_2e2);
}
OPTIMIZER2F_(cint2e_spgsp1_optimizer);
C2Fo_(cint2e_spgsp1)
/* <SIGMA DOT P k G i|R12 |j SIGMA DOT P l> : i,j \in electron 1; k,l \in electron 2
 * = (G i j|R12 |SIGMA DOT P k SIGMA DOT P l) */
static void CINTgout2e_cint2e_g1spsp2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double *g5 = g4 + envs->g_size * 3;
double *g6 = g5 + envs->g_size * 3;
double *g7 = g6 + envs->g_size * 3;
double *g8 = g7 + envs->g_size * 3;
double s[27];
double rirj[3];
rirj[0] = ri[0] - rj[0];
rirj[1] = ri[1] - rj[1];
rirj[2] = ri[2] - rj[2];
double c[3];
c[0] = 1 * rirj[0];
c[1] = 1 * rirj[1];
c[2] = 1 * rirj[2];
G2E_D_L(g1, g0, i_l+1, j_l+0, k_l+1, l_l+0);
G2E_D_K(g2, g0, i_l+1, j_l+0, k_l+0, l_l);
G2E_D_K(g3, g1, i_l+1, j_l+0, k_l+0, l_l);
G2E_R0I(g4, g0, i_l+0, j_l, k_l, l_l);
G2E_R0I(g5, g1, i_l+0, j_l, k_l, l_l);
G2E_R0I(g6, g2, i_l+0, j_l, k_l, l_l);
G2E_R0I(g7, g3, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
CINTdset0(27, s);
for (i = 0; i < envs->nrys_roots; i++) {
s[0] += g7[ix+i] * g0[iy+i] * g0[iz+i];
s[1] += g6[ix+i] * g1[iy+i] * g0[iz+i];
s[2] += g6[ix+i] * g0[iy+i] * g1[iz+i];
s[3] += g5[ix+i] * g2[iy+i] * g0[iz+i];
s[4] += g4[ix+i] * g3[iy+i] * g0[iz+i];
s[5] += g4[ix+i] * g2[iy+i] * g1[iz+i];
s[6] += g5[ix+i] * g0[iy+i] * g2[iz+i];
s[7] += g4[ix+i] * g1[iy+i] * g2[iz+i];
s[8] += g4[ix+i] * g0[iy+i] * g3[iz+i];
s[9] += g3[ix+i] * g4[iy+i] * g0[iz+i];
s[10] += g2[ix+i] * g5[iy+i] * g0[iz+i];
s[11] += g2[ix+i] * g4[iy+i] * g1[iz+i];
s[12] += g1[ix+i] * g6[iy+i] * g0[iz+i];
s[13] += g0[ix+i] * g7[iy+i] * g0[iz+i];
s[14] += g0[ix+i] * g6[iy+i] * g1[iz+i];
s[15] += g1[ix+i] * g4[iy+i] * g2[iz+i];
s[16] += g0[ix+i] * g5[iy+i] * g2[iz+i];
s[17] += g0[ix+i] * g4[iy+i] * g3[iz+i];
s[18] += g3[ix+i] * g0[iy+i] * g4[iz+i];
s[19] += g2[ix+i] * g1[iy+i] * g4[iz+i];
s[20] += g2[ix+i] * g0[iy+i] * g5[iz+i];
s[21] += g1[ix+i] * g2[iy+i] * g4[iz+i];
s[22] += g0[ix+i] * g3[iy+i] * g4[iz+i];
s[23] += g0[ix+i] * g2[iy+i] * g5[iz+i];
s[24] += g1[ix+i] * g0[iy+i] * g6[iz+i];
s[25] += g0[ix+i] * g1[iy+i] * g6[iz+i];
s[26] += g0[ix+i] * g0[iy+i] * g7[iz+i];
}
if (gout_empty) {
gout[0] = + c[1]*s[23] + (-1*c[2]*s[14]) + (-1*c[1]*s[25]) + c[2]*s[16];
gout[1] = + c[1]*s[24] + (-1*c[2]*s[15]) + (-1*c[1]*s[20]) + c[2]*s[11];
gout[2] = + c[1]*s[19] + (-1*c[2]*s[10]) + (-1*c[1]*s[21]) + c[2]*s[12];
gout[3] = + c[1]*s[18] + (-1*c[2]*s[9]) + c[1]*s[22] + (-1*c[2]*s[13]) + c[1]*s[26] + (-1*c[2]*s[17]);
gout[4] = + c[2]*s[5] + (-1*c[0]*s[23]) + (-1*c[2]*s[7]) + c[0]*s[25];
gout[5] = + c[2]*s[6] + (-1*c[0]*s[24]) + (-1*c[2]*s[2]) + c[0]*s[20];
gout[6] = + c[2]*s[1] + (-1*c[0]*s[19]) + (-1*c[2]*s[3]) + c[0]*s[21];
gout[7] = + c[2]*s[0] + (-1*c[0]*s[18]) + c[2]*s[4] + (-1*c[0]*s[22]) + c[2]*s[8] + (-1*c[0]*s[26]);
gout[8] = + c[0]*s[14] + (-1*c[1]*s[5]) + (-1*c[0]*s[16]) + c[1]*s[7];
gout[9] = + c[0]*s[15] + (-1*c[1]*s[6]) + (-1*c[0]*s[11]) + c[1]*s[2];
gout[10] = + c[0]*s[10] + (-1*c[1]*s[1]) + (-1*c[0]*s[12]) + c[1]*s[3];
gout[11] = + c[0]*s[9] + (-1*c[1]*s[0]) + c[0]*s[13] + (-1*c[1]*s[4]) + c[0]*s[17] + (-1*c[1]*s[8]);
gout += 12;
} else {
gout[0] += + c[1]*s[23] + (-1*c[2]*s[14]) + (-1*c[1]*s[25]) + c[2]*s[16];
gout[1] += + c[1]*s[24] + (-1*c[2]*s[15]) + (-1*c[1]*s[20]) + c[2]*s[11];
gout[2] += + c[1]*s[19] + (-1*c[2]*s[10]) + (-1*c[1]*s[21]) + c[2]*s[12];
gout[3] += + c[1]*s[18] + (-1*c[2]*s[9]) + c[1]*s[22] + (-1*c[2]*s[13]) + c[1]*s[26] + (-1*c[2]*s[17]);
gout[4] += + c[2]*s[5] + (-1*c[0]*s[23]) + (-1*c[2]*s[7]) + c[0]*s[25];
gout[5] += + c[2]*s[6] + (-1*c[0]*s[24]) + (-1*c[2]*s[2]) + c[0]*s[20];
gout[6] += + c[2]*s[1] + (-1*c[0]*s[19]) + (-1*c[2]*s[3]) + c[0]*s[21];
gout[7] += + c[2]*s[0] + (-1*c[0]*s[18]) + c[2]*s[4] + (-1*c[0]*s[22]) + c[2]*s[8] + (-1*c[0]*s[26]);
gout[8] += + c[0]*s[14] + (-1*c[1]*s[5]) + (-1*c[0]*s[16]) + c[1]*s[7];
gout[9] += + c[0]*s[15] + (-1*c[1]*s[6]) + (-1*c[0]*s[11]) + c[1]*s[2];
gout[10] += + c[0]*s[10] + (-1*c[1]*s[1]) + (-1*c[0]*s[12]) + c[1]*s[3];
gout[11] += + c[0]*s[9] + (-1*c[1]*s[0]) + c[0]*s[13] + (-1*c[1]*s[4]) + c[0]*s[17] + (-1*c[1]*s[8]);
gout += 12;
}}}
void cint2e_g1spsp2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 1, 1, 3, 1, 4, 3};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_g1spsp2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 0, 1, 1, 3, 1, 4, 3};
const FINT i_sh = shls[0];
const FINT j_sh = shls[1];
const FINT k_sh = shls[2];
const FINT l_sh = shls[3];
if (bas(ATOM_OF, i_sh) == bas(ATOM_OF, j_sh)) {
FINT ip = CINTlen_spinor(i_sh, bas) * bas(NCTR_OF,i_sh);
FINT jp = CINTlen_spinor(j_sh, bas) * bas(NCTR_OF,j_sh);
FINT kp = CINTlen_spinor(k_sh, bas) * bas(NCTR_OF,k_sh);
FINT lp = CINTlen_spinor(l_sh, bas) * bas(NCTR_OF,l_sh);
CINTdset0(kp * ip * jp * lp * OF_CMPLX * ng[TENSOR], opijkl);
return 0; }
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_g1spsp2;
envs.common_factor *= 0.5;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_sf_2e1i, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_g1spsp2_optimizer);
C2Fo_(cint2e_g1spsp2)
/* <SIGMA DOT P k G SIGMA DOT P i|R12 |SIGMA DOT P j SIGMA DOT P l> : i,j \in electron 1; k,l \in electron 2
 * = (G SIGMA DOT P i SIGMA DOT P j|R12 |SIGMA DOT P k SIGMA DOT P l) */
static void CINTgout2e_cint2e_spgsp1spsp2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double *g5 = g4 + envs->g_size * 3;
double *g6 = g5 + envs->g_size * 3;
double *g7 = g6 + envs->g_size * 3;
double *g8 = g7 + envs->g_size * 3;
double *g9 = g8 + envs->g_size * 3;
double *g10 = g9 + envs->g_size * 3;
double *g11 = g10 + envs->g_size * 3;
double *g12 = g11 + envs->g_size * 3;
double *g13 = g12 + envs->g_size * 3;
double *g14 = g13 + envs->g_size * 3;
double *g15 = g14 + envs->g_size * 3;
double *g16 = g15 + envs->g_size * 3;
double *g17 = g16 + envs->g_size * 3;
double *g18 = g17 + envs->g_size * 3;
double *g19 = g18 + envs->g_size * 3;
double *g20 = g19 + envs->g_size * 3;
double *g21 = g20 + envs->g_size * 3;
double *g22 = g21 + envs->g_size * 3;
double *g23 = g22 + envs->g_size * 3;
double *g24 = g23 + envs->g_size * 3;
double *g25 = g24 + envs->g_size * 3;
double *g26 = g25 + envs->g_size * 3;
double *g27 = g26 + envs->g_size * 3;
double *g28 = g27 + envs->g_size * 3;
double *g29 = g28 + envs->g_size * 3;
double *g30 = g29 + envs->g_size * 3;
double *g31 = g30 + envs->g_size * 3;
double *g32 = g31 + envs->g_size * 3;
double s[243];
double rirj[3];
rirj[0] = ri[0] - rj[0];
rirj[1] = ri[1] - rj[1];
rirj[2] = ri[2] - rj[2];
double c[3];
c[0] = 1 * rirj[0];
c[1] = 1 * rirj[1];
c[2] = 1 * rirj[2];
G2E_D_L(g1, g0, i_l+2, j_l+1, k_l+1, l_l+0);
G2E_D_K(g2, g0, i_l+2, j_l+1, k_l+0, l_l);
G2E_D_K(g3, g1, i_l+2, j_l+1, k_l+0, l_l);
G2E_D_J(g4, g0, i_l+2, j_l+0, k_l, l_l);
G2E_D_J(g5, g1, i_l+2, j_l+0, k_l, l_l);
G2E_D_J(g6, g2, i_l+2, j_l+0, k_l, l_l);
G2E_D_J(g7, g3, i_l+2, j_l+0, k_l, l_l);
G2E_R0I(g8, g0, i_l+1, j_l, k_l, l_l);
G2E_R0I(g9, g1, i_l+1, j_l, k_l, l_l);
G2E_R0I(g10, g2, i_l+1, j_l, k_l, l_l);
G2E_R0I(g11, g3, i_l+1, j_l, k_l, l_l);
G2E_R0I(g12, g4, i_l+1, j_l, k_l, l_l);
G2E_R0I(g13, g5, i_l+1, j_l, k_l, l_l);
G2E_R0I(g14, g6, i_l+1, j_l, k_l, l_l);
G2E_R0I(g15, g7, i_l+1, j_l, k_l, l_l);
G2E_D_I(g16, g0, i_l+0, j_l, k_l, l_l);
G2E_D_I(g17, g1, i_l+0, j_l, k_l, l_l);
G2E_D_I(g18, g2, i_l+0, j_l, k_l, l_l);
G2E_D_I(g19, g3, i_l+0, j_l, k_l, l_l);
G2E_D_I(g20, g4, i_l+0, j_l, k_l, l_l);
G2E_D_I(g21, g5, i_l+0, j_l, k_l, l_l);
G2E_D_I(g22, g6, i_l+0, j_l, k_l, l_l);
G2E_D_I(g23, g7, i_l+0, j_l, k_l, l_l);
G2E_D_I(g24, g8, i_l+0, j_l, k_l, l_l);
G2E_D_I(g25, g9, i_l+0, j_l, k_l, l_l);
G2E_D_I(g26, g10, i_l+0, j_l, k_l, l_l);
G2E_D_I(g27, g11, i_l+0, j_l, k_l, l_l);
G2E_D_I(g28, g12, i_l+0, j_l, k_l, l_l);
G2E_D_I(g29, g13, i_l+0, j_l, k_l, l_l);
G2E_D_I(g30, g14, i_l+0, j_l, k_l, l_l);
G2E_D_I(g31, g15, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
CINTdset0(243, s);
for (i = 0; i < envs->nrys_roots; i++) {
s[0] += g31[ix+i] * g0[iy+i] * g0[iz+i];
s[1] += g30[ix+i] * g1[iy+i] * g0[iz+i];
s[2] += g30[ix+i] * g0[iy+i] * g1[iz+i];
s[3] += g29[ix+i] * g2[iy+i] * g0[iz+i];
s[4] += g28[ix+i] * g3[iy+i] * g0[iz+i];
s[5] += g28[ix+i] * g2[iy+i] * g1[iz+i];
s[6] += g29[ix+i] * g0[iy+i] * g2[iz+i];
s[7] += g28[ix+i] * g1[iy+i] * g2[iz+i];
s[8] += g28[ix+i] * g0[iy+i] * g3[iz+i];
s[9] += g27[ix+i] * g4[iy+i] * g0[iz+i];
s[10] += g26[ix+i] * g5[iy+i] * g0[iz+i];
s[11] += g26[ix+i] * g4[iy+i] * g1[iz+i];
s[12] += g25[ix+i] * g6[iy+i] * g0[iz+i];
s[13] += g24[ix+i] * g7[iy+i] * g0[iz+i];
s[14] += g24[ix+i] * g6[iy+i] * g1[iz+i];
s[15] += g25[ix+i] * g4[iy+i] * g2[iz+i];
s[16] += g24[ix+i] * g5[iy+i] * g2[iz+i];
s[17] += g24[ix+i] * g4[iy+i] * g3[iz+i];
s[18] += g27[ix+i] * g0[iy+i] * g4[iz+i];
s[19] += g26[ix+i] * g1[iy+i] * g4[iz+i];
s[20] += g26[ix+i] * g0[iy+i] * g5[iz+i];
s[21] += g25[ix+i] * g2[iy+i] * g4[iz+i];
s[22] += g24[ix+i] * g3[iy+i] * g4[iz+i];
s[23] += g24[ix+i] * g2[iy+i] * g5[iz+i];
s[24] += g25[ix+i] * g0[iy+i] * g6[iz+i];
s[25] += g24[ix+i] * g1[iy+i] * g6[iz+i];
s[26] += g24[ix+i] * g0[iy+i] * g7[iz+i];
s[27] += g23[ix+i] * g8[iy+i] * g0[iz+i];
s[28] += g22[ix+i] * g9[iy+i] * g0[iz+i];
s[29] += g22[ix+i] * g8[iy+i] * g1[iz+i];
s[30] += g21[ix+i] * g10[iy+i] * g0[iz+i];
s[31] += g20[ix+i] * g11[iy+i] * g0[iz+i];
s[32] += g20[ix+i] * g10[iy+i] * g1[iz+i];
s[33] += g21[ix+i] * g8[iy+i] * g2[iz+i];
s[34] += g20[ix+i] * g9[iy+i] * g2[iz+i];
s[35] += g20[ix+i] * g8[iy+i] * g3[iz+i];
s[36] += g19[ix+i] * g12[iy+i] * g0[iz+i];
s[37] += g18[ix+i] * g13[iy+i] * g0[iz+i];
s[38] += g18[ix+i] * g12[iy+i] * g1[iz+i];
s[39] += g17[ix+i] * g14[iy+i] * g0[iz+i];
s[40] += g16[ix+i] * g15[iy+i] * g0[iz+i];
s[41] += g16[ix+i] * g14[iy+i] * g1[iz+i];
s[42] += g17[ix+i] * g12[iy+i] * g2[iz+i];
s[43] += g16[ix+i] * g13[iy+i] * g2[iz+i];
s[44] += g16[ix+i] * g12[iy+i] * g3[iz+i];
s[45] += g19[ix+i] * g8[iy+i] * g4[iz+i];
s[46] += g18[ix+i] * g9[iy+i] * g4[iz+i];
s[47] += g18[ix+i] * g8[iy+i] * g5[iz+i];
s[48] += g17[ix+i] * g10[iy+i] * g4[iz+i];
s[49] += g16[ix+i] * g11[iy+i] * g4[iz+i];
s[50] += g16[ix+i] * g10[iy+i] * g5[iz+i];
s[51] += g17[ix+i] * g8[iy+i] * g6[iz+i];
s[52] += g16[ix+i] * g9[iy+i] * g6[iz+i];
s[53] += g16[ix+i] * g8[iy+i] * g7[iz+i];
s[54] += g23[ix+i] * g0[iy+i] * g8[iz+i];
s[55] += g22[ix+i] * g1[iy+i] * g8[iz+i];
s[56] += g22[ix+i] * g0[iy+i] * g9[iz+i];
s[57] += g21[ix+i] * g2[iy+i] * g8[iz+i];
s[58] += g20[ix+i] * g3[iy+i] * g8[iz+i];
s[59] += g20[ix+i] * g2[iy+i] * g9[iz+i];
s[60] += g21[ix+i] * g0[iy+i] * g10[iz+i];
s[61] += g20[ix+i] * g1[iy+i] * g10[iz+i];
s[62] += g20[ix+i] * g0[iy+i] * g11[iz+i];
s[63] += g19[ix+i] * g4[iy+i] * g8[iz+i];
s[64] += g18[ix+i] * g5[iy+i] * g8[iz+i];
s[65] += g18[ix+i] * g4[iy+i] * g9[iz+i];
s[66] += g17[ix+i] * g6[iy+i] * g8[iz+i];
s[67] += g16[ix+i] * g7[iy+i] * g8[iz+i];
s[68] += g16[ix+i] * g6[iy+i] * g9[iz+i];
s[69] += g17[ix+i] * g4[iy+i] * g10[iz+i];
s[70] += g16[ix+i] * g5[iy+i] * g10[iz+i];
s[71] += g16[ix+i] * g4[iy+i] * g11[iz+i];
s[72] += g19[ix+i] * g0[iy+i] * g12[iz+i];
s[73] += g18[ix+i] * g1[iy+i] * g12[iz+i];
s[74] += g18[ix+i] * g0[iy+i] * g13[iz+i];
s[75] += g17[ix+i] * g2[iy+i] * g12[iz+i];
s[76] += g16[ix+i] * g3[iy+i] * g12[iz+i];
s[77] += g16[ix+i] * g2[iy+i] * g13[iz+i];
s[78] += g17[ix+i] * g0[iy+i] * g14[iz+i];
s[79] += g16[ix+i] * g1[iy+i] * g14[iz+i];
s[80] += g16[ix+i] * g0[iy+i] * g15[iz+i];
s[81] += g15[ix+i] * g16[iy+i] * g0[iz+i];
s[82] += g14[ix+i] * g17[iy+i] * g0[iz+i];
s[83] += g14[ix+i] * g16[iy+i] * g1[iz+i];
s[84] += g13[ix+i] * g18[iy+i] * g0[iz+i];
s[85] += g12[ix+i] * g19[iy+i] * g0[iz+i];
s[86] += g12[ix+i] * g18[iy+i] * g1[iz+i];
s[87] += g13[ix+i] * g16[iy+i] * g2[iz+i];
s[88] += g12[ix+i] * g17[iy+i] * g2[iz+i];
s[89] += g12[ix+i] * g16[iy+i] * g3[iz+i];
s[90] += g11[ix+i] * g20[iy+i] * g0[iz+i];
s[91] += g10[ix+i] * g21[iy+i] * g0[iz+i];
s[92] += g10[ix+i] * g20[iy+i] * g1[iz+i];
s[93] += g9[ix+i] * g22[iy+i] * g0[iz+i];
s[94] += g8[ix+i] * g23[iy+i] * g0[iz+i];
s[95] += g8[ix+i] * g22[iy+i] * g1[iz+i];
s[96] += g9[ix+i] * g20[iy+i] * g2[iz+i];
s[97] += g8[ix+i] * g21[iy+i] * g2[iz+i];
s[98] += g8[ix+i] * g20[iy+i] * g3[iz+i];
s[99] += g11[ix+i] * g16[iy+i] * g4[iz+i];
s[100] += g10[ix+i] * g17[iy+i] * g4[iz+i];
s[101] += g10[ix+i] * g16[iy+i] * g5[iz+i];
s[102] += g9[ix+i] * g18[iy+i] * g4[iz+i];
s[103] += g8[ix+i] * g19[iy+i] * g4[iz+i];
s[104] += g8[ix+i] * g18[iy+i] * g5[iz+i];
s[105] += g9[ix+i] * g16[iy+i] * g6[iz+i];
s[106] += g8[ix+i] * g17[iy+i] * g6[iz+i];
s[107] += g8[ix+i] * g16[iy+i] * g7[iz+i];
s[108] += g7[ix+i] * g24[iy+i] * g0[iz+i];
s[109] += g6[ix+i] * g25[iy+i] * g0[iz+i];
s[110] += g6[ix+i] * g24[iy+i] * g1[iz+i];
s[111] += g5[ix+i] * g26[iy+i] * g0[iz+i];
s[112] += g4[ix+i] * g27[iy+i] * g0[iz+i];
s[113] += g4[ix+i] * g26[iy+i] * g1[iz+i];
s[114] += g5[ix+i] * g24[iy+i] * g2[iz+i];
s[115] += g4[ix+i] * g25[iy+i] * g2[iz+i];
s[116] += g4[ix+i] * g24[iy+i] * g3[iz+i];
s[117] += g3[ix+i] * g28[iy+i] * g0[iz+i];
s[118] += g2[ix+i] * g29[iy+i] * g0[iz+i];
s[119] += g2[ix+i] * g28[iy+i] * g1[iz+i];
s[120] += g1[ix+i] * g30[iy+i] * g0[iz+i];
s[121] += g0[ix+i] * g31[iy+i] * g0[iz+i];
s[122] += g0[ix+i] * g30[iy+i] * g1[iz+i];
s[123] += g1[ix+i] * g28[iy+i] * g2[iz+i];
s[124] += g0[ix+i] * g29[iy+i] * g2[iz+i];
s[125] += g0[ix+i] * g28[iy+i] * g3[iz+i];
s[126] += g3[ix+i] * g24[iy+i] * g4[iz+i];
s[127] += g2[ix+i] * g25[iy+i] * g4[iz+i];
s[128] += g2[ix+i] * g24[iy+i] * g5[iz+i];
s[129] += g1[ix+i] * g26[iy+i] * g4[iz+i];
s[130] += g0[ix+i] * g27[iy+i] * g4[iz+i];
s[131] += g0[ix+i] * g26[iy+i] * g5[iz+i];
s[132] += g1[ix+i] * g24[iy+i] * g6[iz+i];
s[133] += g0[ix+i] * g25[iy+i] * g6[iz+i];
s[134] += g0[ix+i] * g24[iy+i] * g7[iz+i];
s[135] += g7[ix+i] * g16[iy+i] * g8[iz+i];
s[136] += g6[ix+i] * g17[iy+i] * g8[iz+i];
s[137] += g6[ix+i] * g16[iy+i] * g9[iz+i];
s[138] += g5[ix+i] * g18[iy+i] * g8[iz+i];
s[139] += g4[ix+i] * g19[iy+i] * g8[iz+i];
s[140] += g4[ix+i] * g18[iy+i] * g9[iz+i];
s[141] += g5[ix+i] * g16[iy+i] * g10[iz+i];
s[142] += g4[ix+i] * g17[iy+i] * g10[iz+i];
s[143] += g4[ix+i] * g16[iy+i] * g11[iz+i];
s[144] += g3[ix+i] * g20[iy+i] * g8[iz+i];
s[145] += g2[ix+i] * g21[iy+i] * g8[iz+i];
s[146] += g2[ix+i] * g20[iy+i] * g9[iz+i];
s[147] += g1[ix+i] * g22[iy+i] * g8[iz+i];
s[148] += g0[ix+i] * g23[iy+i] * g8[iz+i];
s[149] += g0[ix+i] * g22[iy+i] * g9[iz+i];
s[150] += g1[ix+i] * g20[iy+i] * g10[iz+i];
s[151] += g0[ix+i] * g21[iy+i] * g10[iz+i];
s[152] += g0[ix+i] * g20[iy+i] * g11[iz+i];
s[153] += g3[ix+i] * g16[iy+i] * g12[iz+i];
s[154] += g2[ix+i] * g17[iy+i] * g12[iz+i];
s[155] += g2[ix+i] * g16[iy+i] * g13[iz+i];
s[156] += g1[ix+i] * g18[iy+i] * g12[iz+i];
s[157] += g0[ix+i] * g19[iy+i] * g12[iz+i];
s[158] += g0[ix+i] * g18[iy+i] * g13[iz+i];
s[159] += g1[ix+i] * g16[iy+i] * g14[iz+i];
s[160] += g0[ix+i] * g17[iy+i] * g14[iz+i];
s[161] += g0[ix+i] * g16[iy+i] * g15[iz+i];
s[162] += g15[ix+i] * g0[iy+i] * g16[iz+i];
s[163] += g14[ix+i] * g1[iy+i] * g16[iz+i];
s[164] += g14[ix+i] * g0[iy+i] * g17[iz+i];
s[165] += g13[ix+i] * g2[iy+i] * g16[iz+i];
s[166] += g12[ix+i] * g3[iy+i] * g16[iz+i];
s[167] += g12[ix+i] * g2[iy+i] * g17[iz+i];
s[168] += g13[ix+i] * g0[iy+i] * g18[iz+i];
s[169] += g12[ix+i] * g1[iy+i] * g18[iz+i];
s[170] += g12[ix+i] * g0[iy+i] * g19[iz+i];
s[171] += g11[ix+i] * g4[iy+i] * g16[iz+i];
s[172] += g10[ix+i] * g5[iy+i] * g16[iz+i];
s[173] += g10[ix+i] * g4[iy+i] * g17[iz+i];
s[174] += g9[ix+i] * g6[iy+i] * g16[iz+i];
s[175] += g8[ix+i] * g7[iy+i] * g16[iz+i];
s[176] += g8[ix+i] * g6[iy+i] * g17[iz+i];
s[177] += g9[ix+i] * g4[iy+i] * g18[iz+i];
s[178] += g8[ix+i] * g5[iy+i] * g18[iz+i];
s[179] += g8[ix+i] * g4[iy+i] * g19[iz+i];
s[180] += g11[ix+i] * g0[iy+i] * g20[iz+i];
s[181] += g10[ix+i] * g1[iy+i] * g20[iz+i];
s[182] += g10[ix+i] * g0[iy+i] * g21[iz+i];
s[183] += g9[ix+i] * g2[iy+i] * g20[iz+i];
s[184] += g8[ix+i] * g3[iy+i] * g20[iz+i];
s[185] += g8[ix+i] * g2[iy+i] * g21[iz+i];
s[186] += g9[ix+i] * g0[iy+i] * g22[iz+i];
s[187] += g8[ix+i] * g1[iy+i] * g22[iz+i];
s[188] += g8[ix+i] * g0[iy+i] * g23[iz+i];
s[189] += g7[ix+i] * g8[iy+i] * g16[iz+i];
s[190] += g6[ix+i] * g9[iy+i] * g16[iz+i];
s[191] += g6[ix+i] * g8[iy+i] * g17[iz+i];
s[192] += g5[ix+i] * g10[iy+i] * g16[iz+i];
s[193] += g4[ix+i] * g11[iy+i] * g16[iz+i];
s[194] += g4[ix+i] * g10[iy+i] * g17[iz+i];
s[195] += g5[ix+i] * g8[iy+i] * g18[iz+i];
s[196] += g4[ix+i] * g9[iy+i] * g18[iz+i];
s[197] += g4[ix+i] * g8[iy+i] * g19[iz+i];
s[198] += g3[ix+i] * g12[iy+i] * g16[iz+i];
s[199] += g2[ix+i] * g13[iy+i] * g16[iz+i];
s[200] += g2[ix+i] * g12[iy+i] * g17[iz+i];
s[201] += g1[ix+i] * g14[iy+i] * g16[iz+i];
s[202] += g0[ix+i] * g15[iy+i] * g16[iz+i];
s[203] += g0[ix+i] * g14[iy+i] * g17[iz+i];
s[204] += g1[ix+i] * g12[iy+i] * g18[iz+i];
s[205] += g0[ix+i] * g13[iy+i] * g18[iz+i];
s[206] += g0[ix+i] * g12[iy+i] * g19[iz+i];
s[207] += g3[ix+i] * g8[iy+i] * g20[iz+i];
s[208] += g2[ix+i] * g9[iy+i] * g20[iz+i];
s[209] += g2[ix+i] * g8[iy+i] * g21[iz+i];
s[210] += g1[ix+i] * g10[iy+i] * g20[iz+i];
s[211] += g0[ix+i] * g11[iy+i] * g20[iz+i];
s[212] += g0[ix+i] * g10[iy+i] * g21[iz+i];
s[213] += g1[ix+i] * g8[iy+i] * g22[iz+i];
s[214] += g0[ix+i] * g9[iy+i] * g22[iz+i];
s[215] += g0[ix+i] * g8[iy+i] * g23[iz+i];
s[216] += g7[ix+i] * g0[iy+i] * g24[iz+i];
s[217] += g6[ix+i] * g1[iy+i] * g24[iz+i];
s[218] += g6[ix+i] * g0[iy+i] * g25[iz+i];
s[219] += g5[ix+i] * g2[iy+i] * g24[iz+i];
s[220] += g4[ix+i] * g3[iy+i] * g24[iz+i];
s[221] += g4[ix+i] * g2[iy+i] * g25[iz+i];
s[222] += g5[ix+i] * g0[iy+i] * g26[iz+i];
s[223] += g4[ix+i] * g1[iy+i] * g26[iz+i];
s[224] += g4[ix+i] * g0[iy+i] * g27[iz+i];
s[225] += g3[ix+i] * g4[iy+i] * g24[iz+i];
s[226] += g2[ix+i] * g5[iy+i] * g24[iz+i];
s[227] += g2[ix+i] * g4[iy+i] * g25[iz+i];
s[228] += g1[ix+i] * g6[iy+i] * g24[iz+i];
s[229] += g0[ix+i] * g7[iy+i] * g24[iz+i];
s[230] += g0[ix+i] * g6[iy+i] * g25[iz+i];
s[231] += g1[ix+i] * g4[iy+i] * g26[iz+i];
s[232] += g0[ix+i] * g5[iy+i] * g26[iz+i];
s[233] += g0[ix+i] * g4[iy+i] * g27[iz+i];
s[234] += g3[ix+i] * g0[iy+i] * g28[iz+i];
s[235] += g2[ix+i] * g1[iy+i] * g28[iz+i];
s[236] += g2[ix+i] * g0[iy+i] * g29[iz+i];
s[237] += g1[ix+i] * g2[iy+i] * g28[iz+i];
s[238] += g0[ix+i] * g3[iy+i] * g28[iz+i];
s[239] += g0[ix+i] * g2[iy+i] * g29[iz+i];
s[240] += g1[ix+i] * g0[iy+i] * g30[iz+i];
s[241] += g0[ix+i] * g1[iy+i] * g30[iz+i];
s[242] += g0[ix+i] * g0[iy+i] * g31[iz+i];
}
if (gout_empty) {
gout[0] = + c[1]*s[158] + (-1*c[2]*s[131]) + (-1*c[1]*s[230]) + c[2]*s[203] + (-1*c[1]*s[160]) + c[2]*s[133] + c[1]*s[232] + (-1*c[2]*s[205]);
gout[1] = + c[1]*s[221] + (-1*c[2]*s[194]) + (-1*c[1]*s[77]) + c[2]*s[50] + (-1*c[1]*s[223]) + c[2]*s[196] + c[1]*s[79] + (-1*c[2]*s[52]);
gout[2] = + c[1]*s[68] + (-1*c[2]*s[41]) + (-1*c[1]*s[140]) + c[2]*s[113] + (-1*c[1]*s[70]) + c[2]*s[43] + c[1]*s[142] + (-1*c[2]*s[115]);
gout[3] = + c[1]*s[59] + (-1*c[2]*s[32]) + c[1]*s[149] + (-1*c[2]*s[122]) + c[1]*s[239] + (-1*c[2]*s[212]) + (-1*c[1]*s[61]) + c[2]*s[34] + (-1*c[1]*s[151]) + c[2]*s[124] + (-1*c[1]*s[241]) + c[2]*s[214];
gout[4] = + c[1]*s[159] + (-1*c[2]*s[132]) + (-1*c[1]*s[231]) + c[2]*s[204] + (-1*c[1]*s[155]) + c[2]*s[128] + c[1]*s[227] + (-1*c[2]*s[200]);
gout[5] = + c[1]*s[222] + (-1*c[2]*s[195]) + (-1*c[1]*s[78]) + c[2]*s[51] + (-1*c[1]*s[218]) + c[2]*s[191] + c[1]*s[74] + (-1*c[2]*s[47]);
gout[6] = + c[1]*s[69] + (-1*c[2]*s[42]) + (-1*c[1]*s[141]) + c[2]*s[114] + (-1*c[1]*s[65]) + c[2]*s[38] + c[1]*s[137] + (-1*c[2]*s[110]);
gout[7] = + c[1]*s[60] + (-1*c[2]*s[33]) + c[1]*s[150] + (-1*c[2]*s[123]) + c[1]*s[240] + (-1*c[2]*s[213]) + (-1*c[1]*s[56]) + c[2]*s[29] + (-1*c[1]*s[146]) + c[2]*s[119] + (-1*c[1]*s[236]) + c[2]*s[209];
gout[8] = + c[1]*s[154] + (-1*c[2]*s[127]) + (-1*c[1]*s[226]) + c[2]*s[199] + (-1*c[1]*s[156]) + c[2]*s[129] + c[1]*s[228] + (-1*c[2]*s[201]);
gout[9] = + c[1]*s[217] + (-1*c[2]*s[190]) + (-1*c[1]*s[73]) + c[2]*s[46] + (-1*c[1]*s[219]) + c[2]*s[192] + c[1]*s[75] + (-1*c[2]*s[48]);
gout[10] = + c[1]*s[64] + (-1*c[2]*s[37]) + (-1*c[1]*s[136]) + c[2]*s[109] + (-1*c[1]*s[66]) + c[2]*s[39] + c[1]*s[138] + (-1*c[2]*s[111]);
gout[11] = + c[1]*s[55] + (-1*c[2]*s[28]) + c[1]*s[145] + (-1*c[2]*s[118]) + c[1]*s[235] + (-1*c[2]*s[208]) + (-1*c[1]*s[57]) + c[2]*s[30] + (-1*c[1]*s[147]) + c[2]*s[120] + (-1*c[1]*s[237]) + c[2]*s[210];
gout[12] = + c[1]*s[153] + (-1*c[2]*s[126]) + (-1*c[1]*s[225]) + c[2]*s[198] + c[1]*s[157] + (-1*c[2]*s[130]) + (-1*c[1]*s[229]) + c[2]*s[202] + c[1]*s[161] + (-1*c[2]*s[134]) + (-1*c[1]*s[233]) + c[2]*s[206];
gout[13] = + c[1]*s[216] + (-1*c[2]*s[189]) + (-1*c[1]*s[72]) + c[2]*s[45] + c[1]*s[220] + (-1*c[2]*s[193]) + (-1*c[1]*s[76]) + c[2]*s[49] + c[1]*s[224] + (-1*c[2]*s[197]) + (-1*c[1]*s[80]) + c[2]*s[53];
gout[14] = + c[1]*s[63] + (-1*c[2]*s[36]) + (-1*c[1]*s[135]) + c[2]*s[108] + c[1]*s[67] + (-1*c[2]*s[40]) + (-1*c[1]*s[139]) + c[2]*s[112] + c[1]*s[71] + (-1*c[2]*s[44]) + (-1*c[1]*s[143]) + c[2]*s[116];
gout[15] = + c[1]*s[54] + (-1*c[2]*s[27]) + c[1]*s[144] + (-1*c[2]*s[117]) + c[1]*s[234] + (-1*c[2]*s[207]) + c[1]*s[58] + (-1*c[2]*s[31]) + c[1]*s[148] + (-1*c[2]*s[121]) + c[1]*s[238] + (-1*c[2]*s[211]) + c[1]*s[62] + (-1*c[2]*s[35]) + c[1]*s[152] + (-1*c[2]*s[125]) + c[1]*s[242] + (-1*c[2]*s[215]);
gout[16] = + c[2]*s[104] + (-1*c[0]*s[158]) + (-1*c[2]*s[176]) + c[0]*s[230] + (-1*c[2]*s[106]) + c[0]*s[160] + c[2]*s[178] + (-1*c[0]*s[232]);
gout[17] = + c[2]*s[167] + (-1*c[0]*s[221]) + (-1*c[2]*s[23]) + c[0]*s[77] + (-1*c[2]*s[169]) + c[0]*s[223] + c[2]*s[25] + (-1*c[0]*s[79]);
gout[18] = + c[2]*s[14] + (-1*c[0]*s[68]) + (-1*c[2]*s[86]) + c[0]*s[140] + (-1*c[2]*s[16]) + c[0]*s[70] + c[2]*s[88] + (-1*c[0]*s[142]);
gout[19] = + c[2]*s[5] + (-1*c[0]*s[59]) + c[2]*s[95] + (-1*c[0]*s[149]) + c[2]*s[185] + (-1*c[0]*s[239]) + (-1*c[2]*s[7]) + c[0]*s[61] + (-1*c[2]*s[97]) + c[0]*s[151] + (-1*c[2]*s[187]) + c[0]*s[241];
gout[20] = + c[2]*s[105] + (-1*c[0]*s[159]) + (-1*c[2]*s[177]) + c[0]*s[231] + (-1*c[2]*s[101]) + c[0]*s[155] + c[2]*s[173] + (-1*c[0]*s[227]);
gout[21] = + c[2]*s[168] + (-1*c[0]*s[222]) + (-1*c[2]*s[24]) + c[0]*s[78] + (-1*c[2]*s[164]) + c[0]*s[218] + c[2]*s[20] + (-1*c[0]*s[74]);
gout[22] = + c[2]*s[15] + (-1*c[0]*s[69]) + (-1*c[2]*s[87]) + c[0]*s[141] + (-1*c[2]*s[11]) + c[0]*s[65] + c[2]*s[83] + (-1*c[0]*s[137]);
gout[23] = + c[2]*s[6] + (-1*c[0]*s[60]) + c[2]*s[96] + (-1*c[0]*s[150]) + c[2]*s[186] + (-1*c[0]*s[240]) + (-1*c[2]*s[2]) + c[0]*s[56] + (-1*c[2]*s[92]) + c[0]*s[146] + (-1*c[2]*s[182]) + c[0]*s[236];
gout[24] = + c[2]*s[100] + (-1*c[0]*s[154]) + (-1*c[2]*s[172]) + c[0]*s[226] + (-1*c[2]*s[102]) + c[0]*s[156] + c[2]*s[174] + (-1*c[0]*s[228]);
gout[25] = + c[2]*s[163] + (-1*c[0]*s[217]) + (-1*c[2]*s[19]) + c[0]*s[73] + (-1*c[2]*s[165]) + c[0]*s[219] + c[2]*s[21] + (-1*c[0]*s[75]);
gout[26] = + c[2]*s[10] + (-1*c[0]*s[64]) + (-1*c[2]*s[82]) + c[0]*s[136] + (-1*c[2]*s[12]) + c[0]*s[66] + c[2]*s[84] + (-1*c[0]*s[138]);
gout[27] = + c[2]*s[1] + (-1*c[0]*s[55]) + c[2]*s[91] + (-1*c[0]*s[145]) + c[2]*s[181] + (-1*c[0]*s[235]) + (-1*c[2]*s[3]) + c[0]*s[57] + (-1*c[2]*s[93]) + c[0]*s[147] + (-1*c[2]*s[183]) + c[0]*s[237];
gout[28] = + c[2]*s[99] + (-1*c[0]*s[153]) + (-1*c[2]*s[171]) + c[0]*s[225] + c[2]*s[103] + (-1*c[0]*s[157]) + (-1*c[2]*s[175]) + c[0]*s[229] + c[2]*s[107] + (-1*c[0]*s[161]) + (-1*c[2]*s[179]) + c[0]*s[233];
gout[29] = + c[2]*s[162] + (-1*c[0]*s[216]) + (-1*c[2]*s[18]) + c[0]*s[72] + c[2]*s[166] + (-1*c[0]*s[220]) + (-1*c[2]*s[22]) + c[0]*s[76] + c[2]*s[170] + (-1*c[0]*s[224]) + (-1*c[2]*s[26]) + c[0]*s[80];
gout[30] = + c[2]*s[9] + (-1*c[0]*s[63]) + (-1*c[2]*s[81]) + c[0]*s[135] + c[2]*s[13] + (-1*c[0]*s[67]) + (-1*c[2]*s[85]) + c[0]*s[139] + c[2]*s[17] + (-1*c[0]*s[71]) + (-1*c[2]*s[89]) + c[0]*s[143];
gout[31] = + c[2]*s[0] + (-1*c[0]*s[54]) + c[2]*s[90] + (-1*c[0]*s[144]) + c[2]*s[180] + (-1*c[0]*s[234]) + c[2]*s[4] + (-1*c[0]*s[58]) + c[2]*s[94] + (-1*c[0]*s[148]) + c[2]*s[184] + (-1*c[0]*s[238]) + c[2]*s[8] + (-1*c[0]*s[62]) + c[2]*s[98] + (-1*c[0]*s[152]) + c[2]*s[188] + (-1*c[0]*s[242]);
gout[32] = + c[0]*s[131] + (-1*c[1]*s[104]) + (-1*c[0]*s[203]) + c[1]*s[176] + (-1*c[0]*s[133]) + c[1]*s[106] + c[0]*s[205] + (-1*c[1]*s[178]);
gout[33] = + c[0]*s[194] + (-1*c[1]*s[167]) + (-1*c[0]*s[50]) + c[1]*s[23] + (-1*c[0]*s[196]) + c[1]*s[169] + c[0]*s[52] + (-1*c[1]*s[25]);
gout[34] = + c[0]*s[41] + (-1*c[1]*s[14]) + (-1*c[0]*s[113]) + c[1]*s[86] + (-1*c[0]*s[43]) + c[1]*s[16] + c[0]*s[115] + (-1*c[1]*s[88]);
gout[35] = + c[0]*s[32] + (-1*c[1]*s[5]) + c[0]*s[122] + (-1*c[1]*s[95]) + c[0]*s[212] + (-1*c[1]*s[185]) + (-1*c[0]*s[34]) + c[1]*s[7] + (-1*c[0]*s[124]) + c[1]*s[97] + (-1*c[0]*s[214]) + c[1]*s[187];
gout[36] = + c[0]*s[132] + (-1*c[1]*s[105]) + (-1*c[0]*s[204]) + c[1]*s[177] + (-1*c[0]*s[128]) + c[1]*s[101] + c[0]*s[200] + (-1*c[1]*s[173]);
gout[37] = + c[0]*s[195] + (-1*c[1]*s[168]) + (-1*c[0]*s[51]) + c[1]*s[24] + (-1*c[0]*s[191]) + c[1]*s[164] + c[0]*s[47] + (-1*c[1]*s[20]);
gout[38] = + c[0]*s[42] + (-1*c[1]*s[15]) + (-1*c[0]*s[114]) + c[1]*s[87] + (-1*c[0]*s[38]) + c[1]*s[11] + c[0]*s[110] + (-1*c[1]*s[83]);
gout[39] = + c[0]*s[33] + (-1*c[1]*s[6]) + c[0]*s[123] + (-1*c[1]*s[96]) + c[0]*s[213] + (-1*c[1]*s[186]) + (-1*c[0]*s[29]) + c[1]*s[2] + (-1*c[0]*s[119]) + c[1]*s[92] + (-1*c[0]*s[209]) + c[1]*s[182];
gout[40] = + c[0]*s[127] + (-1*c[1]*s[100]) + (-1*c[0]*s[199]) + c[1]*s[172] + (-1*c[0]*s[129]) + c[1]*s[102] + c[0]*s[201] + (-1*c[1]*s[174]);
gout[41] = + c[0]*s[190] + (-1*c[1]*s[163]) + (-1*c[0]*s[46]) + c[1]*s[19] + (-1*c[0]*s[192]) + c[1]*s[165] + c[0]*s[48] + (-1*c[1]*s[21]);
gout[42] = + c[0]*s[37] + (-1*c[1]*s[10]) + (-1*c[0]*s[109]) + c[1]*s[82] + (-1*c[0]*s[39]) + c[1]*s[12] + c[0]*s[111] + (-1*c[1]*s[84]);
gout[43] = + c[0]*s[28] + (-1*c[1]*s[1]) + c[0]*s[118] + (-1*c[1]*s[91]) + c[0]*s[208] + (-1*c[1]*s[181]) + (-1*c[0]*s[30]) + c[1]*s[3] + (-1*c[0]*s[120]) + c[1]*s[93] + (-1*c[0]*s[210]) + c[1]*s[183];
gout[44] = + c[0]*s[126] + (-1*c[1]*s[99]) + (-1*c[0]*s[198]) + c[1]*s[171] + c[0]*s[130] + (-1*c[1]*s[103]) + (-1*c[0]*s[202]) + c[1]*s[175] + c[0]*s[134] + (-1*c[1]*s[107]) + (-1*c[0]*s[206]) + c[1]*s[179];
gout[45] = + c[0]*s[189] + (-1*c[1]*s[162]) + (-1*c[0]*s[45]) + c[1]*s[18] + c[0]*s[193] + (-1*c[1]*s[166]) + (-1*c[0]*s[49]) + c[1]*s[22] + c[0]*s[197] + (-1*c[1]*s[170]) + (-1*c[0]*s[53]) + c[1]*s[26];
gout[46] = + c[0]*s[36] + (-1*c[1]*s[9]) + (-1*c[0]*s[108]) + c[1]*s[81] + c[0]*s[40] + (-1*c[1]*s[13]) + (-1*c[0]*s[112]) + c[1]*s[85] + c[0]*s[44] + (-1*c[1]*s[17]) + (-1*c[0]*s[116]) + c[1]*s[89];
gout[47] = + c[0]*s[27] + (-1*c[1]*s[0]) + c[0]*s[117] + (-1*c[1]*s[90]) + c[0]*s[207] + (-1*c[1]*s[180]) + c[0]*s[31] + (-1*c[1]*s[4]) + c[0]*s[121] + (-1*c[1]*s[94]) + c[0]*s[211] + (-1*c[1]*s[184]) + c[0]*s[35] + (-1*c[1]*s[8]) + c[0]*s[125] + (-1*c[1]*s[98]) + c[0]*s[215] + (-1*c[1]*s[188]);
gout += 48;
} else {
gout[0] += + c[1]*s[158] + (-1*c[2]*s[131]) + (-1*c[1]*s[230]) + c[2]*s[203] + (-1*c[1]*s[160]) + c[2]*s[133] + c[1]*s[232] + (-1*c[2]*s[205]);
gout[1] += + c[1]*s[221] + (-1*c[2]*s[194]) + (-1*c[1]*s[77]) + c[2]*s[50] + (-1*c[1]*s[223]) + c[2]*s[196] + c[1]*s[79] + (-1*c[2]*s[52]);
gout[2] += + c[1]*s[68] + (-1*c[2]*s[41]) + (-1*c[1]*s[140]) + c[2]*s[113] + (-1*c[1]*s[70]) + c[2]*s[43] + c[1]*s[142] + (-1*c[2]*s[115]);
gout[3] += + c[1]*s[59] + (-1*c[2]*s[32]) + c[1]*s[149] + (-1*c[2]*s[122]) + c[1]*s[239] + (-1*c[2]*s[212]) + (-1*c[1]*s[61]) + c[2]*s[34] + (-1*c[1]*s[151]) + c[2]*s[124] + (-1*c[1]*s[241]) + c[2]*s[214];
gout[4] += + c[1]*s[159] + (-1*c[2]*s[132]) + (-1*c[1]*s[231]) + c[2]*s[204] + (-1*c[1]*s[155]) + c[2]*s[128] + c[1]*s[227] + (-1*c[2]*s[200]);
gout[5] += + c[1]*s[222] + (-1*c[2]*s[195]) + (-1*c[1]*s[78]) + c[2]*s[51] + (-1*c[1]*s[218]) + c[2]*s[191] + c[1]*s[74] + (-1*c[2]*s[47]);
gout[6] += + c[1]*s[69] + (-1*c[2]*s[42]) + (-1*c[1]*s[141]) + c[2]*s[114] + (-1*c[1]*s[65]) + c[2]*s[38] + c[1]*s[137] + (-1*c[2]*s[110]);
gout[7] += + c[1]*s[60] + (-1*c[2]*s[33]) + c[1]*s[150] + (-1*c[2]*s[123]) + c[1]*s[240] + (-1*c[2]*s[213]) + (-1*c[1]*s[56]) + c[2]*s[29] + (-1*c[1]*s[146]) + c[2]*s[119] + (-1*c[1]*s[236]) + c[2]*s[209];
gout[8] += + c[1]*s[154] + (-1*c[2]*s[127]) + (-1*c[1]*s[226]) + c[2]*s[199] + (-1*c[1]*s[156]) + c[2]*s[129] + c[1]*s[228] + (-1*c[2]*s[201]);
gout[9] += + c[1]*s[217] + (-1*c[2]*s[190]) + (-1*c[1]*s[73]) + c[2]*s[46] + (-1*c[1]*s[219]) + c[2]*s[192] + c[1]*s[75] + (-1*c[2]*s[48]);
gout[10] += + c[1]*s[64] + (-1*c[2]*s[37]) + (-1*c[1]*s[136]) + c[2]*s[109] + (-1*c[1]*s[66]) + c[2]*s[39] + c[1]*s[138] + (-1*c[2]*s[111]);
gout[11] += + c[1]*s[55] + (-1*c[2]*s[28]) + c[1]*s[145] + (-1*c[2]*s[118]) + c[1]*s[235] + (-1*c[2]*s[208]) + (-1*c[1]*s[57]) + c[2]*s[30] + (-1*c[1]*s[147]) + c[2]*s[120] + (-1*c[1]*s[237]) + c[2]*s[210];
gout[12] += + c[1]*s[153] + (-1*c[2]*s[126]) + (-1*c[1]*s[225]) + c[2]*s[198] + c[1]*s[157] + (-1*c[2]*s[130]) + (-1*c[1]*s[229]) + c[2]*s[202] + c[1]*s[161] + (-1*c[2]*s[134]) + (-1*c[1]*s[233]) + c[2]*s[206];
gout[13] += + c[1]*s[216] + (-1*c[2]*s[189]) + (-1*c[1]*s[72]) + c[2]*s[45] + c[1]*s[220] + (-1*c[2]*s[193]) + (-1*c[1]*s[76]) + c[2]*s[49] + c[1]*s[224] + (-1*c[2]*s[197]) + (-1*c[1]*s[80]) + c[2]*s[53];
gout[14] += + c[1]*s[63] + (-1*c[2]*s[36]) + (-1*c[1]*s[135]) + c[2]*s[108] + c[1]*s[67] + (-1*c[2]*s[40]) + (-1*c[1]*s[139]) + c[2]*s[112] + c[1]*s[71] + (-1*c[2]*s[44]) + (-1*c[1]*s[143]) + c[2]*s[116];
gout[15] += + c[1]*s[54] + (-1*c[2]*s[27]) + c[1]*s[144] + (-1*c[2]*s[117]) + c[1]*s[234] + (-1*c[2]*s[207]) + c[1]*s[58] + (-1*c[2]*s[31]) + c[1]*s[148] + (-1*c[2]*s[121]) + c[1]*s[238] + (-1*c[2]*s[211]) + c[1]*s[62] + (-1*c[2]*s[35]) + c[1]*s[152] + (-1*c[2]*s[125]) + c[1]*s[242] + (-1*c[2]*s[215]);
gout[16] += + c[2]*s[104] + (-1*c[0]*s[158]) + (-1*c[2]*s[176]) + c[0]*s[230] + (-1*c[2]*s[106]) + c[0]*s[160] + c[2]*s[178] + (-1*c[0]*s[232]);
gout[17] += + c[2]*s[167] + (-1*c[0]*s[221]) + (-1*c[2]*s[23]) + c[0]*s[77] + (-1*c[2]*s[169]) + c[0]*s[223] + c[2]*s[25] + (-1*c[0]*s[79]);
gout[18] += + c[2]*s[14] + (-1*c[0]*s[68]) + (-1*c[2]*s[86]) + c[0]*s[140] + (-1*c[2]*s[16]) + c[0]*s[70] + c[2]*s[88] + (-1*c[0]*s[142]);
gout[19] += + c[2]*s[5] + (-1*c[0]*s[59]) + c[2]*s[95] + (-1*c[0]*s[149]) + c[2]*s[185] + (-1*c[0]*s[239]) + (-1*c[2]*s[7]) + c[0]*s[61] + (-1*c[2]*s[97]) + c[0]*s[151] + (-1*c[2]*s[187]) + c[0]*s[241];
gout[20] += + c[2]*s[105] + (-1*c[0]*s[159]) + (-1*c[2]*s[177]) + c[0]*s[231] + (-1*c[2]*s[101]) + c[0]*s[155] + c[2]*s[173] + (-1*c[0]*s[227]);
gout[21] += + c[2]*s[168] + (-1*c[0]*s[222]) + (-1*c[2]*s[24]) + c[0]*s[78] + (-1*c[2]*s[164]) + c[0]*s[218] + c[2]*s[20] + (-1*c[0]*s[74]);
gout[22] += + c[2]*s[15] + (-1*c[0]*s[69]) + (-1*c[2]*s[87]) + c[0]*s[141] + (-1*c[2]*s[11]) + c[0]*s[65] + c[2]*s[83] + (-1*c[0]*s[137]);
gout[23] += + c[2]*s[6] + (-1*c[0]*s[60]) + c[2]*s[96] + (-1*c[0]*s[150]) + c[2]*s[186] + (-1*c[0]*s[240]) + (-1*c[2]*s[2]) + c[0]*s[56] + (-1*c[2]*s[92]) + c[0]*s[146] + (-1*c[2]*s[182]) + c[0]*s[236];
gout[24] += + c[2]*s[100] + (-1*c[0]*s[154]) + (-1*c[2]*s[172]) + c[0]*s[226] + (-1*c[2]*s[102]) + c[0]*s[156] + c[2]*s[174] + (-1*c[0]*s[228]);
gout[25] += + c[2]*s[163] + (-1*c[0]*s[217]) + (-1*c[2]*s[19]) + c[0]*s[73] + (-1*c[2]*s[165]) + c[0]*s[219] + c[2]*s[21] + (-1*c[0]*s[75]);
gout[26] += + c[2]*s[10] + (-1*c[0]*s[64]) + (-1*c[2]*s[82]) + c[0]*s[136] + (-1*c[2]*s[12]) + c[0]*s[66] + c[2]*s[84] + (-1*c[0]*s[138]);
gout[27] += + c[2]*s[1] + (-1*c[0]*s[55]) + c[2]*s[91] + (-1*c[0]*s[145]) + c[2]*s[181] + (-1*c[0]*s[235]) + (-1*c[2]*s[3]) + c[0]*s[57] + (-1*c[2]*s[93]) + c[0]*s[147] + (-1*c[2]*s[183]) + c[0]*s[237];
gout[28] += + c[2]*s[99] + (-1*c[0]*s[153]) + (-1*c[2]*s[171]) + c[0]*s[225] + c[2]*s[103] + (-1*c[0]*s[157]) + (-1*c[2]*s[175]) + c[0]*s[229] + c[2]*s[107] + (-1*c[0]*s[161]) + (-1*c[2]*s[179]) + c[0]*s[233];
gout[29] += + c[2]*s[162] + (-1*c[0]*s[216]) + (-1*c[2]*s[18]) + c[0]*s[72] + c[2]*s[166] + (-1*c[0]*s[220]) + (-1*c[2]*s[22]) + c[0]*s[76] + c[2]*s[170] + (-1*c[0]*s[224]) + (-1*c[2]*s[26]) + c[0]*s[80];
gout[30] += + c[2]*s[9] + (-1*c[0]*s[63]) + (-1*c[2]*s[81]) + c[0]*s[135] + c[2]*s[13] + (-1*c[0]*s[67]) + (-1*c[2]*s[85]) + c[0]*s[139] + c[2]*s[17] + (-1*c[0]*s[71]) + (-1*c[2]*s[89]) + c[0]*s[143];
gout[31] += + c[2]*s[0] + (-1*c[0]*s[54]) + c[2]*s[90] + (-1*c[0]*s[144]) + c[2]*s[180] + (-1*c[0]*s[234]) + c[2]*s[4] + (-1*c[0]*s[58]) + c[2]*s[94] + (-1*c[0]*s[148]) + c[2]*s[184] + (-1*c[0]*s[238]) + c[2]*s[8] + (-1*c[0]*s[62]) + c[2]*s[98] + (-1*c[0]*s[152]) + c[2]*s[188] + (-1*c[0]*s[242]);
gout[32] += + c[0]*s[131] + (-1*c[1]*s[104]) + (-1*c[0]*s[203]) + c[1]*s[176] + (-1*c[0]*s[133]) + c[1]*s[106] + c[0]*s[205] + (-1*c[1]*s[178]);
gout[33] += + c[0]*s[194] + (-1*c[1]*s[167]) + (-1*c[0]*s[50]) + c[1]*s[23] + (-1*c[0]*s[196]) + c[1]*s[169] + c[0]*s[52] + (-1*c[1]*s[25]);
gout[34] += + c[0]*s[41] + (-1*c[1]*s[14]) + (-1*c[0]*s[113]) + c[1]*s[86] + (-1*c[0]*s[43]) + c[1]*s[16] + c[0]*s[115] + (-1*c[1]*s[88]);
gout[35] += + c[0]*s[32] + (-1*c[1]*s[5]) + c[0]*s[122] + (-1*c[1]*s[95]) + c[0]*s[212] + (-1*c[1]*s[185]) + (-1*c[0]*s[34]) + c[1]*s[7] + (-1*c[0]*s[124]) + c[1]*s[97] + (-1*c[0]*s[214]) + c[1]*s[187];
gout[36] += + c[0]*s[132] + (-1*c[1]*s[105]) + (-1*c[0]*s[204]) + c[1]*s[177] + (-1*c[0]*s[128]) + c[1]*s[101] + c[0]*s[200] + (-1*c[1]*s[173]);
gout[37] += + c[0]*s[195] + (-1*c[1]*s[168]) + (-1*c[0]*s[51]) + c[1]*s[24] + (-1*c[0]*s[191]) + c[1]*s[164] + c[0]*s[47] + (-1*c[1]*s[20]);
gout[38] += + c[0]*s[42] + (-1*c[1]*s[15]) + (-1*c[0]*s[114]) + c[1]*s[87] + (-1*c[0]*s[38]) + c[1]*s[11] + c[0]*s[110] + (-1*c[1]*s[83]);
gout[39] += + c[0]*s[33] + (-1*c[1]*s[6]) + c[0]*s[123] + (-1*c[1]*s[96]) + c[0]*s[213] + (-1*c[1]*s[186]) + (-1*c[0]*s[29]) + c[1]*s[2] + (-1*c[0]*s[119]) + c[1]*s[92] + (-1*c[0]*s[209]) + c[1]*s[182];
gout[40] += + c[0]*s[127] + (-1*c[1]*s[100]) + (-1*c[0]*s[199]) + c[1]*s[172] + (-1*c[0]*s[129]) + c[1]*s[102] + c[0]*s[201] + (-1*c[1]*s[174]);
gout[41] += + c[0]*s[190] + (-1*c[1]*s[163]) + (-1*c[0]*s[46]) + c[1]*s[19] + (-1*c[0]*s[192]) + c[1]*s[165] + c[0]*s[48] + (-1*c[1]*s[21]);
gout[42] += + c[0]*s[37] + (-1*c[1]*s[10]) + (-1*c[0]*s[109]) + c[1]*s[82] + (-1*c[0]*s[39]) + c[1]*s[12] + c[0]*s[111] + (-1*c[1]*s[84]);
gout[43] += + c[0]*s[28] + (-1*c[1]*s[1]) + c[0]*s[118] + (-1*c[1]*s[91]) + c[0]*s[208] + (-1*c[1]*s[181]) + (-1*c[0]*s[30]) + c[1]*s[3] + (-1*c[0]*s[120]) + c[1]*s[93] + (-1*c[0]*s[210]) + c[1]*s[183];
gout[44] += + c[0]*s[126] + (-1*c[1]*s[99]) + (-1*c[0]*s[198]) + c[1]*s[171] + c[0]*s[130] + (-1*c[1]*s[103]) + (-1*c[0]*s[202]) + c[1]*s[175] + c[0]*s[134] + (-1*c[1]*s[107]) + (-1*c[0]*s[206]) + c[1]*s[179];
gout[45] += + c[0]*s[189] + (-1*c[1]*s[162]) + (-1*c[0]*s[45]) + c[1]*s[18] + c[0]*s[193] + (-1*c[1]*s[166]) + (-1*c[0]*s[49]) + c[1]*s[22] + c[0]*s[197] + (-1*c[1]*s[170]) + (-1*c[0]*s[53]) + c[1]*s[26];
gout[46] += + c[0]*s[36] + (-1*c[1]*s[9]) + (-1*c[0]*s[108]) + c[1]*s[81] + c[0]*s[40] + (-1*c[1]*s[13]) + (-1*c[0]*s[112]) + c[1]*s[85] + c[0]*s[44] + (-1*c[1]*s[17]) + (-1*c[0]*s[116]) + c[1]*s[89];
gout[47] += + c[0]*s[27] + (-1*c[1]*s[0]) + c[0]*s[117] + (-1*c[1]*s[90]) + c[0]*s[207] + (-1*c[1]*s[180]) + c[0]*s[31] + (-1*c[1]*s[4]) + c[0]*s[121] + (-1*c[1]*s[94]) + c[0]*s[211] + (-1*c[1]*s[184]) + c[0]*s[35] + (-1*c[1]*s[8]) + c[0]*s[125] + (-1*c[1]*s[98]) + c[0]*s[215] + (-1*c[1]*s[188]);
gout += 48;
}}}
void cint2e_spgsp1spsp2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {2, 1, 1, 1, 5, 4, 4, 3};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_spgsp1spsp2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {2, 1, 1, 1, 5, 4, 4, 3};
const FINT i_sh = shls[0];
const FINT j_sh = shls[1];
const FINT k_sh = shls[2];
const FINT l_sh = shls[3];
if (bas(ATOM_OF, i_sh) == bas(ATOM_OF, j_sh)) {
FINT ip = CINTlen_spinor(i_sh, bas) * bas(NCTR_OF,i_sh);
FINT jp = CINTlen_spinor(j_sh, bas) * bas(NCTR_OF,j_sh);
FINT kp = CINTlen_spinor(k_sh, bas) * bas(NCTR_OF,k_sh);
FINT lp = CINTlen_spinor(l_sh, bas) * bas(NCTR_OF,l_sh);
CINTdset0(kp * ip * jp * lp * OF_CMPLX * ng[TENSOR], opijkl);
return 0; }
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_spgsp1spsp2;
envs.common_factor *= 0.5;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1i, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_spgsp1spsp2_optimizer);
C2Fo_(cint2e_spgsp1spsp2)
/* <SIGMA DOT P i|OVLP |SIGMA DOT P SIGMA DOT P j> */
static void CINTgout1e_cint1e_spspsp(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double *g3 = g2  + envs->g_size * 3;
double *g4 = g3  + envs->g_size * 3;
double *g5 = g4  + envs->g_size * 3;
double *g6 = g5  + envs->g_size * 3;
double *g7 = g6  + envs->g_size * 3;
double *g8 = g7  + envs->g_size * 3;
double s[27];
G1E_D_J(g1, g0, i_l+1, j_l+0);
G1E_D_J(g2, g0, i_l+1, j_l+1);
G1E_D_J(g3, g2, i_l+1, j_l+0);
G1E_D_I(g4, g0, i_l+0, j_l);
G1E_D_I(g5, g1, i_l+0, j_l);
G1E_D_I(g6, g2, i_l+0, j_l);
G1E_D_I(g7, g3, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g7[ix] * g0[iy] * g0[iz];
s[1] = g6[ix] * g1[iy] * g0[iz];
s[2] = g6[ix] * g0[iy] * g1[iz];
s[3] = g5[ix] * g2[iy] * g0[iz];
s[4] = g4[ix] * g3[iy] * g0[iz];
s[5] = g4[ix] * g2[iy] * g1[iz];
s[6] = g5[ix] * g0[iy] * g2[iz];
s[7] = g4[ix] * g1[iy] * g2[iz];
s[8] = g4[ix] * g0[iy] * g3[iz];
s[9] = g3[ix] * g4[iy] * g0[iz];
s[10] = g2[ix] * g5[iy] * g0[iz];
s[11] = g2[ix] * g4[iy] * g1[iz];
s[12] = g1[ix] * g6[iy] * g0[iz];
s[13] = g0[ix] * g7[iy] * g0[iz];
s[14] = g0[ix] * g6[iy] * g1[iz];
s[15] = g1[ix] * g4[iy] * g2[iz];
s[16] = g0[ix] * g5[iy] * g2[iz];
s[17] = g0[ix] * g4[iy] * g3[iz];
s[18] = g3[ix] * g0[iy] * g4[iz];
s[19] = g2[ix] * g1[iy] * g4[iz];
s[20] = g2[ix] * g0[iy] * g5[iz];
s[21] = g1[ix] * g2[iy] * g4[iz];
s[22] = g0[ix] * g3[iy] * g4[iz];
s[23] = g0[ix] * g2[iy] * g5[iz];
s[24] = g1[ix] * g0[iy] * g6[iz];
s[25] = g0[ix] * g1[iy] * g6[iz];
s[26] = g0[ix] * g0[iy] * g7[iz];
gout[0] += + (-1*s[0]) + (-1*s[12]) + (-1*s[24]);
gout[1] += + (-1*s[1]) + (-1*s[13]) + (-1*s[25]);
gout[2] += + (-1*s[2]) + (-1*s[14]) + (-1*s[26]);
gout[3] += 0;
gout += 4;
}}
FINT cint1e_spspsp(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 2, 0, 0, 3, 4, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_spspsp;
return CINT1e_drv(opij, &envs, 1, &c2s_si_1e); }
C2F_(cint1e_spspsp)
/* <SIGMA DOT P i|NUC |j> */
static void CINTgout1e_cint1e_spnuc(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
FINT ix, iy, iz, n;
double *g0 = g;
double *g1 = g0  + envs->g_size * 3;
double *g2 = g1  + envs->g_size * 3;
double s[3];
G1E_D_I(g1, g0, i_l+0, j_l);
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
s[0] = g1[ix] * g0[iy] * g0[iz];
s[1] = g0[ix] * g1[iy] * g0[iz];
s[2] = g0[ix] * g0[iy] * g1[iz];
gout[0] += + s[0];
gout[1] += + s[1];
gout[2] += + s[2];
gout[3] += 0;
gout += 4;
}}
FINT cint1e_spnuc(double *opij, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 0, 0, 1, 4, 0, 1};
CINTEnvVars envs;
CINTinit_int1e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout1e_cint1e_spnuc;
return CINT1e_nuc_drv(opij, &envs, 1, &c2s_si_1e); }
C2F_(cint1e_spnuc)
/* <k SIGMA DOT P i|R12 |j l> : i,j \in electron 1; k,l \in electron 2
 * = (SIGMA DOT P i j|R12 |k l) */
static void CINTgout2e_cint2e_spv1(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double s[3];
G2E_D_I(g1, g0, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g1[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g0[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g0[ix+0]*g0[iy+0]*g1[iz+0];
break;
case 2:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
break;
case 3:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g1[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g0[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g0[ix+2] * g0[iy+2] * g1[iz+2];
break;
case 4:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
break;
case 5:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g1[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g0[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g0[ix+4] * g0[iy+4] * g1[iz+4];
break;
case 6:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
break;
case 7:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g1[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g0[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g0[ix+6] * g0[iy+6] * g1[iz+6];
break;
default:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g1[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g0[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g0[ix+i] * g0[iy+i] * g1[iz+i];}
break;}
if (gout_empty) {
gout[0] = + s[0];
gout[1] = + s[1];
gout[2] = + s[2];
gout[3] = 0;
gout += 4;
} else {
gout[0] += + s[0];
gout[1] += + s[1];
gout[2] += + s[2];
gout[3] += 0;
gout += 4;
}}}
void cint2e_spv1_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 0, 0, 1, 4, 1, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_spv1(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 0, 0, 0, 1, 4, 1, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_spv1;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_sf_2e2);
}
OPTIMIZER2F_(cint2e_spv1_optimizer);
C2Fo_(cint2e_spv1)
/* <k i|R12 |SIGMA DOT P j l> : i,j \in electron 1; k,l \in electron 2
 * = (i SIGMA DOT P j|R12 |k l) */
static void CINTgout2e_cint2e_vsp1(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double s[3];
G2E_D_J(g1, g0, i_l+0, j_l+0, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g1[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g0[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g0[ix+0]*g0[iy+0]*g1[iz+0];
break;
case 2:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
break;
case 3:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g1[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g0[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g0[ix+2] * g0[iy+2] * g1[iz+2];
break;
case 4:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
break;
case 5:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g1[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g0[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g0[ix+4] * g0[iy+4] * g1[iz+4];
break;
case 6:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
break;
case 7:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g1[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g0[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g0[ix+6] * g0[iy+6] * g1[iz+6];
break;
default:
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g1[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g0[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g0[ix+i] * g0[iy+i] * g1[iz+i];}
break;}
if (gout_empty) {
gout[0] = + (-1*s[0]);
gout[1] = + (-1*s[1]);
gout[2] = + (-1*s[2]);
gout[3] = 0;
gout += 4;
} else {
gout[0] += + (-1*s[0]);
gout[1] += + (-1*s[1]);
gout[2] += + (-1*s[2]);
gout[3] += 0;
gout += 4;
}}}
void cint2e_vsp1_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {0, 1, 0, 0, 1, 4, 1, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_vsp1(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {0, 1, 0, 0, 1, 4, 1, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_vsp1;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_sf_2e2);
}
OPTIMIZER2F_(cint2e_vsp1_optimizer);
C2Fo_(cint2e_vsp1)
/* <SIGMA DOT P k i|R12 |j SIGMA DOT P l> : i,j \in electron 1; k,l \in electron 2
 * = (i j|R12 |SIGMA DOT P k SIGMA DOT P l) */
static void CINTgout2e_cint2e_spsp2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double s[9];
G2E_D_L(g1, g0, i_l+0, j_l+0, k_l+1, l_l+0);
G2E_D_K(g2, g0, i_l+0, j_l+0, k_l+0, l_l);
G2E_D_K(g3, g1, i_l+0, j_l+0, k_l+0, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g3[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g2[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g2[ix+0]*g0[iy+0]*g1[iz+0];
s[3] = + g1[ix+0]*g2[iy+0]*g0[iz+0];
s[4] = + g0[ix+0]*g3[iy+0]*g0[iz+0];
s[5] = + g0[ix+0]*g2[iy+0]*g1[iz+0];
s[6] = + g1[ix+0]*g0[iy+0]*g2[iz+0];
s[7] = + g0[ix+0]*g1[iy+0]*g2[iz+0];
s[8] = + g0[ix+0]*g0[iy+0]*g3[iz+0];
break;
case 2:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 3:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+2] * g0[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+2] * g2[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+2] * g3[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+2] * g2[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+2] * g0[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+2] * g1[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+2] * g0[iy+2] * g3[iz+2];
break;
case 4:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 5:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+4] * g0[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+4] * g2[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+4] * g3[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+4] * g2[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+4] * g0[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+4] * g1[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+4] * g0[iy+4] * g3[iz+4];
break;
case 6:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 7:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+6] * g0[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+6] * g2[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+6] * g3[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+6] * g2[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+6] * g0[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+6] * g1[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+6] * g0[iy+6] * g3[iz+6];
break;
default:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g3[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g2[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g2[ix+i] * g0[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[3] += g1[ix+i] * g2[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g3+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[4] += g0[ix+i] * g3[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[5] += g0[ix+i] * g2[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[6] += g1[ix+i] * g0[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[7] += g0[ix+i] * g1[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[8] += g0[ix+i] * g0[iy+i] * g3[iz+i];}
break;}
if (gout_empty) {
gout[0] = + s[5] + (-1*s[7]);
gout[1] = + s[6] + (-1*s[2]);
gout[2] = + s[1] + (-1*s[3]);
gout[3] = + s[0] + s[4] + s[8];
gout += 4;
} else {
gout[0] += + s[5] + (-1*s[7]);
gout[1] += + s[6] + (-1*s[2]);
gout[2] += + s[1] + (-1*s[3]);
gout[3] += + s[0] + s[4] + s[8];
gout += 4;
}}}
void cint2e_spsp2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {0, 0, 1, 1, 2, 1, 4, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_spsp2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {0, 0, 1, 1, 2, 1, 4, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_spsp2;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_sf_2e1, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_spsp2_optimizer);
C2Fo_(cint2e_spsp2)
/* <SIGMA DOT P k SIGMA DOT P i|R12 |j l> : i,j \in electron 1; k,l \in electron 2
 * = (SIGMA DOT P i j|R12 |SIGMA DOT P k l) */
static void CINTgout2e_cint2e_spv1spv2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double s[9];
G2E_D_K(g1, g0, i_l+1, j_l+0, k_l+0, l_l);
G2E_D_I(g2, g0, i_l+0, j_l, k_l, l_l);
G2E_D_I(g3, g1, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g3[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g2[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g2[ix+0]*g0[iy+0]*g1[iz+0];
s[3] = + g1[ix+0]*g2[iy+0]*g0[iz+0];
s[4] = + g0[ix+0]*g3[iy+0]*g0[iz+0];
s[5] = + g0[ix+0]*g2[iy+0]*g1[iz+0];
s[6] = + g1[ix+0]*g0[iy+0]*g2[iz+0];
s[7] = + g0[ix+0]*g1[iy+0]*g2[iz+0];
s[8] = + g0[ix+0]*g0[iy+0]*g3[iz+0];
break;
case 2:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 3:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+2] * g0[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+2] * g2[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+2] * g3[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+2] * g2[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+2] * g0[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+2] * g1[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+2] * g0[iy+2] * g3[iz+2];
break;
case 4:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 5:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+4] * g0[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+4] * g2[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+4] * g3[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+4] * g2[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+4] * g0[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+4] * g1[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+4] * g0[iy+4] * g3[iz+4];
break;
case 6:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 7:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+6] * g0[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+6] * g2[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+6] * g3[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+6] * g2[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+6] * g0[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+6] * g1[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+6] * g0[iy+6] * g3[iz+6];
break;
default:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g3[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g2[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g2[ix+i] * g0[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[3] += g1[ix+i] * g2[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g3+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[4] += g0[ix+i] * g3[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[5] += g0[ix+i] * g2[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[6] += g1[ix+i] * g0[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[7] += g0[ix+i] * g1[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[8] += g0[ix+i] * g0[iy+i] * g3[iz+i];}
break;}
if (gout_empty) {
gout[0] = + s[0];
gout[1] = + s[3];
gout[2] = + s[6];
gout[3] = 0;
gout[4] = + s[1];
gout[5] = + s[4];
gout[6] = + s[7];
gout[7] = 0;
gout[8] = + s[2];
gout[9] = + s[5];
gout[10] = + s[8];
gout[11] = 0;
gout[12] = 0;
gout[13] = 0;
gout[14] = 0;
gout[15] = 0;
gout += 16;
} else {
gout[0] += + s[0];
gout[1] += + s[3];
gout[2] += + s[6];
gout[3] += 0;
gout[4] += + s[1];
gout[5] += + s[4];
gout[6] += + s[7];
gout[7] += 0;
gout[8] += + s[2];
gout[9] += + s[5];
gout[10] += + s[8];
gout[11] += 0;
gout[12] += 0;
gout[13] += 0;
gout[14] += 0;
gout[15] += 0;
gout += 16;
}}}
void cint2e_spv1spv2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 1, 0, 2, 4, 4, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_spv1spv2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 0, 1, 0, 2, 4, 4, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_spv1spv2;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_spv1spv2_optimizer);
C2Fo_(cint2e_spv1spv2)
/* <SIGMA DOT P k i|R12 |SIGMA DOT P j l> : i,j \in electron 1; k,l \in electron 2
 * = (i SIGMA DOT P j|R12 |SIGMA DOT P k l) */
static void CINTgout2e_cint2e_vsp1spv2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double s[9];
G2E_D_K(g1, g0, i_l+0, j_l+1, k_l+0, l_l);
G2E_D_J(g2, g0, i_l+0, j_l+0, k_l, l_l);
G2E_D_J(g3, g1, i_l+0, j_l+0, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g3[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g2[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g2[ix+0]*g0[iy+0]*g1[iz+0];
s[3] = + g1[ix+0]*g2[iy+0]*g0[iz+0];
s[4] = + g0[ix+0]*g3[iy+0]*g0[iz+0];
s[5] = + g0[ix+0]*g2[iy+0]*g1[iz+0];
s[6] = + g1[ix+0]*g0[iy+0]*g2[iz+0];
s[7] = + g0[ix+0]*g1[iy+0]*g2[iz+0];
s[8] = + g0[ix+0]*g0[iy+0]*g3[iz+0];
break;
case 2:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 3:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+2] * g0[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+2] * g2[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+2] * g3[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+2] * g2[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+2] * g0[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+2] * g1[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+2] * g0[iy+2] * g3[iz+2];
break;
case 4:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 5:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+4] * g0[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+4] * g2[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+4] * g3[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+4] * g2[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+4] * g0[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+4] * g1[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+4] * g0[iy+4] * g3[iz+4];
break;
case 6:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 7:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+6] * g0[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+6] * g2[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+6] * g3[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+6] * g2[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+6] * g0[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+6] * g1[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+6] * g0[iy+6] * g3[iz+6];
break;
default:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g3[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g2[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g2[ix+i] * g0[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[3] += g1[ix+i] * g2[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g3+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[4] += g0[ix+i] * g3[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[5] += g0[ix+i] * g2[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[6] += g1[ix+i] * g0[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[7] += g0[ix+i] * g1[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[8] += g0[ix+i] * g0[iy+i] * g3[iz+i];}
break;}
if (gout_empty) {
gout[0] = + (-1*s[0]);
gout[1] = + (-1*s[3]);
gout[2] = + (-1*s[6]);
gout[3] = 0;
gout[4] = + (-1*s[1]);
gout[5] = + (-1*s[4]);
gout[6] = + (-1*s[7]);
gout[7] = 0;
gout[8] = + (-1*s[2]);
gout[9] = + (-1*s[5]);
gout[10] = + (-1*s[8]);
gout[11] = 0;
gout[12] = 0;
gout[13] = 0;
gout[14] = 0;
gout[15] = 0;
gout += 16;
} else {
gout[0] += + (-1*s[0]);
gout[1] += + (-1*s[3]);
gout[2] += + (-1*s[6]);
gout[3] += 0;
gout[4] += + (-1*s[1]);
gout[5] += + (-1*s[4]);
gout[6] += + (-1*s[7]);
gout[7] += 0;
gout[8] += + (-1*s[2]);
gout[9] += + (-1*s[5]);
gout[10] += + (-1*s[8]);
gout[11] += 0;
gout[12] += 0;
gout[13] += 0;
gout[14] += 0;
gout[15] += 0;
gout += 16;
}}}
void cint2e_vsp1spv2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {0, 1, 1, 0, 2, 4, 4, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_vsp1spv2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {0, 1, 1, 0, 2, 4, 4, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_vsp1spv2;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_vsp1spv2_optimizer);
C2Fo_(cint2e_vsp1spv2)
/* <k SIGMA DOT P i|R12 |j SIGMA DOT P l> : i,j \in electron 1; k,l \in electron 2
 * = (SIGMA DOT P i j|R12 |k SIGMA DOT P l) */
static void CINTgout2e_cint2e_spv1vsp2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double s[9];
G2E_D_L(g1, g0, i_l+1, j_l+0, k_l+0, l_l+0);
G2E_D_I(g2, g0, i_l+0, j_l, k_l, l_l);
G2E_D_I(g3, g1, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g3[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g2[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g2[ix+0]*g0[iy+0]*g1[iz+0];
s[3] = + g1[ix+0]*g2[iy+0]*g0[iz+0];
s[4] = + g0[ix+0]*g3[iy+0]*g0[iz+0];
s[5] = + g0[ix+0]*g2[iy+0]*g1[iz+0];
s[6] = + g1[ix+0]*g0[iy+0]*g2[iz+0];
s[7] = + g0[ix+0]*g1[iy+0]*g2[iz+0];
s[8] = + g0[ix+0]*g0[iy+0]*g3[iz+0];
break;
case 2:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 3:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+2] * g0[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+2] * g2[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+2] * g3[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+2] * g2[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+2] * g0[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+2] * g1[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+2] * g0[iy+2] * g3[iz+2];
break;
case 4:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 5:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+4] * g0[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+4] * g2[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+4] * g3[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+4] * g2[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+4] * g0[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+4] * g1[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+4] * g0[iy+4] * g3[iz+4];
break;
case 6:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 7:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+6] * g0[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+6] * g2[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+6] * g3[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+6] * g2[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+6] * g0[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+6] * g1[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+6] * g0[iy+6] * g3[iz+6];
break;
default:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g3[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g2[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g2[ix+i] * g0[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[3] += g1[ix+i] * g2[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g3+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[4] += g0[ix+i] * g3[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[5] += g0[ix+i] * g2[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[6] += g1[ix+i] * g0[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[7] += g0[ix+i] * g1[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[8] += g0[ix+i] * g0[iy+i] * g3[iz+i];}
break;}
if (gout_empty) {
gout[0] = + (-1*s[0]);
gout[1] = + (-1*s[3]);
gout[2] = + (-1*s[6]);
gout[3] = 0;
gout[4] = + (-1*s[1]);
gout[5] = + (-1*s[4]);
gout[6] = + (-1*s[7]);
gout[7] = 0;
gout[8] = + (-1*s[2]);
gout[9] = + (-1*s[5]);
gout[10] = + (-1*s[8]);
gout[11] = 0;
gout[12] = 0;
gout[13] = 0;
gout[14] = 0;
gout[15] = 0;
gout += 16;
} else {
gout[0] += + (-1*s[0]);
gout[1] += + (-1*s[3]);
gout[2] += + (-1*s[6]);
gout[3] += 0;
gout[4] += + (-1*s[1]);
gout[5] += + (-1*s[4]);
gout[6] += + (-1*s[7]);
gout[7] += 0;
gout[8] += + (-1*s[2]);
gout[9] += + (-1*s[5]);
gout[10] += + (-1*s[8]);
gout[11] += 0;
gout[12] += 0;
gout[13] += 0;
gout[14] += 0;
gout[15] += 0;
gout += 16;
}}}
void cint2e_spv1vsp2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 0, 1, 2, 4, 4, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_spv1vsp2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 0, 0, 1, 2, 4, 4, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_spv1vsp2;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_spv1vsp2_optimizer);
C2Fo_(cint2e_spv1vsp2)
/* <k i|R12 |SIGMA DOT P j SIGMA DOT P l> : i,j \in electron 1; k,l \in electron 2
 * = (i SIGMA DOT P j|R12 |k SIGMA DOT P l) */
static void CINTgout2e_cint2e_vsp1vsp2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double s[9];
G2E_D_L(g1, g0, i_l+0, j_l+1, k_l+0, l_l+0);
G2E_D_J(g2, g0, i_l+0, j_l+0, k_l, l_l);
G2E_D_J(g3, g1, i_l+0, j_l+0, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
switch (envs->nrys_roots) {
case 1:
s[0] = + g3[ix+0]*g0[iy+0]*g0[iz+0];
s[1] = + g2[ix+0]*g1[iy+0]*g0[iz+0];
s[2] = + g2[ix+0]*g0[iy+0]*g1[iz+0];
s[3] = + g1[ix+0]*g2[iy+0]*g0[iz+0];
s[4] = + g0[ix+0]*g3[iy+0]*g0[iz+0];
s[5] = + g0[ix+0]*g2[iy+0]*g1[iz+0];
s[6] = + g1[ix+0]*g0[iy+0]*g2[iz+0];
s[7] = + g0[ix+0]*g1[iy+0]*g2[iz+0];
s[8] = + g0[ix+0]*g0[iy+0]*g3[iz+0];
break;
case 2:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 3:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+2] * g0[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+2] * g1[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+2] * g0[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+2] * g2[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+2] * g3[iy+2] * g0[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+2] * g2[iy+2] * g1[iz+2];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+2] * g0[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+2] * g1[iy+2] * g2[iz+2];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+2] * g0[iy+2] * g3[iz+2];
break;
case 4:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 5:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+4] * g0[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+4] * g1[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+4] * g0[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+4] * g2[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+4] * g3[iy+4] * g0[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+4] * g2[iy+4] * g1[iz+4];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+4] * g0[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+4] * g1[iy+4] * g2[iz+4];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+4] * g0[iy+4] * g3[iz+4];
break;
case 6:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
break;
case 7:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
s[0] += g3[ix+6] * g0[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
s[1] += g2[ix+6] * g1[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
s[2] += g2[ix+6] * g0[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
s[3] += g1[ix+6] * g2[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
s[4] += g0[ix+6] * g3[iy+6] * g0[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
s[5] += g0[ix+6] * g2[iy+6] * g1[iz+6];
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
s[6] += g1[ix+6] * g0[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
s[7] += g0[ix+6] * g1[iy+6] * g2[iz+6];
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
s[8] += g0[ix+6] * g0[iy+6] * g3[iz+6];
break;
default:
r0 = _mm_load_pd(g3+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g3+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g3+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+0, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[0] += g3[ix+i] * g0[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+1, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[1] += g2[ix+i] * g1[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g2+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g2+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g2+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+2, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[2] += g2[ix+i] * g0[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+3, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[3] += g1[ix+i] * g2[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g3+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g0+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g3+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g3+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g3+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g0+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+4, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[4] += g0[ix+i] * g3[iy+i] * g0[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g2+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g1+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g2+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g2+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g2+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g1+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+5, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[5] += g0[ix+i] * g2[iy+i] * g1[iz+i];}
r0 = _mm_load_pd(g1+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g1+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g1+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+6, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[6] += g1[ix+i] * g0[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g1+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g2+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g1+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g1+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g1+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g2+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+7, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[7] += g0[ix+i] * g1[iy+i] * g2[iz+i];}
r0 = _mm_load_pd(g0+ix  );
r1 = _mm_load_pd(g0+iy  );
r0 = _mm_mul_pd(r0, r1);
r1 = _mm_load_pd(g3+iz  );
r0 = _mm_mul_pd(r0, r1);
r2 = _mm_load_pd(g0+ix+2);
r1 = _mm_load_pd(g0+iy+2);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+2);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+4);
r1 = _mm_load_pd(g0+iy+4);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+4);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r2 = _mm_load_pd(g0+ix+6);
r1 = _mm_load_pd(g0+iy+6);
r2 = _mm_mul_pd(r2, r1);
r1 = _mm_load_pd(g3+iz+6);
r2 = _mm_mul_pd(r2, r1);
r0 = _mm_add_pd(r0, r2);
r0 = _mm_hadd_pd(r0, r0);
_mm_storeh_pd(s+8, r0);
for (i = 8; i < envs->nrys_roots; i++) {
s[8] += g0[ix+i] * g0[iy+i] * g3[iz+i];}
break;}
if (gout_empty) {
gout[0] = + s[0];
gout[1] = + s[3];
gout[2] = + s[6];
gout[3] = 0;
gout[4] = + s[1];
gout[5] = + s[4];
gout[6] = + s[7];
gout[7] = 0;
gout[8] = + s[2];
gout[9] = + s[5];
gout[10] = + s[8];
gout[11] = 0;
gout[12] = 0;
gout[13] = 0;
gout[14] = 0;
gout[15] = 0;
gout += 16;
} else {
gout[0] += + s[0];
gout[1] += + s[3];
gout[2] += + s[6];
gout[3] += 0;
gout[4] += + s[1];
gout[5] += + s[4];
gout[6] += + s[7];
gout[7] += 0;
gout[8] += + s[2];
gout[9] += + s[5];
gout[10] += + s[8];
gout[11] += 0;
gout[12] += 0;
gout[13] += 0;
gout[14] += 0;
gout[15] += 0;
gout += 16;
}}}
void cint2e_vsp1vsp2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {0, 1, 0, 1, 2, 4, 4, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_vsp1vsp2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {0, 1, 0, 1, 2, 4, 4, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_vsp1vsp2;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_vsp1vsp2_optimizer);
C2Fo_(cint2e_vsp1vsp2)
/* <SIGMA DOT P k SIGMA DOT P i|R12 |j SIGMA DOT P l> : i,j \in electron 1; k,l \in electron 2
 * = (SIGMA DOT P i j|R12 |SIGMA DOT P k SIGMA DOT P l) */
static void CINTgout2e_cint2e_spv1spsp2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double *g5 = g4 + envs->g_size * 3;
double *g6 = g5 + envs->g_size * 3;
double *g7 = g6 + envs->g_size * 3;
double *g8 = g7 + envs->g_size * 3;
double s[27];
G2E_D_L(g1, g0, i_l+1, j_l+0, k_l+1, l_l+0);
G2E_D_K(g2, g0, i_l+1, j_l+0, k_l+0, l_l);
G2E_D_K(g3, g1, i_l+1, j_l+0, k_l+0, l_l);
G2E_D_I(g4, g0, i_l+0, j_l, k_l, l_l);
G2E_D_I(g5, g1, i_l+0, j_l, k_l, l_l);
G2E_D_I(g6, g2, i_l+0, j_l, k_l, l_l);
G2E_D_I(g7, g3, i_l+0, j_l, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
CINTdset0(27, s);
for (i = 0; i < envs->nrys_roots; i++) {
s[0] += g7[ix+i] * g0[iy+i] * g0[iz+i];
s[1] += g6[ix+i] * g1[iy+i] * g0[iz+i];
s[2] += g6[ix+i] * g0[iy+i] * g1[iz+i];
s[3] += g5[ix+i] * g2[iy+i] * g0[iz+i];
s[4] += g4[ix+i] * g3[iy+i] * g0[iz+i];
s[5] += g4[ix+i] * g2[iy+i] * g1[iz+i];
s[6] += g5[ix+i] * g0[iy+i] * g2[iz+i];
s[7] += g4[ix+i] * g1[iy+i] * g2[iz+i];
s[8] += g4[ix+i] * g0[iy+i] * g3[iz+i];
s[9] += g3[ix+i] * g4[iy+i] * g0[iz+i];
s[10] += g2[ix+i] * g5[iy+i] * g0[iz+i];
s[11] += g2[ix+i] * g4[iy+i] * g1[iz+i];
s[12] += g1[ix+i] * g6[iy+i] * g0[iz+i];
s[13] += g0[ix+i] * g7[iy+i] * g0[iz+i];
s[14] += g0[ix+i] * g6[iy+i] * g1[iz+i];
s[15] += g1[ix+i] * g4[iy+i] * g2[iz+i];
s[16] += g0[ix+i] * g5[iy+i] * g2[iz+i];
s[17] += g0[ix+i] * g4[iy+i] * g3[iz+i];
s[18] += g3[ix+i] * g0[iy+i] * g4[iz+i];
s[19] += g2[ix+i] * g1[iy+i] * g4[iz+i];
s[20] += g2[ix+i] * g0[iy+i] * g5[iz+i];
s[21] += g1[ix+i] * g2[iy+i] * g4[iz+i];
s[22] += g0[ix+i] * g3[iy+i] * g4[iz+i];
s[23] += g0[ix+i] * g2[iy+i] * g5[iz+i];
s[24] += g1[ix+i] * g0[iy+i] * g6[iz+i];
s[25] += g0[ix+i] * g1[iy+i] * g6[iz+i];
s[26] += g0[ix+i] * g0[iy+i] * g7[iz+i];
}
if (gout_empty) {
gout[0] = + s[5] + (-1*s[7]);
gout[1] = + s[14] + (-1*s[16]);
gout[2] = + s[23] + (-1*s[25]);
gout[3] = 0;
gout[4] = + s[6] + (-1*s[2]);
gout[5] = + s[15] + (-1*s[11]);
gout[6] = + s[24] + (-1*s[20]);
gout[7] = 0;
gout[8] = + s[1] + (-1*s[3]);
gout[9] = + s[10] + (-1*s[12]);
gout[10] = + s[19] + (-1*s[21]);
gout[11] = 0;
gout[12] = + s[0] + s[4] + s[8];
gout[13] = + s[9] + s[13] + s[17];
gout[14] = + s[18] + s[22] + s[26];
gout[15] = 0;
gout += 16;
} else {
gout[0] += + s[5] + (-1*s[7]);
gout[1] += + s[14] + (-1*s[16]);
gout[2] += + s[23] + (-1*s[25]);
gout[3] += 0;
gout[4] += + s[6] + (-1*s[2]);
gout[5] += + s[15] + (-1*s[11]);
gout[6] += + s[24] + (-1*s[20]);
gout[7] += 0;
gout[8] += + s[1] + (-1*s[3]);
gout[9] += + s[10] + (-1*s[12]);
gout[10] += + s[19] + (-1*s[21]);
gout[11] += 0;
gout[12] += + s[0] + s[4] + s[8];
gout[13] += + s[9] + s[13] + s[17];
gout[14] += + s[18] + s[22] + s[26];
gout[15] += 0;
gout += 16;
}}}
void cint2e_spv1spsp2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {1, 0, 1, 1, 3, 4, 4, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_spv1spsp2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {1, 0, 1, 1, 3, 4, 4, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_spv1spsp2;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_spv1spsp2_optimizer);
C2Fo_(cint2e_spv1spsp2)
/* <SIGMA DOT P k i|R12 |SIGMA DOT P j SIGMA DOT P l> : i,j \in electron 1; k,l \in electron 2
 * = (i SIGMA DOT P j|R12 |SIGMA DOT P k SIGMA DOT P l) */
static void CINTgout2e_cint2e_vsp1spsp2(double *g,
double *gout, const FINT *idx, const CINTEnvVars *envs, FINT gout_empty) {
const double *env = envs->env;
const FINT nf = envs->nf;
const FINT i_l = envs->i_l;
const FINT j_l = envs->j_l;
const FINT k_l = envs->k_l;
const FINT l_l = envs->l_l;
const double *ri = envs->ri;
const double *rj = envs->rj;
const double *rk = envs->rk;
const double *rl = envs->rl;
FINT ix, iy, iz, i, n;
double *g0 = g;
double *g1 = g0 + envs->g_size * 3;
double *g2 = g1 + envs->g_size * 3;
double *g3 = g2 + envs->g_size * 3;
double *g4 = g3 + envs->g_size * 3;
double *g5 = g4 + envs->g_size * 3;
double *g6 = g5 + envs->g_size * 3;
double *g7 = g6 + envs->g_size * 3;
double *g8 = g7 + envs->g_size * 3;
double s[27];
G2E_D_L(g1, g0, i_l+0, j_l+1, k_l+1, l_l+0);
G2E_D_K(g2, g0, i_l+0, j_l+1, k_l+0, l_l);
G2E_D_K(g3, g1, i_l+0, j_l+1, k_l+0, l_l);
G2E_D_J(g4, g0, i_l+0, j_l+0, k_l, l_l);
G2E_D_J(g5, g1, i_l+0, j_l+0, k_l, l_l);
G2E_D_J(g6, g2, i_l+0, j_l+0, k_l, l_l);
G2E_D_J(g7, g3, i_l+0, j_l+0, k_l, l_l);
__m128d r0, r1, r2, r3;
for (n = 0; n < nf; n++, idx+=3) {
ix = idx[0];
iy = idx[1];
iz = idx[2];
CINTdset0(27, s);
for (i = 0; i < envs->nrys_roots; i++) {
s[0] += g7[ix+i] * g0[iy+i] * g0[iz+i];
s[1] += g6[ix+i] * g1[iy+i] * g0[iz+i];
s[2] += g6[ix+i] * g0[iy+i] * g1[iz+i];
s[3] += g5[ix+i] * g2[iy+i] * g0[iz+i];
s[4] += g4[ix+i] * g3[iy+i] * g0[iz+i];
s[5] += g4[ix+i] * g2[iy+i] * g1[iz+i];
s[6] += g5[ix+i] * g0[iy+i] * g2[iz+i];
s[7] += g4[ix+i] * g1[iy+i] * g2[iz+i];
s[8] += g4[ix+i] * g0[iy+i] * g3[iz+i];
s[9] += g3[ix+i] * g4[iy+i] * g0[iz+i];
s[10] += g2[ix+i] * g5[iy+i] * g0[iz+i];
s[11] += g2[ix+i] * g4[iy+i] * g1[iz+i];
s[12] += g1[ix+i] * g6[iy+i] * g0[iz+i];
s[13] += g0[ix+i] * g7[iy+i] * g0[iz+i];
s[14] += g0[ix+i] * g6[iy+i] * g1[iz+i];
s[15] += g1[ix+i] * g4[iy+i] * g2[iz+i];
s[16] += g0[ix+i] * g5[iy+i] * g2[iz+i];
s[17] += g0[ix+i] * g4[iy+i] * g3[iz+i];
s[18] += g3[ix+i] * g0[iy+i] * g4[iz+i];
s[19] += g2[ix+i] * g1[iy+i] * g4[iz+i];
s[20] += g2[ix+i] * g0[iy+i] * g5[iz+i];
s[21] += g1[ix+i] * g2[iy+i] * g4[iz+i];
s[22] += g0[ix+i] * g3[iy+i] * g4[iz+i];
s[23] += g0[ix+i] * g2[iy+i] * g5[iz+i];
s[24] += g1[ix+i] * g0[iy+i] * g6[iz+i];
s[25] += g0[ix+i] * g1[iy+i] * g6[iz+i];
s[26] += g0[ix+i] * g0[iy+i] * g7[iz+i];
}
if (gout_empty) {
gout[0] = + (-1*s[5]) + s[7];
gout[1] = + (-1*s[14]) + s[16];
gout[2] = + (-1*s[23]) + s[25];
gout[3] = 0;
gout[4] = + (-1*s[6]) + s[2];
gout[5] = + (-1*s[15]) + s[11];
gout[6] = + (-1*s[24]) + s[20];
gout[7] = 0;
gout[8] = + (-1*s[1]) + s[3];
gout[9] = + (-1*s[10]) + s[12];
gout[10] = + (-1*s[19]) + s[21];
gout[11] = 0;
gout[12] = + (-1*s[0]) + (-1*s[4]) + (-1*s[8]);
gout[13] = + (-1*s[9]) + (-1*s[13]) + (-1*s[17]);
gout[14] = + (-1*s[18]) + (-1*s[22]) + (-1*s[26]);
gout[15] = 0;
gout += 16;
} else {
gout[0] += + (-1*s[5]) + s[7];
gout[1] += + (-1*s[14]) + s[16];
gout[2] += + (-1*s[23]) + s[25];
gout[3] += 0;
gout[4] += + (-1*s[6]) + s[2];
gout[5] += + (-1*s[15]) + s[11];
gout[6] += + (-1*s[24]) + s[20];
gout[7] += 0;
gout[8] += + (-1*s[1]) + s[3];
gout[9] += + (-1*s[10]) + s[12];
gout[10] += + (-1*s[19]) + s[21];
gout[11] += 0;
gout[12] += + (-1*s[0]) + (-1*s[4]) + (-1*s[8]);
gout[13] += + (-1*s[9]) + (-1*s[13]) + (-1*s[17]);
gout[14] += + (-1*s[18]) + (-1*s[22]) + (-1*s[26]);
gout[15] += 0;
gout += 16;
}}}
void cint2e_vsp1spsp2_optimizer(CINTOpt **opt, const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env) {
FINT ng[] = {0, 1, 1, 1, 3, 4, 4, 1};
CINTuse_all_optimizer(opt, ng, atm, natm, bas, nbas, env);
}
FINT cint2e_vsp1spsp2(double *opijkl, const FINT *shls,
const FINT *atm, const FINT natm,
const FINT *bas, const FINT nbas, const double *env, CINTOpt *opt) {
FINT ng[] = {0, 1, 1, 1, 3, 4, 4, 1};
CINTEnvVars envs;
CINTinit_int2e_EnvVars(&envs, ng, shls, atm, natm, bas, nbas, env);
envs.f_gout = &CINTgout2e_cint2e_vsp1spsp2;
envs.common_factor *= 1;
return CINT2e_spinor_drv(opijkl, &envs, opt, &c2s_si_2e1, &c2s_si_2e2);
}
OPTIMIZER2F_(cint2e_vsp1spsp2_optimizer);
C2Fo_(cint2e_vsp1spsp2)
